/*******************************************************************************
* FILE NAME: ibuffer7.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of the editing functions of          *
*   the class IBuffer.                                                         *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

/*==============================================================================
  Notes:
    1. Each of these functions returns a pointer to an IBuffer object.
       The resulting buffer contains the edited result.  It may or not
       be a different buffer than the receiver, depending on the nature
       of the change and the use count of the receiver.
    2. It is presumed that the caller has properly adjusted the use
       count for the receiver prior to calling any of these functions
       (that is, the use count is accurate).
    3. It is presumed that the caller will attach the resulting buffer
       to an IString so that the built-in use count in the returned
       buffer is accurate.
    4. Typical usage is as follows:
         ...
         // Preserve old contents.
         IString old( aString.pBuffer );
         ...
         // Reset aString to point to (possibly) new edited buffer:
         aString.pBuffer = aString.pBuffer->xxxxxxx();
         ...
==============================================================================*/
extern "C"
  {
  #include <string.h>
  }

#include <ibuffer.hpp>

#include <istrtest.hpp>


// Segment definitions
#ifdef IC_PAGETUNE
  #define _IBUFFER7_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IBuffer::center                                                              |
|                                                                              |
| If the requested length is equal to the length of the receiver,              |
| then the reciever is returned unmodified.  Otherwise, the                    |
| receiver's length is reset to the argument length and space is               |
| allocated for the centered buffer.  The new buffer is built                  |
| one of two ways, depending on whether the new length is larger               |
| or smaller than the original length of the reciever's buffer.                |
|                                                                              |
| In the former case, the amount of padding required before and                |
| after the centered string is calculated.  Then, the new string               |
| is filled with pad characters, the old contents are copied, and              |
| then the remainder is filled with more pad characters.                       |
|                                                                              |
| In the latter case, a substring of the receiver (the center portion          |
| of the requested length) is copied to the result buffer.                     |
------------------------------------------------------------------------------*/
IBuffer * IBuffer :: center ( unsigned newLength,
                              char     padCharacter )
  {
  IBuffer *result = this;

  if ( newLength != this->length() )
    {
    // Initialize parameters to likely values:
    unsigned prefix       = 0,
             suffix       = 0,
             fromReceiver = this->length(),
             startPos     = 1;

    // Adjust initialization parameters:
    if ( newLength > this->length() )
      prefix = ( newLength - this->length() ) / 2,
      suffix = newLength - this->length() - prefix;
    else
      fromReceiver = newLength,
      startPos     = ( this->length() - fromReceiver ) / 2 + 1;

    // Allocate space and copy receiver to middle:
    result = this->newBuffer( 0,
                              prefix,
                              this->contents() + startPos - 1,
                              fromReceiver,
                              0,
                              suffix,
                              padCharacter );
    }
  else
    this->addRef();

  return result;
  }

// Class to maintain a linked list of occurrences of patterns.
struct Occurrence {
// Pointer to next occurrence.
Occurrence
 *pNext;
// Index of this occurrence.
unsigned int
  pos;
Occurrence ( unsigned int i )
  : pNext(0), pos(i) { }
~Occurrence ( )
  { if ( pNext != 0 ) delete pNext; };
};

/*------------------------------------------------------------------------------
| IBuffer::change                                                              |
|                                                                              |
| This function is implemented by:                                             |
|   1. Making a pass through the receiver, noting the number of occurrences    |
|      of the pattern and the position of each.                                |
|   2. Calculating the new length of the receiver, based on the relative       |
|      lengths of the pattern and replacement strings.                         |
|   3. Filling the new buffer, alternating pieces of the old buffer with the   |
|      new replacement string.                                                 |
|                                                                              |
| The information collected during the the first pass through the receiver is  |
| maintained in a linked list of Occurrence objects.  Each has a data member   |
| to record the position in the receiver of the pattern occurrence and a       |
| pointer to the Occurrence object for the next occurrence of the pattern.     |
------------------------------------------------------------------------------*/
IBuffer * IBuffer :: change ( const char *pPattern,
                              unsigned    patternLen,
                              const char *pReplacement,
                              unsigned    replacementLen,
                              unsigned    startPos,
                              unsigned    numChanges )
  {
  IBuffer *result = this;

  unsigned count = 0; // Number of occurrences.

  // Initialize occurrence chain with a dummy entry at head.
  // pOccurrence is used to step through the chain; it points
  // to the last occurrence during search, current occurrence
  // during pattern replacement.  Note that when "head" is
  // destructed, it will delete all the other Occurrences.
  Occurrence  head        = Occurrence( 0 );
  Occurrence *pOccurrence = &head;

  // Search for occurrences till all (we need) are found:
  while ( count < numChanges )
    if ( ( startPos = this->indexOf( pPattern,
                                     patternLen,
                                     startPos ) ) != 0 )
      {
      // Occurrence found...
      count++;
      pOccurrence = pOccurrence -> pNext = new Occurrence( startPos );
      startPos += patternLen;
      }
    else
       // No more occurrences...
       break;

  if ( count != 0 )
    {
    // Occurrences were found, replace them...
    // Add dummy occurrence at end to ensure tail end of receiver
    // gets copied to destination:
    pOccurrence -> pNext = new Occurrence( this->length() + 1 );
    // Start at first actual occurrence:
    pOccurrence = head.pNext;
    const char *pSource = this->contents() + pOccurrence->pos - 1;

    // Allocate new buffer if size is changing:
    if ( patternLen != replacementLen )
      {
      unsigned newLen = this->length();
      if ( patternLen > replacementLen )
        newLen -= checkMultiplication( patternLen - replacementLen, count );
      else
        newLen = checkAddition( newLen,
                    checkMultiplication(replacementLen - patternLen, count ) );
      // Allocate new buffer and fill with old contents up to
      // first occurrence of the pattern:
      result = this->newBuffer( this->contents(), pSource - this->contents(),
                                0, newLen - ( pSource - this->contents() ),
                                0, 0,
                                0 );
      }
    else
      { // See if we need new buffer for receiver:
      if ( this->useCount() == 1 )
        // Hang on to this buffer.
        this->addRef();
      else
        // Copy shared buffer.
        result = this->newBuffer( this->contents(), this->length() );
      }

    char *pDest = result->contents() + ( pOccurrence -> pos ) - 1;

    // For each occurrence, replace pattern and copy up to next
    // occurrence:
    while ( count-- )
      {
      unsigned previousPos = pOccurrence -> pos;
      pOccurrence = pOccurrence -> pNext;
      memcpy( pDest, pReplacement, replacementLen );
      pDest   += replacementLen;
      pSource += patternLen;
      // Copy source string up to next occurrence:
      unsigned n = ( pOccurrence -> pos - previousPos ) - patternLen;
      if ( patternLen != replacementLen )
        memcpy( pDest, pSource, n );
      pDest   += n;
      pSource += n;
      }
    }
  else
    // No change, reuse receiver:
    this->addRef();

  return result;
  }

/*------------------------------------------------------------------------------
| IBuffer::copy                                                                |
|                                                                              |
| If the number of copies is 1, or the receiver is a null buffer,              |
| then the receiver is returned unmodified.                                    |
|                                                                              |
| If the number of copies is zero, then the result is the null buffer.         |
|                                                                              |
| Otherwise, the actual number of copies to be made is calculated and          |
| space allocated.  The old receiver contents are copied to the new            |
| buffer.  Then, we loop, copying as much as we have (or need) from            |
| the beginning of the buffer to the end.                                      |
------------------------------------------------------------------------------*/
IBuffer * IBuffer :: copy ( unsigned int numCopies )
  {
  IBuffer *result = this;

  if ( this->length() != 0
       &&
       numCopies != 1 )
    { // IBuffer contents will change.
    if ( numCopies )
      {
      // Initialize new buffer with first copy:
      result = this->newBuffer( this->contents(), this->length(),
                                0, checkMultiplication( --numCopies,
                                                        this->length() ) );
      // Make additional copies:
      // Where second copy goes.
      char *pDest = result->contents() + this->length();
      while ( numCopies > 0 )
        { // Duplicate as much as we can from what we've already got:
        unsigned n = (
                     pDest - result->contents() < numCopies * this->length() ) ?
                         pDest - result->contents()
                         :
                         numCopies * this->length();
        memcpy( pDest, result->contents(), n );
        pDest += n;
        numCopies -= n / this->length();
        }
      }
    else
      {
      result = this->null();
      result->addRef();
      }
    }
  else
    // This buffer will be maintained:
    this->addRef();

  return result;
  }

/*------------------------------------------------------------------------------
| IBuffer::insert                                                              |
|                                                                              |
| This function accepts as arguments:                                          |
|   - pointer to string to be inserted                                         |
|   - length  of string to be inserted                                         |
|   - position at which to insert the string                                   |
|   - pad character.                                                           |
|                                                                              |
|  It returns the receiver unmodified if both the string to be                 |
|  inserted is null and the index at which it is to be inserted is             |
|  not off the end of the receiver buffer.                                     |
|                                                                              |
|  Otherwise, space is allocated for the combined buffers.  The                |
|  insertion is accomplished in four steps:                                    |
|                                                                              |
|    1. The portion of the receiver prior to the index at which                |
|       the new string is to be inserted is copied (may be null).              |
|    2. If the receiver isn't this long, then it is padded out                 |
|       to the insert index using the pad character.                           |
|    3. The to-be-inserted string is copied (may be null).                     |
|    4. The portion of the receiver following the index at which               |
|       the new string is to be inserted is copied (may be null).              |
|                                                                              |
|  All this is accomplished with a single call to newBuffer().                 |
------------------------------------------------------------------------------*/
IBuffer * IBuffer :: insert ( const char *pInsert,
                              unsigned    len,
                              unsigned    index,
                              char        padCharacter )
  {
  IBuffer *result = this;

  if ( len != 0
       ||
       index > this->length() )
    {
    unsigned    len1 = ( this->length() < index ) ?
                         this->length()
                         :
                         index,
                len2,
                len3;
    const char *p1   = this->contents(),
               *p2,
               *p3;
    // See if we need to pad the receiver:
    if ( index > this->length() )
      {
      // Second initializer is pad characters:
      len2 = index - this->length(),
      p2   = 0,
      // Third initializer is the string to be inserted:
      len3 = len,
      p3   = pInsert;
      }
    else
      {
      // Second initializer is the string to be inserted:
      len2 = len,
      p2   = pInsert,
      // Third initializer is the rest of the receiver:
      len3 = this->length() - index,
      p3   = this->contents() + index;
      }
    // Return new buffer:
    result = this->newBuffer( p1, len1,
                              p2, len2,
                              p3, len3,
                              padCharacter );
    }
  else
    this->addRef();

  return result;
  }

/*------------------------------------------------------------------------------
| IBuffer::leftJustify                                                         |
|                                                                              |
| If the new length is the same as the receiver's current length, then the     |
| receiver is returned unmodified.                                             |
|                                                                              |
| Otherwise, the new buffer is generated via an invocation of the              |
| newBuffer() function.  The new contents are built from the receiver (either  |
| its full length or truncated at the new length if that is shorter) and a     |
| string of pad characters (of sufficient length to fill the result out to     |
| the specified length).                                                       |
------------------------------------------------------------------------------*/
IBuffer * IBuffer :: leftJustify ( unsigned newLength,
                                   char     padCharacter )
  {
  IBuffer *result = this;
  if ( newLength != this->length() )
    result = this->newBuffer( this->contents(),
                              ( this->length() < newLength ) ?
                                this->length()
                                :
                                newLength,
                              0,
                              ( newLength > this->length() ) ?
                                newLength - this->length()
                                :
                                0,
                              0,
                              0,
                              padCharacter );
  else
    this->addRef();

  return result;
  }

/*------------------------------------------------------------------------------
| IBuffer :: lowerCase                                                         |
|                                                                              |
| Starting at the beginning of the buffer, scan for the next capital           |
| letter and if found, convert it to lower case.  Quit as soon as              |
| the search for a capital (via indexOfAnyOf()) fails.                         |
------------------------------------------------------------------------------*/
  static const char caps[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";

IBuffer * IBuffer :: lowerCase ( )
  {
  IBuffer *result = this;

  unsigned pos = this->indexOfAnyOf( caps, 26, 1 );
  if ( pos )
    {
    if ( this->useCount() > 1 )
      result = this->newBuffer( this->contents(), this->length() );
    else
      this->addRef();
    char *p = result->contents();
    while ( pos )
      {
      p[ pos - 1 ] += 'a' - 'A';
      pos = this->indexOfAnyOf( caps, 26, pos + 1 );
      }
    }
  else
    this->addRef();

  return result;
  }

/*------------------------------------------------------------------------------
| IBuffer::overlayWith                                                         |
|                                                                              |
| The member function overlayWith() returns the receiver                       |
| unmodified if both the length of the overlay string is zero and              |
| the overlay occurs within the current receiver (not off the end).            |
|                                                                              |
| Otherwise, space is allocated for the combined buffers.  The                 |
| overlay is accomplished in four steps:                                       |
|                                                                              |
|   1. The portion of the receiver prior to the index at which                 |
|      the new string is to be overlayed is copied (may be null).              |
|   2. If the receiver isn't long enough, then it is padded out                |
|      to this index using the pad character.                                  |
|   3. The overlay string is copied (may be null).                             |
|   4. The portion of the receiver following the portion of it                 |
|      which was overlaid is copied (may be null).                             |
------------------------------------------------------------------------------*/
IBuffer * IBuffer :: overlayWith ( const char *pOverlay,
                                   unsigned    len,
                                   unsigned    index,
                                   char        padCharacter )
  {
  IBuffer *result = this;

  if ( len != 0
       ||
       index > this->length() )
    {
    // Treat index 0 like index 1:
    if ( index == 0 ) index = 1;
    unsigned  len1 = ( this->length() < index - 1 ) ?
                       this->length()
                       :
                       index - 1,
              len2,
              len3;
    const char *p1 = this->contents(),
               *p2,
               *p3;
    // Rest is pad+overlay or overlay+rest:
    if ( len1 < index - 1 )
      {
      // Second initializer is pad characters:
      len2 = ( index - 1 ) - len1;
      p2   = 0;
      // Third initializer is the string to be overlayed:
      len3 = len;
      p3   = pOverlay;
      }
    else
      {
      // Second initializer is the string to be overlayed:
      len2 = len;
      p2   = pOverlay;
      // Third initializer is the rest of the receiver:
      if ( this->length() >= index + len )
        {
        len3 = this->length() - index - len + 1;
        p3   = this->contents() + index + len - 1;
        }
      else
        {
        len3 = 0;
        p3   = 0;
        }
      }
    // Return new buffer:
    result = this->newBuffer( p1, len1,
                             p2, len2,
                             p3, len3,
                             padCharacter );
    }
  else
    this->addRef();

  return result;
  }

/*------------------------------------------------------------------------------
| IBuffer::reverse                                                             |
|                                                                              |
| If the string is not null, reverse its bytes.  Since the IBuffer is          |
| not guaranteed to not have null characters, we can't use strrev()            |
| and instead roll our own.                                                    |
------------------------------------------------------------------------------*/
IBuffer * IBuffer :: reverse ( )
  {
  IBuffer *result = this;
  if ( this->length() > 1 )
    {
    if ( this->useCount() > 1 )
      result = this->newBuffer( this->contents(), this->length() );
    else
      this->addRef();
    char* p1 = result->contents();
    char* p2 = result->contents() + result->length() - 1;
    while ( p1 < p2 )
      {
      char c = *p1;
      *p1++ = *p2;
      *p2-- = c;
      }
    }
  else
    this->addRef();

  return result;
  }

/*------------------------------------------------------------------------------
| IBuffer::strip                                                               |
|                                                                              |
| Accepts a string of characters to be stripped. A "mode" indicator specifies  |
| whether leading, trailing, or, both leading and trailing characters meeting  |
| the criteria are to be stripped.                                             |
|                                                                              |
| This method is implemented with the basic algorithm:                         |
|    1. Find the first character that fails the criteria                       |
|    2. Find the last character that fails the criteria                        |
|    3. Allocate space for and copy the intervening characters                 |
|    4. Return the resulting buffer.                                           |
------------------------------------------------------------------------------*/
IBuffer * IBuffer :: strip ( const char             *pChars,
                             unsigned                len,
                             IStringEnum::StripMode  mode )
  {
  IBuffer *result = this;

  if ( this->length() != 0 )
    {
    unsigned start = 1,
             stop  = this->length();
    if ( mode != IStringEnum::trailing )
      start = this->indexOfAnyBut( pChars, len, 1 );
    if ( mode != IStringEnum::leading
         &&
         start != 0 )
      stop = this->lastIndexOfAnyBut( pChars, len, this->length() );
    if ( start != 1
         ||
         stop != this->length() )
      {
      if ( start != 0 )
        result = this->newBuffer( this->contents() + start - 1,
                                 stop - start + 1 );
      else
        {
        result = this->null();
        result->addRef();
        }
      }
    else
      this->addRef();
    }
  else
    this->addRef();

  return result;
  }

/*------------------------------------------------------------------------------
| IBuffer::strip                                                               |
|                                                                              |
| This method uses an IStringTest object to determine which characters are to  |
| be stripped. A "mode" indicator specifies whether leading, trailing, or,     |
| both leading and trailing characters meeting the criteria are to be          |
| stripped.                                                                    |
|                                                                              |
| 1. Find the first character that fails the criteria                          |
| 2. Find the last character that fails the criteria                           |
| 3. Allocate space for and copy the intervening characters                    |
| 4. Return the resulting buffer.                                              |
------------------------------------------------------------------------------*/
IBuffer * IBuffer :: strip ( const IStringTest      &aTest,
                             IStringEnum::StripMode  mode )
  {
  IBuffer *result = this;

  if ( this->length() != 0 )
    {
    unsigned start = 1,
             stop  = this->length();
    if ( mode != IStringEnum::trailing )
      start = this->indexOfAnyBut( aTest, 1 );
    if ( mode != IStringEnum::leading
         &&
         start != 0 )
      stop = this->lastIndexOfAnyBut( aTest, this->length() );
    if ( start != 1
         ||
         stop != this->length() )
      {
      if ( start != 0 )
        result = this->newBuffer( this->contents() + start - 1,
                                 stop - start + 1 );
      else
        {
        result = this->null();
        result->addRef();
        }
      }
    else
      this->addRef();
    }
  else
    this->addRef();

  return result;
  }

/*------------------------------------------------------------------------------
| IBuffer::translate                                                           |
|                                                                              |
| The translation is accomplished by searching for the next occurrence of one  |
| of the input characters (via indexOfAnyOf()) and then converting it to the   |
| corresponding output character. The index into the character arrays is       |
| calculated via memchr(). If the output character array isn't this long,      |
| then it is effectively padded with the pad character.  The process stops     |
| when the call to indexOfAnyOf() fails.                                       |
------------------------------------------------------------------------------*/
IBuffer * IBuffer :: translate ( const char *pInputChars,
                                 unsigned    inputLen,
                                 const char *pOutputChars,
                                 unsigned    outputLen,
                                 char        padCharacter )
  {
  IBuffer *result = this;
  unsigned pos = this->indexOfAnyOf( pInputChars, inputLen, 1 );
  if ( pos )
    {
    // Changes will occur, make sure string contents are private:
    if ( this->useCount() > 1 )
      result = this->newBuffer( this->contents(), this->length() );
    else
      this->addRef();
    char *p = result->contents();
    // Create copies of source/target characters to avoid
    // problems if either overlap the receiver:
    char
     *pIn  = (char*) memcpy( new char[ inputLen ], pInputChars, inputLen ),
     *pOut = (char*) memcpy( new char[ inputLen ],
                             pOutputChars,
                             ( outputLen < inputLen ) ? outputLen : inputLen );
    if ( outputLen < inputLen )
      memset( pOut+outputLen, padCharacter, inputLen - outputLen );
    // Translate all occurrences of input chars:
    while ( pos )
       {
       unsigned i = (char*) memchr( pIn, p[ pos - 1 ], inputLen ) - pIn;
       p[ pos - 1 ] = pOut[ i ];
       pos = this->indexOfAnyOf( pIn, inputLen, pos + 1 );
       }
    delete pIn;
    delete pOut;
    }
  else
    this->addRef();

  return result;
  }

/*------------------------------------------------------------------------------
| IBuffer::upperCase                                                           |
|                                                                              |
| Starting at the beginning of the string, scan for the next lower case letter |
| and if found, convert it to upper case.  Quit as soon as the search for a    |
| lowercase (via indexOfAnyOf()) fails.                                        |
------------------------------------------------------------------------------*/
  static const char lowers[] = "abcdefghijklmnopqrstuvwxyz";

IBuffer * IBuffer :: upperCase ( )
  {
  IBuffer *result = this;

  unsigned pos = this->indexOfAnyOf( lowers, 26, 1 );

  if ( pos )
    {
    if ( this->useCount() > 1 )
      result = this->newBuffer( this->contents(), this->length() );
    else
      this->addRef();
    char *p = result->contents();
    while ( pos )
      {
      p[ pos - 1 ] += 'A' - 'a';
      pos = this->indexOfAnyOf( lowers, 26, pos + 1 );
      }
    }
  else
    this->addRef();

  return result;
  }
