/*******************************************************************************
* FILE NAME: iecmpimp.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Implementation of the classes:                                             *
*     IEmbeddedComponentImp                                                    *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
#include <precomp.hpp>

#include "iecmpimp.hpp"

#include <iembmod.hpp>
#include <istatnry.hpp>
#include <iguibndl.hpp>
#include <icmpfram.hpp>
#include <ibasstrm.hpp>
#include <iview.hpp>

/*******************************************************************************
*                                                                              *
*                        IEmbeddedComponentImp                                 *
*                                                                              *
*******************************************************************************/

// -----------------------------------------------------------------------------
// IEmbeddedComponentImp
// -----------------------------------------------------------------------------
// default constructor
IEmbeddedComponentImp::IEmbeddedComponentImp( IEmbeddedComponent* comp,
                                              IComponent::EAspect aspect ) :
    CAswContainedObject( IComponentStationery::stationery() ),
    fpEmbeddedComp( comp ),
    fpAspect( aspect == IComponent::kIcon ? &fIconAspect : &fContentAspect ),
    fContentAspect(),
    fIconAspect(),
    fRendDev( IComponentStationery::stationery() ),
    fPresSite( *this ),
    fInitialized( FALSE )
{
}


// -----------------------------------------------------------------------------
// IEmbeddedComponentImp
// -----------------------------------------------------------------------------
// copy constructor
// IEmbeddedComponentImp::IEmbeddedComponentImp( const IEmbeddedComponentImp& theCopy ) :
//  CAswContainedObject( IComponentStationery::stationery() )
// {
// }

// -----------------------------------------------------------------------------
// ~IEmbeddedComponentImp
// -----------------------------------------------------------------------------
// destructor
IEmbeddedComponentImp::~IEmbeddedComponentImp()
{
    uninitialize();

}
        
void
IEmbeddedComponentImp::uninitialize()
{
    if ( fInitialized )
    {
        fPresSite.Shutdown();
        fRendDev.Shutdown();
        fContentAspect.Shutdown();
        fIconAspect.Shutdown();

        // shutdown ourself
        Shutdown();

        fInitialized = FALSE;
    }
    
}

// -----------------------------------------------------------------------------
// initialize
// -----------------------------------------------------------------------------
// this method creates an Aswan aspect (contents) and aswan rendering device (screen)
// and a pres site (out of place for now)
void
IEmbeddedComponentImp::initialize( Boolean isNew, Boolean bActivate )
{

    // create the requried aspect object
    CAswResult ar = fpAspect->Setup( this, fpEmbeddedComp->aspect() );
    IASWSHOWERROR( ar );

    // create a rendering device for the screen
    ar = fRendDev.Setup( this, fpAspect );
    IASWSHOWERROR( ar );

    IComponentFrameWindow* pFrame = (IComponentFrameWindow*)
        &fpEmbeddedComp->bundle().frameWindow();

    // create a presentation site in the container's client area
    ar = fPresSite.Setup( this, &fRendDev, *pFrame, NULL,
                          (HWND)(void*)IComponentStationery::stationery().bundle().objectView().handle() );
    IASWSHOWERROR( ar );

    if( isNew )
    {
        // Get the initial size from the server and assume this is the size
        SIZEL sizeHi, sizePix;
        fpAspect->GetExtent( &sizeHi );
        // size is in himetric units - convert to pixels
        XformSizeInHimetricToPixels( NULL, &sizeHi, &sizePix );

        // for in-place activation do not allow a window that is so small
        // that the users can't do anyting in it!
        if ( fpAspect->GetAspect() == DVASPECT_CONTENT )
        {
            // Use 200 x 200 for content aspect to compensate for Excel bug
            // Warning: ANY app that asks for a 65x17 will get 200x200
            if( sizePix.cx == 65 && sizePix.cy == 17 )
            {
                sizePix.cx = 200;
                sizePix.cy = 200;
            }
        }

        // offset the object by 50 pixels in each direction
        // (to allow for adornments)
        // Note: if you change this constant make sure you change the corresponding
        // value in idataex.cpp (IAswPasteTarget::ObjectPasted)
        fpEmbeddedComp->setArea( IRectangle( IPoint(50, 50),
                                             ISize(sizePix.cx, sizePix.cy) ) );

        // start caching the object
        ar = fRendDev.Cache();
        IASWSHOWERROR( ar );

    }
    
    // start change notifications
    ar = fpAspect->SetAdvise();
    IASWSHOWERROR( ar );

    if( bActivate )
    {
        // call the server to activate
        ar = Activate( NULL, &fPresSite );
        IASWSHOWERROR( ar );
    }
    else
        if( !isNew )
        {
            ar = Load();
            IASWSHOWERROR( ar );
        }

    fInitialized = TRUE;

}

//--------------------------------------------------
// area
//--------------------------------------------------
// returns the area rectangle in the embedded object
IRectangle
IEmbeddedComponentImp::area()
{
    return fpEmbeddedComp->area();
}

//--------------------------------------------------
// setArea
//--------------------------------------------------
// sets the area rectangle in the embedded object
void
IEmbeddedComponentImp::setArea( IRectangle rect )
{
    fpEmbeddedComp->setArea( rect );

    IGrabHandles* hdls = fpEmbeddedComp->grabHandles();
    if ( hdls )
        hdls->changeTo( rect );

}

//--------------------------------------------------
// switchAspect
//--------------------------------------------------
void                   
IEmbeddedComponentImp::switchAspect( IAswAspect& newAspect )
{
    if ( isInitialized() )
        fpAspect->SwitchAspect( &newAspect, false, false );

    fpAspect = &newAspect;
}

//--------------------------------------------------
// ObjectIsEmpty
//--------------------------------------------------
void                   
IEmbeddedComponentImp::ObjectIsEmpty( void )
{
    // we failed to save ourself, so we better delete ourself
    
    ::PostMessage( (HWND)(void*)fpEmbeddedComp->bundle().frameWindow().handle(),
                   WM_COMMAND, kEditDelete, 0 );
    
}


/*******************************************************************************
*                                                                              *
*                        IAswAspect                                            *
*                                                                              *
*******************************************************************************/

//--------------------------------------------------
// Construction/Destruction
//--------------------------------------------------
IAswAspect::IAswAspect() :
    CAswAspect( IComponentStationery::stationery() )
{
}

IAswAspect::~IAswAspect()
{
    Shutdown( TRUE );
}

//--------------------------------------------------
// OnViewChange
//--------------------------------------------------
// Notification that an aspect's presentation view has changed. 
BOOL
IAswAspect::OnViewChange( void )
{

    // Aswan thinks all sites are in place if they are derived from InPlacePresSite
    // so we have to check ourselves. This behaviour will be changed in Rel.2 of Aswan
    IAswPresSite& presSite = ((IEmbeddedComponentImp*)GetContainedObject())->presSite();
    if( !presSite.isInPlace() )
        presSite.RerenderNeeded( TRUE );

    return TRUE;
}

//--------------------------------------------------
// OnExtentChange
//--------------------------------------------------
// Notification that the object changed its extents for presentations of this aspect.
void
IAswAspect::OnExtentChange( SIZEL sizelNewExtentsHiMetric )
{
#ifdef IC_SCREEN_DEBUG
cout << "OnExtentChange called" << endl;
#endif

    // size is in himetric units - convert to pixels
    SIZEL sizePix;    
    XformSizeInHimetricToPixels( NULL, &sizelNewExtentsHiMetric, &sizePix );

    // set the size in the parent rectangle
    IEmbeddedComponentImp* pParent = (IEmbeddedComponentImp*)GetContainedObject();
    IRectangle rectNew = pParent->area().sizeTo( ISize(sizePix.cx, sizePix.cy) );
#ifdef IC_SCREEN_DEBUG
cout << " size: " << sizePix.cx << "," << sizePix.cy << endl;
#endif

    pParent->setArea( rectNew );

}


/*******************************************************************************
*                                                                              *
*                        IAswPresSite                                          *
*                                                                              *
*******************************************************************************/

// -----------------------------------------------------------------------------
// IAswPresSite
// -----------------------------------------------------------------------------
// default constructor
IAswPresSite::IAswPresSite( IEmbeddedComponentImp& Parent ) :
    CAswInPlacePresSite( IComponentStationery::stationery() ),
    fbInPlace( false ),
    fParent( Parent )
{
}


// -----------------------------------------------------------------------------
// ~IAswPresSite
// -----------------------------------------------------------------------------
// destructor
IAswPresSite::~IAswPresSite()
{
    Shutdown( TRUE );
}


// -----------------------------------------------------------------------------
// GetPosition
// -----------------------------------------------------------------------------
// called by the framework to get the current area (in pixels) for the object
// as well as a handle to the parents window
CAswResult
IAswPresSite::GetPosition( HWND ASWPTR* phWnd, LPRECT lprcPixelPos )
{

    lprcPixelPos->left   = fParent.area().minX();
    lprcPixelPos->top    = fParent.area().minY();
    lprcPixelPos->right  = fParent.area().maxX();
    lprcPixelPos->bottom = fParent.area().maxY();

//FBEMB *phWnd = (HWND)(void*)IComponentStationery::stationery().bundle().view().handle();
    *phWnd = (HWND)(void*)IComponentStationery::stationery().bundle().objectView().handle();
    
}


// -----------------------------------------------------------------------------
// InPlaceObjectRequestsNewPosition
// -----------------------------------------------------------------------------
// Notification that the inplace presentation changed its extents, and is
// thus asking for more or less position to edit in place. 
void ASWPUBMETHOD
IAswPresSite::InPlaceObjectRequestsNewPosition( LPCRECT lprcPosClientPixelsIn,
                                                LPRECT  lprcPosClientPixelsOut )
{

    // set our area to match the request
    IPoint ptTopLeft( lprcPosClientPixelsIn->left, lprcPosClientPixelsIn->top );
    IPoint ptBottomRight( lprcPosClientPixelsIn->right, lprcPosClientPixelsIn->bottom );
    fParent.setArea( IRectangle( ptTopLeft, ptBottomRight ) );

    // return the requested rect back to the server to indicate we've accepted the request
    *lprcPosClientPixelsOut = *lprcPosClientPixelsIn;

}

