/*******************************************************************************
* FILE NAME: iframhdr.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in iframhdr.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C"
  {
  #define INCL_WINFRAMEMGR
  #define INCL_WINTRACKRECT
  #include <iwindefs.h>
  }

#include <iframhdr.hpp>

#include <iframprv.hpp>
#include <iframevt.hpp>
#include <iswp.hpp>
#include <itrace.hpp>
#include <icolor.hpp>
#include <icontrol.hpp>
#include <iexcept.hpp>
#include <ihandle.hpp>
#include <icmdevt.hpp>
#include <iframe.hpp>
#include <inotifev.hpp>
#include <icoordsy.hpp>
#include <isysmenu.hpp>
#include <iplatfrm.hpp>
#include <ititle.hpp>
#ifdef IC_WIN
  #include <icvhdr.hpp>
  #include <ithread.hpp>
  #include <iwcname.hpp>
  #include <iwinlsts.hpp>
  #include <ihelp.hpp>
#endif

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IFRAMHDR_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IFrameHandler::dispatchHandlerEvent                                          |
|                                                                              |
| Intercepts all window events and dispatches frame-specific events to the     |
| various handle event functions.                                              |
| Note: The following messages are of interest:                                |
|    For WM_CALCFRAMERECT:                                                     |
|      Dispatch the calcRect function.                                         |
|                                                                              |
|    For WM_QUERYFRAMECTLCOUNT:                                                |
|      1. Get the number of standard controls by invoking the default frame    |
|         window procedure.                                                    |
|      2. Get the number of frame extensions via a call to numberOfExtensions()|
|      3. Return the sum in the event result.                                  |
|                                                                              |
|    For WM_FORMATFRAME:                                                       |
|      Dispatch the format() function.                                         |
|                                                                              |
|    For WM_CLOSE:                                                             |
|      Dispatch the closed() function.                                         |
|                                                                              |
|    For WM_ACTIVATE:                                                          |
|      Dispatch either activated() or deactivated() based on mp1.              |
|                                                                              |
|    For WM_SAVEAPPLICATION:                                                   |
|      Dispatch saved().                                                       |
|                                                                              |
|    For WM_PAINT:                                                             |
|      Dispatch draw().                                                        |
|                                                                              |
|    For WM_COMMAND:                                                           |
|      Dispatch command();                                                     |
|                                                                              |
|    For WM_DESTROY:                                                           |
|      Call frame->removeFromWindowList();                                     |
|                                                                              |
|    For WM_DRAWITEM: or WM_MEASUREITEM:                                       |
|      Return true if already sent to client.                                  |
|                                                                              |
|    For WM_CONTROLPOINTER:                                                    |
|      Return the handle of the mouse pointer to use in the event result.      |
|                                                                              |
|    This handler returns false for all other events.                          |
------------------------------------------------------------------------------*/
Boolean IFrameHandler::dispatchHandlerEvent ( IEvent &event )
  {
  IMODTRACE_ALL("IFrameHandler::dispatchHandlerEvent");
  Boolean
    result = false;

  switch( event.eventId() )
    {
#ifdef   IC_WIN
    case IC_UM_CALCFRAMERECT:
      // This message is sent by the ICalcFrameRect function to compute
      // size of the frame rectangle.  Parameter1 is a Windows RECT
      // structure, Parameter2 is 0 if the RECT is a client rectangle,
      // otherwise the RECT is a window rectangle.  Message similar to
      // PM WM_CALCFRAMERECT
#endif   //IC_WIN
#ifndef IC_WIN
    case WM_CALCFRAMERECT:
#endif
      {
      IMODTRACE_DEVELOP( "IFrameHandler WM_CALCFRAMERECT" );

      IFrameEvent
        frameEvent( event );

      // Dispatch virtual function...
      result = this -> calcRect( frameEvent );

      // If handled, copy result to input event structure...
      if ( result )
        {
        event.setResult( frameEvent.result() );
        }
      }
      break;

#ifdef IC_WIN
    case IC_UM_UPDATEFRAME:
      {
      IMODTRACE_DEVELOP( "IFrameHandler IC_UM_UPDATEFRAME" );

      // In Windows, the frame extension formatting is driven by
      // this IC_UM_UPDATEFRAME message.  It is generated by the
      // IFrameWindow::update function.

      IFrameWindow* frame = ((IFrameWindow*)(event.window()));

      // Handle initializing the dialog template unwrappered children list.
      // For IWindow objects, we insure via default handlers that keystroke
      // messages make it to the canvas or frame window to provide
      // keyboard navigation.  With dialogs created via a dialog template,
      // however, we have unwrappered controls in the dialog which by
      // default simply eat keyboard events,  perhaps after doing something
      // with them.  We create an IDialogControls object and keep in it a list
      // of the children of the dialog and their window procedures.  We subclass
      // each control so that we can intercept keyboard events as needed.
      // The subclass window procedure is pfnICControlProc.
      if ( frame->fFrameWindowData->flags & IFrameWindowData::fromDlgTemplate )
         {
         // Get a pointer to the controls list for this thread
         IDialogControls* dialogControls =
            IThread::current().dialogControls();
         // Enumerate the children of the dialog.
         IWindow::ChildCursor child(*frame);
         for (child.setToFirst(); child.isValid(); child.setToNext() )
            {
            IWindowHandle childhwnd = frame->childAt( child );

            // Special case for combobox.  Need to find the entryfield
            // child of the combobox.
            if ( (IWindowClassName(childhwnd) == WC_COMBOBOX) &&
                 (IWindow::windowWithHandle(childhwnd, false) == 0) )
               {
               // Find the edit control child of the combobox.
               childhwnd = GetWindow(childhwnd, GW_CHILD);
               while ( (childhwnd) &&
                       (IWindowClassName(childhwnd) != WC_ENTRYFIELD))
                  childhwnd = GetWindow(childhwnd, GW_HWNDNEXT);
               }

            unsigned long dlgCode =
               childhwnd.sendEvent(WM_GETDLGCODE, 0, 0);
            //
            // It is possible that we could call this more than once.
            // In that case the control is already in the collection
            // and we don't do anything.   We also skip adding to
            // the collection anything that is an IWindow.  It will
            // still work if an IWindow wrapper is created later, but not
            // if the control is an IWindow now.

            if ( (dlgCode & (DLGC_WANTARROWS | DLGC_WANTCHARS | DLGC_BUTTON))
                 &&
                 !(dialogControls->containsElementWithKey(
                                      childhwnd.asUnsigned() ) ||
                   IWindow::windowWithHandle(childhwnd, false) ) )
               {
               IDialogControl control;
               control.fHwnd = childhwnd.asUnsigned();
               control.fOriginalProc = (void*)
                  ISUBCLASSWINDOW(childhwnd,
                                  (unsigned long)pfnwpICControlProc);
               control.fDialog = frame->handle();
               // check here for infinite loop or possible traps
               IASSERTSTATE(control.fOriginalProc &&
                            control.fOriginalProc != pfnwpICControlProc);
               dialogControls->add( control );
               ITRACE_DEVELOP( IString("adding control hwnd=") +
                               IString(control.fHwnd).d2x() );
               }

            }
         }  // end of dialog template case

      // get the client rectangle
      RECT clientRect;
      GetClientRect( event.handle(), &clientRect );

      // prepare and send an IC_UM_FORMATFRAME message to the frame.
      // all of the frame extensions need to be located
      unsigned long
        frameControls = event.handle().sendEvent( IC_UM_QUERYFRAMECTLCOUNT );
      ISWP* swp = new ISWP[ frameControls ];

      // Create ChildCursor to get handles of all child windows,
      // adding them to the ISWP array.  For "standard" frame
      // components, query their position and size and also stuff
      // that information in.  Note that position/size are in native
      // system coordinates, not ICLUI interface coordinates.
      // We only support client attachment at this time.   The client
      // is the only control whose position is filled in in
      // the SWP array.  Extensions are positioned by the format()
      // function later.
      IWindow::ChildCursor cursor(*frame);
      unsigned long index = 0;
      for ( cursor.setToFirst();
            cursor.isValid() ;
            cursor.setToNext())
         {
         IWindowHandle child = frame->childAt( cursor );
         if ( (IIDOF( child ) != FID_CLIENT ) &&
              (index < frameControls) )
            {
            swp[index].hwnd = child;
            index++;
            }
         }

      // Construct & send frame format event.  parameter1 is a pointer to
      // the SWP array, parameter2 is the current client rectangle.
      // Event handled in this handler below.  The format() function is
      // called to position the extensions.
      event.handle().sendEvent( IC_UM_FORMATFRAME,
                                IEventData( swp ),
                                IEventData( &clientRect)  );
      // set the positions of the frame controls
      HDWP  hdwp = BeginDeferWindowPos( (int)frameControls );
      if (!hdwp)
         {
         ITHROWGUIERROR2("BeginDeferWindowPos",
                         IBaseErrorInfo::outOfSystemResource,
                         IException::unrecoverable);
         }
      // set the position of the client window, if any
      IWindowHandle clientHwnd = frame->clientHandle();
      if (clientHwnd != 0)
         {
         // positioning and sizing done in native coordinates
         ISize clientSize(clientRect.right - clientRect.left,
                          clientRect.bottom - clientRect.top );
         ITRACE_ALL( IString("DeferWindowPos clientHwnd=") +
                     IString(clientHwnd.asUnsigned()).d2x() +
                     IString(" position=") +
                     IString(clientRect.left) + IString(",") +
                     IString(clientRect.top) +
                     IString(" size=") +
                     clientSize.asString() );
         hdwp = DeferWindowPos(hdwp,
                               clientHwnd, 0,
                               (int)clientRect.left,
                               (int)clientRect.top,
                               (int)clientSize.width(),
                               (int)clientSize.height(),
                               SWP_NOZORDER );
         }  // if client
      // set position of extensions
      for (index=0;  index < frameControls;  index++)
         {
         ISWP* extSWP = &swp[index];
         if ( extSWP->isMove() || extSWP->isSize() )
            {
            // place child windows.  Note that we work in native coordinates
            // to position extensions so no recoordination is done.
            ITRACE_ALL( IString("DeferWindowPos hwnd=") +
                        IString(extSWP->windowHandle().asUnsigned()).d2x() +
                        IString(" position=") + extSWP->position().asString() +
                        IString(" size=") +
                        extSWP->size().asString() );
            hdwp = DeferWindowPos(hdwp,
                                  extSWP->windowHandle(), 0,
                                  (int)extSWP->position().x(),
                                  (int)extSWP->position().y(),
                                  (int)extSWP->size().width(),
                                  (int)extSWP->size().height(),
                                  extSWP->flags() );
            }  // if move or size
         }  // for

      if ( !EndDeferWindowPos(hdwp) )
         {
         ITHROWGUIERROR2("EndDeferWindowPos",
                         IBaseErrorInfo::invalidParameter,
                         IException::unrecoverable);
         }
      // delete the SWP array
      delete [] swp;

      // if there is a client size it to the frame
//      IWindow*  clientWin = frame->client();
//      if (clientWin)
//         clientWin->moveSizeTo( frame->clientRectFor( IRectangle() ) );
      result = true;
      }
      break;
#endif

#ifdef IC_WIN
    case IC_UM_QUERYFRAMECTLCOUNT:
      {
      IMODTRACE_DEVELOP( "IFrameHandler IC_UM_QUERYFRAMECTLCOUNT" );

      unsigned long
        stdControls = 0;           // No additional standard extensions
#endif // IC_WIN
#ifndef IC_WIN
    case WM_QUERYFRAMECTLCOUNT:
      {
      IMODTRACE_DEVELOP( "IFrameHandler WM_QUERYFRAMECTLCOUNT" );

      // Call default procedure to get number of standard controls...
      unsigned long
        stdControls = this -> defaultProcedure( event );
#endif  // !IC_WIN

      // Compute additional controls count...
      unsigned long
        extensions = this -> numberOfExtensions(
                                         (IFrameWindow*)event.window() );

      // Result is the sum...
      event.setResult( IEventResult( stdControls + extensions ) );

      result = true;
      }
      break;

#ifdef IC_WIN
    // Size handler code for Windows to perform client area resizing
    // whenever frame is resized
    case WM_SIZE:
      {
      IMODTRACE_DEVELOP( "IFrameHandler WM_SIZE" );

      IFrameWindow* frame = ((IFrameWindow*)(event.window()));

      frame->fFrameWindowData->flags |= IFrameWindowData::needsUpdating;
      frame->update();
      }
      break;
#endif  // IC_WIN

#ifdef IC_WIN
    case IC_UM_FORMATFRAME:
#endif
#ifndef IC_WIN
    case WM_FORMATFRAME:
#endif
      {
      IMODTRACE_DEVELOP( "IFrameHandler WM_FORMATFRAME" );

      IFrameFormatEvent
        fmtEvent( event );

      // Dispatch virtual function...
      result = this -> format( fmtEvent );

      // If handled, copy result to input event structure...
      if ( result )
        event.setResult( fmtEvent.result() );
      }
      break;

    case WM_SYSCOMMAND:
    case WM_CLOSE:
    {
      Boolean processClose = false;    // Don't treat most syscmds as close.

      if ( event.eventId() == WM_CLOSE )
      {       // Always close window on WM_CLOSE.
         processClose = true;
      }
#ifdef IC_WIN
      else if ( (((unsigned long)event.parameter1()) & 0xFFF0) == SC_CLOSE )
#else   // !IC_WIN
      else if ( event.parameter1().number1() == SC_CLOSE )
#endif  // !IC_WIN
      {       // Close system command (e.g. from system menu or accelerator).
         processClose = true;   // Almost always close the window.

         // Check if the "Close" system menu choice is disabled.
         // If so, ignore the close request.
         IFrameWindow
          *pFrame = (IFrameWindow*)( event.window() );
#ifdef IC_WIN
         //
         if ( pFrame && ( ISTYLEOF(event.handle()) & WS_SYSMENU ) )
         {
            // Possible multiple ISystemMenu objects OK since menus
            // are not a window
            ISystemMenu sysMenu( pFrame );
            try
            {
               if ( !sysMenu.isItemEnabled(ISystemMenu::idClose) )
               {      // "Close" choice is disabled, ignore the request.
                  processClose = false;
                  result = true;
               }
            }
            catch (IInvalidRequest&)
            {
               // The close menu item has been removed...same as disabled.
               processClose = false;
               result = true;
            }
         }
#else
         IWindowHandle sysMenuHandle =
                         pFrame->handleFor( IFrameWindow::systemMenu );
         if ( sysMenuHandle )
         {         // Frame has a system menu.
            Boolean needToDelete = false;
            ISystemMenu* sysMenu =
              (ISystemMenu*)IWindow::windowWithHandle( sysMenuHandle );
            if ( !sysMenu )
            {
               sysMenu = new ISystemMenu( pFrame );
               needToDelete = true;
            }

            if ( !sysMenu->isItemEnabled( ISystemMenu::idClose ))
            {      // "Close" choice is disabled, ignore the request.
               processClose = false;
            }

            if ( needToDelete )
            {      // Memory clean-up.
               delete sysMenu;
            }
         }
#endif
      }

      if ( processClose )
      {
        IFrameEvent
          frameEvent( event );

        frameEvent.frame()->fFrameWindowData->flags |=
                                        IFrameWindowData::hasBeenClosed;

        result = this -> closed( frameEvent );

        if ( result )
          event.setResult( frameEvent.result() );
      }
#ifdef IC_WIN
      // If this is a system command message
      else if ( event.eventId() == WM_SYSCOMMAND )
      { // and not a close message
        if ( (((unsigned long)event.parameter1()) & 0x7FF0) != SC_CLOSE )
        {
          // Need to call default procedure here and strip off high order
          // bit potentially added earlier for routing purposes.
          IEvent newEvent( event.handle(),
                           event.eventId(),
                           event.parameter1() & 0x7FFFFFFF,
                           event.parameter2() );

          this->defaultProcedure( newEvent );

          // Event handled so don't route on.
          event.setResult( true );
          result = true;
        }
      }
#endif
      break;
    }

#ifdef IC_WIN
    // This is to help behavior of child MDI style windows, athough we
    // don't currently fully support MDI
    case WM_CHILDACTIVATE:
#endif
    case WM_ACTIVATE:
      {
      IFrameEvent
        frameEvent( event );

#ifdef IC_WIN
      if ( ( (unsigned long)frameEvent.parameter1() ) ||
           ( frameEvent.eventId() == WM_CHILDACTIVATE ) )
#else
      if ( (unsigned long)frameEvent.parameter1() )
#endif
        {
        ITRACE_ALL(IString("ACTIVATE frame=")+
                   IString( (unsigned long)(void*)frameEvent.frame()).d2x() );
#ifdef IC_WIN
        // Set the default push button handle if it is not already set.
        // IGWL_DEFBUTTON is an extra window word present in the WC_FRAME
        // window classes that we register.  Not in dialogs!
        IWindowClassName className( event.handle() );
        if ((className.asString().indexOf(WC_FRAME) == 1) )
           {
           // Not dialog template.  Get current value to see if one set.
           unsigned long defPB = IQUERYWINDOWULONG( event.handle(),
                                                    IGWL_DEFBUTTON );
           ITRACE_ALL( IString("Default button HWND=") + IString(defPB).d2x() );
           if (!defPB)
              {
              // value not set...attempt to set now
              defPB = ICanvasStatics::defaultPushButtonHandle(
                                         event.window() ).asUnsigned();
              if (defPB)
                 ISETWINDOWULONG( event.handle(), IGWL_DEFBUTTON, defPB );
              ITRACE_ALL( IString("New default button HWND=") +
                          IString(defPB).d2x() );
              }
           }  // not dialog template
#endif
        frameEvent.frame() -> update();
        result = this -> activated( frameEvent );
        }
      else
        result = this -> deactivated( frameEvent );

      if ( result )
        event.setResult( frameEvent.result() );
      }
      break;
#ifndef IC_WIN
    case WM_SAVEAPPLICATION:
      {
      IFrameEvent
        frameEvent( event );

      result = this -> saved( frameEvent );

      if ( result )
        event.setResult( frameEvent.result() );
      }
      break;
#endif

    case WM_PAINT:
      {
      IFrameEvent
        frameEvent( event );

      result = this -> draw( frameEvent );

      if ( result )
        event.setResult( frameEvent.result() );
      }
      break;

    case WM_COMMAND:
      {
      ICommandEvent cmdEvent( event );
#ifdef IC_PM
      result = this -> command( cmdEvent );
      if ( result )
        event.setResult( cmdEvent.result() );
#endif
#ifdef IC_WIN
      /*---------------------------------------------------------------------
      | Here we are emulating PM system behavior for certain WM_SYSCOMMAND  |
      | messages which tend to be used in applications.  Unfortunately,     |
      | there is no MIS_SYSCOMMAND resource symbol, so we have to process   |
      | the SC_HELP id's as reserved command values.                        |
      ---------------------------------------------------------------------*/
      IHelpWindow::HelpType helptype;
      Boolean      sendToHelp = false;
      Boolean      convertClose = false;
      // See if the event is a help display message.  If so, set
      // type indicator.  Otherwise, ignore it
      switch ( cmdEvent.commandId() )
      {
        case SC_HELPEXTENDED:
           helptype = IHelpWindow::general;
           sendToHelp = true;
           break;
        case SC_HELPKEYS:
           helptype = IHelpWindow::keys;
           sendToHelp = true;
           break;
        case SC_HELPINDEX:
           helptype = IHelpWindow::index;
           sendToHelp = true;
           break;
        case SC_CLOSE:
           convertClose = true;
           break;
      }
      if (sendToHelp)
      {
        // If there is an associated help instance, show the
        // requested help window and indicate message handled.
        IHelpWindow *hwin =
                       IHelpWindow::helpWindow(cmdEvent.dispatchingWindow());
        if (hwin)
        {
          hwin->show( helptype );
          result = true;
        }
      }
      // If SC_CLOSE received, convert to a system command and resend
      // and prevent this message from being passed on.
      else if (convertClose)
      {
        event.handle().sendEvent( WM_SYSCOMMAND,
                                  IEventParameter1( SC_CLOSE, CMDSRC_OTHER ),
                                  IEventParameter2( 0 ) );
        result = true;
      }
      else
      {
        // Normal command so call overload function and save result returned.
        result = this -> command( cmdEvent );
        if ( result )
          event.setResult( cmdEvent.result() );
      }
#endif
      }
      break;

    case WM_DESTROY:
      {
      IFrameWindow
       *pFrame = (IFrameWindow*)( event.window() );
      // If frame wasn't "closed", do so now.
      if ( !( pFrame->fFrameWindowData->flags & IFrameWindowData::hasBeenClosed ) )
        {
        pFrame -> setDestroyOnClose( false ); // Prevent recursion.
        IFrameEvent
          frameEvent( event );
        this -> closed( frameEvent );
        }
#ifdef IC_WIN
      if (!(IPlatform::isWin95() || IPlatform::isNTNewShell()))
        RemoveProp( pFrame->handle(), PROP_MENUBAR_OBJECT );
#endif  // !IC_WIN
      pFrame -> removeFromWindowList();

      //-----------------------------------------------------------------
      // Without special processing, the combination of
      // IFrameWindow::showModally and IWindow::setAutoDeleteObject
      // will cause a trap, during manipulation of the deleted frame
      // object at the end of showModally.  This code causes the
      // auto-delete processing in IWindow to be bypassed, so the object
      // deletion can be done by IFrameWindow instead.  By delaying the
      // setting of the private data flag until now, there should be no
      // out-of-synch conditions between the flag and the auto-delete
      // setting of the frame object (for example, caused by handlers
      // attached to the frame object).
      //-----------------------------------------------------------------
      if ( pFrame->fFrameWindowData->flags & IFrameWindowData::modal  &&
           pFrame->isAutoDeleteObject() )
        {
         pFrame->fFrameWindowData->flags |= IFrameWindowData::modalAutoDelete;
         pFrame->setAutoDeleteObject( false );
        }
      }
      break;

    case WM_DRAWITEM:
    case WM_MEASUREITEM:
      result = ( (unsigned long)event.parameter1() == FID_CLIENT );
      break;

#ifdef IC_WIN
    case WM_SETCURSOR:
#else   // !IC_WIN
    case WM_CONTROLPOINTER:
#endif  // !IC_WIN
      {
      IFrameWindow
       *pFrame = (IFrameWindow*)event.dispatchingWindow();

#ifdef IC_WIN
      if (pFrame->mousePointer())
      {
        ISETPOINTER( pFrame->mousePointer() );
        event.setResult( true );
        result = true;
      }
      else
      {
        result = false;
      }
#else   // !IC_WIN
      if (pFrame->mousePointer())
      {
        event.setResult((unsigned long)pFrame->mousePointer());
      }
      else
      {
        event.setResult((unsigned long)event.parameter2());
      }
      result = true;
#endif  // !IC_WIN
      }
      break;

    case WM_WINDOWPOSCHANGED:
      // If the frame is becoming active and this is a secondary window,
      // be sure that the owner window not minimized.  This is needed to
      // work around the PM behavior for primary and secondary windows
      // being restored from the task list.
      if ((event.window() != 0) &&
#ifdef IC_WIN
         (!(((PWINDOWPOS)(char*)event.parameter2())->flags & SWP_NOACTIVATE)))
#else
         (((unsigned long) event.parameter2()) & AWP_ACTIVATE))
#endif
      {
         IWindow* pOwner = event.window()->owner();
         if (pOwner && pOwner->isFrameWindow() &&
            ((IFrameWindow*)pOwner)->isMinimized())
         {
            ((IFrameWindow*)pOwner)->restore();
         }
      }
      break;

#ifdef IC_WIN
    case WM_QUERYFOCUSCHAIN:
       {
       ITRACE_ALL( IString( "WM_QUERYFOCUSCHAIN " ) );
       switch (event.parameter1().lowNumber())
          {
          case QFC_NEXTINCHAIN  :
          case QFC_FRAME        :
              event.setResult( IEventResult(event.handle().asUnsigned()) );
              result = true;
              break;
          case QFC_PARTOFCHAIN  :
              result = (event.handle().asUnsigned() ==
                        event.parameter2().asUnsignedLong() );
              event.setResult( result );
              break;
//          These are not supported and not used by the Class Library
//            case QFC_ACTIVE       :
//            case QFC_SELECTACTIVE :
          default:
             break;
          }
       }
       break;

    case WM_SETREDRAW:
       {
       ITRACE_DEVELOP(IString("IFrameHandler WM_SETREDRAW flag=") +
                      IString(event.parameter1().asUnsignedLong())  );
       if ( event.parameter1().asUnsignedLong() )
          {
          Boolean wasVisible = event.window()->isVisible();
          if (!wasVisible)
            {
            ITRACE_DEVELOP("was not visible");
            defaultProcedure(event);

            // RDW_ALLCHILDREN may be too much.  Seem to need
            // it to get everything to paint.
            unsigned int rdFlags = RDW_ERASE | RDW_FRAME | RDW_ALLCHILDREN |
                                   RDW_INVALIDATE;
            RedrawWindow( event.handle(),         // this window
                          0,                      // entire window (rect)
                          0,                      // entire window (region)
                          rdFlags );              // flags

            result = true;
            }
          else
            {
            ITRACE_DEVELOP("was visible");
            }
          }
       }
       break;


    // If the frame window has a client, the frame window should not try to
    // paint its background using its background brush.
    case WM_ERASEBKGND:
      if (((IFrameWindow*)event.window())->clientHandle())
      {
        event.setResult( true );
        result = true;
      }
      break;

    case WM_NCPAINT:
    case WM_NCACTIVATE:
       {
       IFrameWindow* frame = (IFrameWindow*)event.window();
       if (frame)
          {
          Boolean drawInactive = false;
          if (event.eventId() == WM_NCACTIVATE)
             {
             if (event.parameter1().asUnsignedLong() == 0)
                drawInactive = true;
             }
          else if (IPlatform::isWin32S())
                 {
                   ITRACE_DEVELOP("Win32s: WM_NCACTIVATE GetActiveWindow()");
                   if ((HWND)event.handle() != GetActiveWindow())
                     {
                     drawInactive = true;
                     }
                 }
               else
                 {
                   if ((HWND)event.handle() != GetForegroundWindow())
                     {
                     drawInactive = true;
                     }
                 }

          if (frame->fFrameWindowData->fhFont)
             {
             // have font set

             // let Windows do its thing...but first need to arrange it
             // so Windows will think text is null.  It will send us
             // a WM_GETTEXT which we handle below.
             frame->fFrameWindowData->flags |=
               IFrameWindowData::inTitlePaint;

             this->defaultProcedure( event );

             frame->fFrameWindowData->flags &=
               ~IFrameWindowData::inTitlePaint;

             // Check if there is a title, otherwise don't draw anything
             if ( ISTYLEOF( frame->handle() ) & WS_CAPTION )
             {
               IWindowHandle hwnd = event.handle();
               HDC hdc = GetWindowDC( hwnd );
               if (!hdc)
                 ITHROWGUIERROR("GetWindowDC");

               // Get rectangle for title.  It is already in WindowRect coordinates.
               // We can create ITitle without worrying about existing
               // user ITitle object since it does not do startHandlingEventsFor
               ITitle title(frame);
               RECTL drawRect =
                 title.nativeRect().asRECTL();

               drawRect.left += 2;   // Add a pad between system menu and text

               // Setup draw flags
               unsigned int dtFlags = DT_NOPREFIX | DT_SINGLELINE | DT_VCENTER;
               if (IPlatform::isWin95() || IPlatform::isNTNewShell())
                 dtFlags |= DT_LEFT;
               else
                 dtFlags |= DT_CENTER;

               // setup colors
               COLORREF oldfg, oldbg;
               if (drawInactive)
                 {
                 oldbg = SetBkColor(hdc,
                            IColor(IColor::inactiveTitleTextBgnd).asRGBLong());
                 oldfg = SetTextColor(hdc,
                              IColor(IColor::inactiveTitleText).asRGBLong());
                 }
               else
                 {
                 oldbg = SetBkColor(hdc,
                            IColor(IColor::activeTitleTextBgnd).asRGBLong());
                 oldfg = SetTextColor(hdc,
                              IColor(IColor::activeTitleText).asRGBLong());
                 }

               // get title text
               IString titleText( title.text() );

               // setup font
               HFONT prevFont = SelectObject( hdc,
                                              frame->fFrameWindowData->fhFont);
               // draw the text
               DrawText(hdc,
                        (const char* )titleText,
                        (int)titleText.length(),
                        (RECT*)&drawRect,
                        dtFlags );

               // cleanup
               SelectObject( hdc, prevFont );
               SetBkColor( hdc, oldbg );
               SetTextColor( hdc, oldfg );
               ReleaseDC( hwnd, hdc );
             } // if frame has a title
             // event handled
             event.setResult( true );
             result = true;
             }  // if has font set
          } // if frame
       }
       break;

    case WM_SETFONT:
       {
       IFrameWindow* frame = ((IFrameWindow*)(event.window()));
       if ( frame )
          {
          // Store a copy of the HFONT object (IWindow/IFont handle refcounting)
          frame->fFrameWindowData->fhFont = (HFONT)event.parameter1();

          // redraw if specified
          if (event.parameter2().asUnsignedLong())
             RedrawWindow( frame->handle(), 0, 0,
                           RDW_FRAME | RDW_INVALIDATE | RDW_ERASENOW );
          result = true;
          }
       }
       break;

    case WM_GETFONT:
       {
       IFrameWindow* frame = ((IFrameWindow*)(event.window()));
       if ( frame && frame->fFrameWindowData->fhFont )
          {
          // Been changed...return our value
          event.setResult( IEventData(frame->fFrameWindowData->fhFont) );
          result = true;
          }
       }
       break;


    case WM_GETTEXT:
       {
       IFrameWindow* frame = ((IFrameWindow*)(event.window()));
       if ( frame && frame->fFrameWindowData->fhFont )
          {
          // Font set ... need to make sure default procedure gets null string
          unsigned long length = event.parameter1().asUnsignedLong();
          if (frame->fFrameWindowData->flags & IFrameWindowData::inTitlePaint)
             {
             // We are in our title paint routine and the default procedure
             // is asking for the text.  Give it a null string.
             *((char*)event.parameter2()) = '\0';
             length = 0;
             event.setResult( IEventData(length) );
             result = true;
             }
          }
       }
       break;

    case WM_GETTEXTLENGTH:
       {
       IFrameWindow* frame = ((IFrameWindow*)(event.window()));
       if ( frame && frame->fFrameWindowData->fhFont )
          {
          // Font set ... need to make sure default procedure gets null string
          if (frame->fFrameWindowData->flags & IFrameWindowData::inTitlePaint)
             {
             // We are in our title paint routine and the default procedure
             // is asking for the text.  Give it a null string.
             event.setResult(0);
             result = true;
             }
          }
       }
       break;

#endif

#ifdef IC_WIN
    case IC_UM_CONTEXT_HELP:
       {
         // construct HELPINFO structure needed for WM_HELP message

        ITRACE_DEVELOP("IC_UM_CONTEXT_HELP");
         POINT nullPoint;
         HELPINFO helpInfo;
         helpInfo.cbSize = sizeof(HELPINFO);
         helpInfo.MousePos = nullPoint;
         helpInfo.dwContextId  = 0;

         // parm2 of IC_UM_CONTEXT_HELP may contain the
         // handle of the window for which help was requested,
         // else, use window with focus

         if ( event.parameter2().asUnsignedLong() )
           helpInfo.hItemHandle = event.parameter2();
         else
           helpInfo.hItemHandle = IQUERYFOCUS(NULL);


         // for combobox, entryfield or list will be returned on GetFocus instead of the control.
         // If so, change the item handle to that of the combo box

         HWND parentHwnd = IPARENTOF(helpInfo.hItemHandle);
         if (parentHwnd)
         {
           IWindowClassName className(parentHwnd);
           if (className == WC_COMBOBOX)
           {
             helpInfo.hItemHandle = parentHwnd;
           }
         }
         IWindowHandle sendTo = helpInfo.hItemHandle;

         // low word of parm1 of IC_UM_CONTEXT_HELP may contain the
         // control id for which help was requested

         if ( event.parameter1().lowNumber() )
           helpInfo.iCtrlId = event.parameter1().lowNumber();
         else
           helpInfo.iCtrlId = IIDOF(helpInfo.hItemHandle);

         if (IsMenu(helpInfo.hItemHandle))
         {
           helpInfo.iContextType = HELPINFO_MENUITEM;
           IMenu *tempMenu = new IMenu( helpInfo.hItemHandle );
           if (tempMenu && tempMenu->owner())
             sendTo = tempMenu->owner()->handle();
           else
             sendTo = event.handle();
         }
         else
           helpInfo.iContextType = HELPINFO_WINDOW;


         // send the WM_HELP message.
         // for win32s, we need to propagate the message up the parent/owner chain

         IEventResult helpRc;
         Boolean helpProcessed=false;
         while ( !helpProcessed && sendTo != 0 )
         {
           ITRACE_DEVELOP("Sending Generated WM_HELP Event");
           helpRc = sendTo.sendEvent( WM_HELP, 0, (LPHELPINFO)&helpInfo );
           if ( helpRc != 0 || !IPlatform::isWin32S() )
             helpProcessed = true;
           else
           {
             IWindowHandle nextWindow = IPARENTOF(sendTo);
             if ( nextWindow == 0 )
               nextWindow = IOWNEROF(sendTo);
             sendTo = nextWindow;
           }
         }
       }
       break;
#endif

    case IC_UM_DESTROY_FRAME:
       {
#ifdef IC_WIN
         // Be sure to unregister the unique frame class after we destroy
         // the frame window.  This will prevent a memory leak since each
         // time we create a frame window, we register a new unique class
         // name.
         IWindowClassName wcName(event.handle());
#endif
         if ( !IDESTROYWINDOW( event.handle() ) )
         {
           ITHROWGUIERROR("IDESTROYWINDOW");
         }
#ifdef IC_WIN
         else
         {
           // Only unregister the frame classes we create.
           // Do not unregister WC_DIALOG, which is a Windows global class
           if ((wcName.asString().indexOf(WC_FRAME) == 1) )
             UnregisterClass(wcName, GetModuleHandle(0));
         }
#endif
         result = true;
       }
       break;

#ifdef IC_PM
    //---------------------------------------------------------------
    // Watch for private class library message to disable the
    // automatic update of frame extensions when the BIDI attributes
    // for the window change or for command to update extensions
    //---------------------------------------------------------------
    case IC_UM_BIDI_CHANGE:
       {
         IFrameWindow *pFrame = (IFrameWindow*)( event.window() );
         if ( event.parameter1().asUnsignedLong() == 0 )
         {
           // Set flag to disable frame extension update
           pFrame->fFrameWindowData->disableBidiChildUpdate = true;
         }
         else
         {
           // Frame extension should be updated
           IBidiSettings bidi(*pFrame);
           IWindow::ChildCursor cwCursor( *pFrame );
           for ( cwCursor.setToFirst(); cwCursor.isValid();
                 cwCursor.setToNext() )
           {
             IWindowHandle whChild = pFrame->childAt( cwCursor );
             IWindow* pwinChild = IWindow::windowWithHandle( whChild );
             if ( pwinChild && ( pwinChild->id() != FID_CLIENT ))
             {
               bidi.apply(*pwinChild);
             }
           }
         }
       }
       break;

    //---------------------------------------------------------------
    // Look for changes to the BIDI attributes by watching for the
    // WM_SETBIDIATTR message.  This message is not in the OS/2 2.1
    // toolkit, so it is hardcoded (0x0BD0) here.
    //---------------------------------------------------------------
    case 0x0BD0:
       {
         if (event.window()->isFrameWindow())
         {
           IFrameWindow *pFrame = (IFrameWindow*)( event.window() );
           if ( pFrame->fFrameWindowData->disableBidiChildUpdate )
           {
             // Frame extensions should not be updated
             pFrame->fFrameWindowData->disableBidiChildUpdate = false;
           }
           else
           {
             // Frame extension should be updated
             pFrame->postEvent(IC_UM_BIDI_CHANGE,true);
           }
         }
       }
       break;
#endif

#ifdef IC_WIN
    // We need to forward this message on to all child controls.  Windows
    // only sends it to top level windows.  The Windows common controls (tab,
    // listview, etc) are documented as needing it.  In addition, our toolbar
    // button painting code needs to see this message.
    case WM_SYSCOLORCHANGE:
       {
       IWindowHandle hwndFrame( event.handle() );
       if ( !(ISTYLEOF( hwndFrame ) & WS_CHILD) )
          {
          // is top level window
          hwndFrame.sendEvents( event.eventId(),
                                event.parameter1(),
                                event.parameter2(),
                                IWindowHandle::descendants );
          }
       }
       break;
#endif

    default:
      break;
    }
  return result;
  }

/*------------------------------------------------------------------------------
| IFrameHandler::numberOfExtensions                                            |
|                                                                              |
| Return the number of frame extensions that the argument frame possesses.     |
------------------------------------------------------------------------------*/
// Moved to iframe.cpp

/*------------------------------------------------------------------------------
| IFrameHandler::format                                                        |
|                                                                              |
| Fills an array of window position structures, one for each frame control.    |
| This function does this for the frame extensions that have been added to the |
| frame using IFrameWindow::addExtension.                                      |
------------------------------------------------------------------------------*/
// Moved to iframe.cpp

/*------------------------------------------------------------------------------
| IFrameHandler::closed                                                        |
|                                                                              |
| Called when the frame window is closed.                                      |
| Notes:  If the frame is modal, dismiss it (presuming DID_CANCEL).            |
|         If there is a client, send it a WM_SYSCOMMAND/SC_CLOSE so that a     |
|           command handler attached to the client window can process the      |
|           close (but first turn on a special bit indicating that the event   |
|           is dispatched from the frame, so a canvas client will not route    |
|           the command event back to the frame).                              |
|         If the frame indicates it should destroy the window on close, then   |
|           WinDestroyWindow.                                                  |
|         Return true (so default window procedure never gets called)          |
------------------------------------------------------------------------------*/
Boolean IFrameHandler::closed ( IFrameEvent &frameEvent )
  {
  IMODTRACE_DEVELOP( "IFrameHandler::closed" );

  if ( frameEvent.frame()->isModal() )
    frameEvent.frame() -> dismiss( DID_CANCEL );

  IWindowHandle
    client =  frameEvent.frame() -> clientHandle();

  /********************************************/
  /* Notify that the frame window is closing. */
  /********************************************/
  frameEvent.window()->notifyObservers(INotificationEvent(
                                       IFrameWindow::closeId,
                                       *(frameEvent.window()), false));

  if ( client )
    {                          // Client area to route events back to
#ifndef IC_WIN
    unsigned long ul2 = (frameEvent.eventId() == WM_SYSCOMMAND) ?
                           (unsigned long)frameEvent.parameter2() :
                           CMDSRC_OTHER;
    // Route event with special flag in mp2 on to client
    client.sendEvent( WM_SYSCOMMAND,
                      SC_CLOSE,
                      IEventParameter2( (unsigned short)(ul2 | 0x8000),
                                        false ) );
#endif  //IC_PM
#ifdef  IC_WIN
    // Route event with special flag in wparam high order word on to client
    // The high word of wparam is not used in Windows.
    if ( IPlatform::isWin32S() && IWindow::windowWithHandle( client, true ))
      S32to16SendMessage( (HWND)client.asUnsigned()
                        , IC_UM_SYSCOMMAND
                        , IEventParameter1( SC_CLOSE
                                          , 0x8000
                                          | CMDSRC_OTHER )
                        , frameEvent.parameter2() );
    else
      client.sendEvent( WM_SYSCOMMAND,
                        IEventParameter1( SC_CLOSE,
                                          0x8000 | CMDSRC_OTHER ),
                        frameEvent.parameter2() );
#endif
    }

  if ( frameEvent.frame()->willDestroyOnClose() )
    {
       frameEvent.frame()->postEvent(IC_UM_DESTROY_FRAME);
    }
  return true;
  }

/*------------------------------------------------------------------------------
| IFrameHandler::activated                                                     |
|                                                                              |
| Called when the frame window is activated.                                   |
| Notes:  The default implementation simply returns false.                     |
------------------------------------------------------------------------------*/
Boolean IFrameHandler::activated ( IFrameEvent &frameEvent )
  {
  return false;
  }

/*------------------------------------------------------------------------------
| IFrameHandler::deactivated                                                   |
|                                                                              |
| Called when the frame is asked to give up the input focus.                   |
| Notes:  The default implementation simply returns false.                     |
------------------------------------------------------------------------------*/
Boolean IFrameHandler::deactivated ( IFrameEvent &frameEvent )
  {
  return false;
  }

/*------------------------------------------------------------------------------
| IFrameHandler::saved                                                         |
|                                                                              |
| Called when the application statis is to be saved.                           |
| Notes:  The default implementation simply returns false.                     |
------------------------------------------------------------------------------*/
Boolean IFrameHandler::saved ( IFrameEvent &frameEvent )
  {
  return false;
  }

/*------------------------------------------------------------------------------
| IFrameHandler::draw                                                          |
|                                                                              |
| Called when the frame window needs to be redrawn.  This function draws the   |
| separator lines between frame extensions and returns false.                  |
| Notes: Iterate extensions, drawing each via IFrameExtension::drawSeparator.  |
------------------------------------------------------------------------------*/
// Moved to iframe.cpp

/*------------------------------------------------------------------------------
| IFrameHandler::command                                                       |
|                                                                              |
| Called when the frame window receives a command event.  This function checks |
| to see if the frame has a client.  If so, it returns false (so the client    |
| can receive the event).  Otherwise, it returns true.                         |
| Notes:  If there's a client, send it the WM_COMMAND with a special bit       |
|           turned on indicating that the event is dispatched from the frame,  |
|           so that a canvas client area won't route it back to the frame.     |
|         Always return true so PM's frame window procedure won't try to route |
|           it again to the client  (without the bit turn on).                 |
------------------------------------------------------------------------------*/
Boolean IFrameHandler::command ( ICommandEvent &cmdEvent )
{
  if ( cmdEvent.commandId() == IC_ID_HELP )
  {  // Command to request help.
     // Hopefully the client isn't expecting to see this command routed
     // from the frame window.  Since there is no standard protocol for
     // a window procedure to indicate whether it has processed a
     // WM_COMMAND message, we cannot first route it to the client with
     // plans to provide default processing if the client does nothing.
     // The application can always override this default processing if
     // it attaches a command handler to the frame window.
#ifdef IC_WIN
     cmdEvent.dispatchingWindow()->sendEvent( IC_UM_CONTEXT_HELP );
#else
     cmdEvent.dispatchingWindow()->sendEvent( WM_HELP );
#endif
  }
  else
  {  // Some other unprocessed command that we'll route to the client.
     IWindowHandle
        client = IWINDOWFROMID( cmdEvent.handle(), FID_CLIENT );
     if ( client )
     {             // Tinker with WM_COMMAND before sending it to client.
#ifdef IC_WIN
        unsigned long wparam = (unsigned long)cmdEvent.parameter1();
        IEventResult evtres =
          client.sendEvent( WM_COMMAND,
                            IEventParameter1( 0x80000000ul | wparam ),
                            cmdEvent.parameter2() );
                                  // Turn on special flag in wNotifyCode.
#endif //IC_WIN
#ifdef IC_PM
        unsigned long ul2 = (unsigned long)cmdEvent.parameter2();

        IEventResult evtres =
          client.sendEvent( WM_COMMAND,
                            cmdEvent.parameter1(),
                            IEventParameter2( (unsigned short)(ul2 | 0x8000),
                                              false ) );
                                  // Turn on special flag in mp2.
#endif //IC_PM

        cmdEvent.setResult(evtres);
     }
  }
  return true;   // Don't let WC_FRAME winproc also forward to client.
}

/*------------------------------------------------------------------------------
| IFrameHandler::calcRect                                                      |
|                                                                              |
| Iterate extensions, expand frame rectangle by height and width of            |
| each extension if given a client rectangle, shrink client rectangle          |
| by height and width of each extension if given a frame rectangle.            |
------------------------------------------------------------------------------*/
// Moved to iframe.cpp

/*------------------------------------------------------------------------------
| IFrameHandler::IFrameHandler                                                 |
|                                                                              |
| Empty constructor here for page tuning.                                      |
------------------------------------------------------------------------------*/
IFrameHandler::IFrameHandler ( )
  {
  }

/*------------------------------------------------------------------------------
| IFrameHandler::~IFrameHandler                                                |
|                                                                              |
| Empty destructor here for page tuning.                                       |
------------------------------------------------------------------------------*/
IFrameHandler::~IFrameHandler ( )
  {
  }

