/*******************************************************************************
* FILE NAME: igraphbt.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in igraphbt.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

extern "C" {
  #define INCL_WINBUTTONS
  #define INCL_WINERRORS
  #define INCL_WINMESSAGEMGR
  #define INCL_WINSYS
  #define INCL_GPIPRIMITIVES
  #include <iwindefs.h>
}

#include <igraphbt.hpp>
#include <iexcept.hpp>
#include <ifont.hpp>
#include <istring.hpp>
#include <ihandle.hpp>
#include <ipoint.hpp>
#include <ireslib.hpp>
#include <iiconctl.hpp>
#include <igrszhdr.hpp>
#include <igraphdr.hpp>


#ifdef IC_PAGETUNE
  #define _IGRAPHBT_CPP_
  #include <ipagetun.h>
#endif


/*------------------------------------------------------------------------------
| Public push button styles.                                                   |
------------------------------------------------------------------------------*/

#pragma data_seg(ICLStaticConst)
const IGraphicPushButton::Style
#ifndef IC_WIN
  IGraphicPushButton::sizeToGraphic     = BS_ICON,
#endif
#ifdef IC_WIN
  IGraphicPushButton::sizeToGraphic  (0,IBS_ICON),
#endif
  IGraphicPushButton::classDefaultStyle = WS_VISIBLE;
#pragma data_seg()

/*------------------------------------------------------------------------------
| Default style for new objects (initial value).                               |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLNonConst)
IGraphicPushButton::Style
  IGraphicPushButton::currentDefaultStyle = WS_VISIBLE;
#pragma data_seg()

/*------------------------------------------------------------------------------
| This class wrappers an object pointer and destroys the pointed               |
| to object when the wrapper is destructed.                                    |
------------------------------------------------------------------------------*/
static struct DefaultHandler {
IGraphicPushButtonResizeHandler*
  resizeHandler();
IGraphicPushButtonHandler*
  handler();

~DefaultHandler ( )
{
  if (size)
  {
     delete size;
     size = 0;
  }
  if (paint)
  {
     delete paint;
     paint = 0;
  } /* endif */
}
IGraphicPushButtonResizeHandler
 *size;
IGraphicPushButtonHandler
 *paint;
} defaultHandler;

/*------------------------------------------------------------------------------
| DefaultHandler :: resizeHandler                                              |
------------------------------------------------------------------------------*/
IGraphicPushButtonResizeHandler* DefaultHandler :: resizeHandler()
{
  if ( !size )
     size = new IGraphicPushButtonResizeHandler;
  return size;
}

/*------------------------------------------------------------------------------
| DefaultHandler :: handler                                                    |
------------------------------------------------------------------------------*/
IGraphicPushButtonHandler* DefaultHandler :: handler()
{
  if ( !paint )
     paint = new IGraphicPushButtonHandler;
  return paint;
}

/*------------------------------------------------------------------------------
| IGraphicPushButton::IGraphicPushButton                                       |
|                                                                              |
| Constructor to create a graphic push button control on a standard window.    |
------------------------------------------------------------------------------*/
IGraphicPushButton :: IGraphicPushButton(unsigned long id,
                             IWindow* parent,
                             IWindow* owner,
                             const IResourceId& bitmapOrIconId,
                             const IRectangle& initial,
                             const Style& style)
              : IPushButton(id, parent,owner,IRectangle(),IWindow::noStyle),
                bSizeTo(false), returnBack(false), border(3,3),
                fGraphicPushButtonData( new IGraphicPushButtonData() )

{
  IASSERTPARM(parent!=0);
  initialize(style, initial, id);
  setGraphic(bitmapOrIconId);               // load and set graphic
  if (style & IWindow::visible)
  {
     show();
     refresh(IWindow::immediate);
  }
}

/*------------------------------------------------------------------------------
| IGraphicPushButton::IGraphicPushButton                                       |
|                                                                              |
| Constructor to create a graphic push button control on a standard window.    |
------------------------------------------------------------------------------*/
IGraphicPushButton :: IGraphicPushButton(unsigned long id,
                             IWindow* parent,
                             IWindow* owner,
                             unsigned long bitmapOrIconId,
                             const IRectangle& initial,
                             const Style& style)
              : IPushButton(id, parent,owner, IRectangle(),IWindow::noStyle),
                bSizeTo(false), returnBack(false), border(3,3),
                fGraphicPushButtonData( new IGraphicPushButtonData()  )
{
  IASSERTPARM(parent!=0);
  initialize(style, initial, id);
  setGraphic(bitmapOrIconId);               // load and set graphic
  if (style & IWindow::visible)
  {
     show();
     refresh(IWindow::immediate);
  }
}


/*------------------------------------------------------------------------------
| IGraphicPushButton::IGraphicPushButton                                       |
|                                                                              |
| Constructor to create a graphic push button control on a standard window.    |
------------------------------------------------------------------------------*/
IGraphicPushButton :: IGraphicPushButton(unsigned long id,
                             IWindow* parent,
                             IWindow* owner,
                             const IBitmapHandle& bitmap,
                             const IRectangle& initial,
                             const Style& style)
              : IPushButton(id, parent,owner, IRectangle(),IWindow::noStyle),
                bSizeTo(false), returnBack(false), border(3,3),
                fGraphicPushButtonData( new IGraphicPushButtonData() )
{
  IASSERTPARM(parent!=0);
  initialize(style, initial, id);
  setGraphic(bitmap);               // load and set graphic
  if (style & IWindow::visible)
  {
     show();
     refresh(IWindow::immediate);
  }
}


/*------------------------------------------------------------------------------
| IGraphicPushButton::IGraphicPushButton                                       |
|                                                                              |
| Constructor to create a graphic push button control on a standard window.    |
------------------------------------------------------------------------------*/
IGraphicPushButton :: IGraphicPushButton(unsigned long id,
                             IWindow* parent,
                             IWindow* owner,
                             const IPointerHandle& icon,
                             const IRectangle& initial,
                             const Style& style)
              : IPushButton(id, parent,owner, IRectangle(),IWindow::noStyle),
                bSizeTo(false), returnBack(false), border(3,3),
                fGraphicPushButtonData( new IGraphicPushButtonData() )
{
  IASSERTPARM(parent!=0);
  initialize(style, initial, id);
  setGraphic(icon);               // load and set graphic
  if (style & IWindow::visible)
  {
     show();
     refresh(IWindow::immediate);
  }
}


/*------------------------------------------------------------------------------
| IGraphicPushButton::IGraphicPushButton                                       |
|                                                                              |
| Constructor to instantiate an object for a dialog template.                  |
------------------------------------------------------------------------------*/
IGraphicPushButton :: IGraphicPushButton(unsigned long ulId,
                           IWindow* pdlgwndParent)
              : IPushButton(ulId,pdlgwndParent),
                bSizeTo(true), returnBack(true), border(3,3),
                fGraphicPushButtonData( new IGraphicPushButtonData() )
{
  IASSERTPARM(pdlgwndParent!=0);

#ifdef IC_PM
  BTNCDATA btn;
  WNDPARAMS wnd;
  wnd.fsStatus  = WPM_CTLDATA;
  wnd.cbCtlData = btn.cb = sizeof(BTNCDATA);
  wnd.pCtlData  = &btn;
  if (!(unsigned long)(handle().sendEvent(WM_QUERYWINDOWPARAMS, IEventParameter1(&wnd),
                                                 IEventParameter2(0))))
     ITHROWGUIERROR("WM_QUERYWINDOWPARAMS");
  LHANDLE hImage = btn.hImage;
#endif

#ifdef IC_WIN
  // The BS_ICON and BS_BITMAP styles are Win95 only.   Attempts to
  // use this function on NT 3.51 will probably result in the exception
  // being thrown.  BM_GETIMAGE is also documented as Win95 only.
  HANDLE hImage = (HANDLE) handle().sendEvent( BM_GETIMAGE );
  if (!hImage)
     ITHROWGUIERROR("BM_GETIMAGE");
#endif

  aIcon = new IIconControl(1,this,this);
  if (style() & BS_BITMAP)
  {
     aIcon->setBitmap(IBitmapHandle(hImage));
     contains = bitmapType;
  }
  else
  {
     aIcon->setIcon(IPointerHandle(hImage));
     contains = iconType;
  }
  this -> addHandler( defaultHandler.resizeHandler() );
  this -> addHandler( defaultHandler.handler() );
}


/*------------------------------------------------------------------------------
| IGraphicPushButton::IGraphicPushButton                                       |
|                                                                              |
| Constructor to instantiate from an existing push button control.             |
------------------------------------------------------------------------------*/
IGraphicPushButton :: IGraphicPushButton(const IWindowHandle& wh)
              : IPushButton(wh),
                bSizeTo(true), returnBack(true), border(3,3),
                fGraphicPushButtonData( new IGraphicPushButtonData() )
{
#ifdef IC_PM
  BTNCDATA btn;
  WNDPARAMS wnd;
  wnd.fsStatus  = WPM_CTLDATA;
  wnd.cbCtlData = btn.cb = sizeof(BTNCDATA);
  wnd.pCtlData  = &btn;
  if (!(unsigned long)(handle().sendEvent(WM_QUERYWINDOWPARAMS, IEventParameter1(&wnd),
                                                 IEventParameter2(0))))
     ITHROWGUIERROR("WM_QUERYWINDOWPARAMS");
  LHANDLE hImage = btn.hImage;
#endif

#ifdef IC_WIN
  // The BS_ICON and BS_BITMAP styles are Win95 only.   Attempts to
  // use this function on NT 3.51 will probably result in the exception
  // being thrown.  BM_GETIMAGE is also documented as Win95 only.
  HANDLE hImage = (HANDLE) handle().sendEvent( BM_GETIMAGE );
  if (!hImage)
     ITHROWGUIERROR("BM_GETIMAGE");
#endif

  aIcon = new IIconControl(1,this,this);
  if (style() & BS_BITMAP)
  {
     aIcon->setBitmap(IBitmapHandle(hImage));
     contains = bitmapType;
  }
  else
  {
     aIcon->setIcon(IPointerHandle(hImage));
     contains = iconType;
  }
  this -> addHandler( defaultHandler.resizeHandler() );
  this -> addHandler( defaultHandler.handler() );
}

/*------------------------------------------------------------------------------
| IGraphicPushButton::~IGraphicPushButton                                      |
------------------------------------------------------------------------------*/
IGraphicPushButton :: ~IGraphicPushButton()
{
   if ( returnBack && isValid() )
   {
#ifdef IC_PM
      if (style() & BS_BITMAP)
         setControlData(bitmap());
      else
         setControlData((IBitmapHandle)icon());
#endif
#ifdef IC_WIN
      if ( currentGraphicType() == bitmapType )
         {
         setStyle( style() & (unsigned long) BS_BITMAP );
         setControlData(bitmap());
         }
      else
         {
         setStyle( style() & (unsigned long) BS_ICON );
         setControlData((IBitmapHandle)icon());
         }
#endif
   } /* endif */
  this -> removeHandler( defaultHandler.resizeHandler() );
  this -> removeHandler( defaultHandler.handler() );
  if (aIcon)
     delete aIcon;
  delete fGraphicPushButtonData;
}


/*------------------------------------------------------------------------------
| IGraphicPushButton::defaultStyle                                             |
------------------------------------------------------------------------------*/
IGraphicPushButton::Style  IGraphicPushButton :: defaultStyle()
{
  return currentDefaultStyle;
}


/*------------------------------------------------------------------------------
| IGraphicPushButton::setDefaultStyle                                          |
------------------------------------------------------------------------------*/
void  IGraphicPushButton :: setDefaultStyle(const IGraphicPushButton::Style& pbsStyle)
{
  currentDefaultStyle = pbsStyle;
}


/*------------------------------------------------------------------------------
| IGraphicPushButton::convertToGUIStyle                                        |
|                                                                              |
| Returns base style for the control by default, or extended style if          |
| extended flag (bExtOnly) is set.                                             |
------------------------------------------------------------------------------*/
unsigned long IGraphicPushButton :: convertToGUIStyle(const IBitFlag& guiStyle,
                                                      Boolean bExtOnly) const
{
  // Obtain the style from the class (IPushButton) that we inherit from
  unsigned long ulStyle = Inherited::convertToGUIStyle( guiStyle, bExtOnly );

  if (bExtOnly)
  {
    // Use mask to only return extended styles in the user defined range
    ulStyle |= guiStyle.asExtendedUnsignedLong() & IS_EXTMASK;
  }
  else
  {
    // Let the clean styles flow thru
    ulStyle |= guiStyle.asUnsignedLong() & IGB_MASK;
#ifndef IC_WIN
    // add the BS_BITMAP style,  BS_PUSHBUTTON will be added by IPushButton
    ulStyle |= BS_BITMAP;
#endif
    // In Windows WS_CHILD and BS_PUSHBUTTON or BS_DEFPUSHBUTTON added
    // by IPushButton.
  }

  return( ulStyle );
}


/*------------------------------------------------------------------------------
| IGraphicPushButton::setGraphic                                               |
------------------------------------------------------------------------------*/
IGraphicPushButton& IGraphicPushButton :: setGraphic(const IResourceId& bitmapOrIconId)
{

  IBitmapHandle bmp = bitmapOrIconId.resourceLibrary().tryToLoadBitmap(bitmapOrIconId);
  if (bmp == 0)
  {
     setGraphic(bitmapOrIconId.resourceLibrary().loadIcon(bitmapOrIconId));
  }
  else
     setGraphic(bmp);
  return *this;
}


/*------------------------------------------------------------------------------
| IGraphicPushButton::setGraphic                                               |
------------------------------------------------------------------------------*/
IGraphicPushButton& IGraphicPushButton :: setGraphic(unsigned long bitmapOrIconId)
{

  setGraphic(IResourceId(bitmapOrIconId));
  return *this;
}


/*------------------------------------------------------------------------------
| IGraphicPushButton::setGraphic                                               |
------------------------------------------------------------------------------*/
IGraphicPushButton&  IGraphicPushButton :: setGraphic(const IBitmapHandle& bitmapHandle)
{
  aIcon->setBitmap(bitmapHandle);
  contains=bitmapType;
  if (bSizeTo)
  {
#ifdef IC_PM
     setStyle((style() & ~BS_ICON) | BS_BITMAP);
     setControlData(bitmapHandle);
#endif
#ifdef IC_WIN
      // size the icon control to its minimum size
      aIcon->sizeTo( aIcon->minimumSize( true ) );
#endif
  }
  refresh();
  return *this;
}


/*------------------------------------------------------------------------------
| IGraphicPushButton::setGraphic                                               |
------------------------------------------------------------------------------*/
IGraphicPushButton& IGraphicPushButton :: setGraphic(const IPointerHandle& handle)
{
  aIcon->setIcon(handle);
  contains=iconType;
  if (bSizeTo)
  {
#ifdef IC_PM
     setStyle((style() & ~BS_BITMAP) | BS_ICON);
     setControlData((IBitmapHandle)handle);
#endif
#ifdef IC_WIN
     // size the icon control to its minimum size
     aIcon->sizeTo( aIcon->minimumSize( true ) );
#endif
  }
  refresh();
  return *this;
}


/*------------------------------------------------------------------------------
| IGraphicPushButton::graphicWindow                                            |
------------------------------------------------------------------------------*/
IIconControl&  IGraphicPushButton :: graphicWindow() const
{
   return *aIcon;
}


/*------------------------------------------------------------------------------
| IGraphicPushButton::bitmap                                                   |
------------------------------------------------------------------------------*/
IBitmapHandle  IGraphicPushButton :: bitmap() const
{
   return aIcon->bitmap();
}


/*------------------------------------------------------------------------------
| IGraphicPushButton::icon                                                     |
------------------------------------------------------------------------------*/
IPointerHandle  IGraphicPushButton :: icon() const
{
   return aIcon->icon();
}


/*------------------------------------------------------------------------------
| IGraphicPushButton::currentGraphicType                                       |
------------------------------------------------------------------------------*/
IGraphicPushButton::GraphicType IGraphicPushButton :: currentGraphicType ( ) const
{
   return contains;
}

/*------------------------------------------------------------------------------
| IGraphicPushButton::setControlData                                           |
------------------------------------------------------------------------------*/
IGraphicPushButton&  IGraphicPushButton :: setControlData(const IBitmapHandle& aHandle)
{
#ifdef IC_PM
   BTNCDATA btn;
   WNDPARAMS wnd;
   wnd.fsStatus  = WPM_CTLDATA;
   wnd.cbCtlData = btn.cb = sizeof(BTNCDATA);
   wnd.pCtlData  = &btn;
   btn.hImage = aHandle;
   btn.fsCheckState = 0;
   btn.fsHiliteState = 0;
   if (!(unsigned long)(handle().sendEvent(WM_SETWINDOWPARAMS, IEventParameter1(&wnd),
                                                  IEventParameter2(0))))
      ITHROWGUIERROR("WM_SETWINDOWPARAMS");
#endif
#ifdef IC_WIN
   // Message documented as Win95 only.  We only call this for wrappered
   // controls being destructed in Windows.
   handle().sendEvent(BM_SETIMAGE,
                      IEventParameter1( 0 ),
                      IEventParameter2( aHandle.asUnsigned() ) );
#endif

   return *this;
}

/*------------------------------------------------------------------------------
| IGraphicPushButton::enableSizeToGraphic                                      |
------------------------------------------------------------------------------*/
IGraphicPushButton&  IGraphicPushButton :: enableSizeToGraphic( Boolean enable)
{
   if (enable && !bSizeTo)
   {
      aIcon->sizeTo(ISize(0,0));
      bSizeTo=true;
#ifdef IC_PM
      if (currentGraphicType() == bitmapType)
      {
         setStyle((style() & ~BS_ICON) | BS_BITMAP);
         setControlData(bitmap());
      }
      else
      {
         setStyle((style() & ~BS_BITMAP) | BS_ICON);
         setControlData((IBitmapHandle)icon());
      }
#endif
#ifdef IC_WIN
      // size the icon control to its minimum size
      aIcon->sizeTo( aIcon->minimumSize( true ) );
#endif
      refresh();
   }
   else if (!enable && bSizeTo)
   {
      bSizeTo=false;
      ISize initial = size();
#ifdef IC_PM
      setStyle((style() & ~BS_ICON) | BS_BITMAP);
      setControlData(IBitmapHandle(0));
#endif
      sizeTo(ISize(initial.width()-1,initial.height()));
      sizeTo(initial);
      refresh();
   } /* endif */
   return *this;
}


/*------------------------------------------------------------------------------
| IGraphicPushButton::disableSizeToGraphic                                     |
------------------------------------------------------------------------------*/
IGraphicPushButton&  IGraphicPushButton :: disableSizeToGraphic()
{
   enableSizeToGraphic(false);
   return *this;
}


/*------------------------------------------------------------------------------
| IGraphicPushButton::isSizeToGraphic                                          |
------------------------------------------------------------------------------*/
Boolean  IGraphicPushButton :: isSizeToGraphic() const
{
   return bSizeTo;
}

/*------------------------------------------------------------------------------
| IGraphicPushButton::calcMinimumSize                                          |
|                                                                              |
| Calculate the minimum screen size needed by the control.  The height is      |
| based on the bitmap/icon height and border height and the width is based on  |
| the bitmap/icon width and the border width.                                  |
------------------------------------------------------------------------------*/
ISize  IGraphicPushButton :: calcMinimumSize() const
{
  ISize size = aIcon->minimumSize();
  return ISize(size.width()+2*border.width(),size.height()+2*border.height());
}

/*------------------------------------------------------------------------------
| IGraphicPushButton::initialize                                               |
------------------------------------------------------------------------------*/
IGraphicPushButton&  IGraphicPushButton :: initialize(const Style& bmsStyle,
                                                      const IRectangle& initial,
                                                      unsigned long id)
{
  aIcon = new IIconControl(id,this,this);

  // Save the extended style to make sure we have a copy of it stored
  setExtendedStyle( extendedStyle() | bmsStyle.asExtendedUnsignedLong() );

  unsigned long ulStyle = convertToGUIStyle( bmsStyle );

  this -> addHandler( defaultHandler.resizeHandler() );
  this -> addHandler( defaultHandler.handler() );

  moveSizeTo(initial);
  if (bmsStyle & IGraphicPushButton::sizeToGraphic )
  {
     bSizeTo = true;
     aIcon->sizeTo(ISize(0,0));
  }

  if (ulStyle & IWindow::visible.asUnsignedLong())
     ulStyle &= (~IWindow::visible.asUnsignedLong());

  setStyle(ulStyle);
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicPushButton::borderSize                                               |
------------------------------------------------------------------------------*/
ISize  IGraphicPushButton :: marginSize() const
{
   return ISize(border);
}

/*------------------------------------------------------------------------------
| IGraphicPushButton::setBorderSize                                            |
------------------------------------------------------------------------------*/
IGraphicPushButton&  IGraphicPushButton :: setMarginSize(const ISize& newSize)
{
   long inith  = newSize.height(),
        initw  = newSize.width(),
        width  = -initw,
        height = -inith;

   border.setHeight(inith).setWidth(initw);
   ISize window = size();
   if ((width += window.width() ) < 0)
      width = 0;
   if ((height+= window.height()) < 0)
      height = 0;
   if (!isSizeToGraphic())
      aIcon->moveSizeTo(IRectangle(initw,inith,width,height));
   return *this;
}

