/*******************************************************************************
* FILE NAME: imenubar.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in imenubar.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

extern "C" {
  #define INCL_WINFRAMEMGR
  #define INCL_WINWINDOWMGR
  #define INCL_WINMENUS
  #include <iwindefs.h>
}

#include <iexcept.hpp>
#include <iframe.hpp>
#include <iframprv.hpp>
#include <imenubar.hpp>
#include <imenuprv.hpp>
#include <iplatfrm.hpp>
#include <ireslib.hpp>
#include <itrace.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IMENUBAR_CPP_
  #include <ipagetun.h>
#endif

#pragma data_seg(ICLStaticConst)
const IMenuBar::Style
  IMenuBar::empty             ( 0, IMBS_EMPTY ),
  IMenuBar::wrapper           ( 0, IMBS_WRAPPER ),
  IMenuBar::classDefaultStyle ( 0, IMBS_EMPTY );
#pragma data_seg()

#pragma data_seg(ICLNonConst)
  IMenuBar::Style
              IMenuBar::currentDefaultStyle ( 0, IMBS_EMPTY );
#pragma data_seg()

/*------------------------------------------------------------------------------
| IMenuBar::IMenuBar                                                           |
|  Create empty or wrapper                                                     |
------------------------------------------------------------------------------*/
IMenuBar::IMenuBar( IFrameWindow*      owner,
                    const Style&       style )
                  : fMenuBarData( 0 )
{
   IASSERTPARM(owner != 0);

   // Set extended style
   setExtendedStyle( extendedStyle() | style.asExtendedUnsignedLong() );

   /* Check for both empty and wrapper being set,
    * or neither set.
    */
   if ( ( (style & wrapper) && (style & empty) ) ||       // both set
        ( !(style & (wrapper | empty)) ) )     // neither set
   {
     ITHROWLIBRARYERROR(IC_INVALIDSTYLE,
                        IBaseErrorInfo::invalidRequest,
                        IException::recoverable);
   }
#ifdef IC_WIN
   // Query for an existing menu and get the handle of it
   HMENU frameMenuBar = GetMenu( owner->handle() );
#else
   HWND frameMenuBar = handleWithParent(FID_MENU, owner->handle());
#endif

   if (style & empty)
   {
#ifdef IC_WIN
     // Create a new empty menu
     fhMenu = CreateMenu();
     fowner = owner->handle();
     fmenuFrame = owner;

     // Set this new menu as the menu for the window
     SetMenu( fowner, fhMenu );

     // If there was a previous menu, destroy it
     if (frameMenuBar)
     {
       DestroyMenu( frameMenuBar );

       // Reset lookup table pointer since menu is now empty
       if (!(IPlatform::isWin95() || IPlatform::isNTNewShell()))
         owner->fFrameWindowData->fpSubmenuList = 0;
     }

     // Update frame window to display with new menubar
     DrawMenuBar( fowner );
#else
     IWindowHandle whMenu =
          this -> create( FID_MENU,
                          0,
                          style.asUnsignedLong() | MS_ACTIONBAR |
                          WS_VISIBLE | WS_SYNCPAINT,
                          WC_MENU,
                          owner->handle(),
                          owner->handle(),
                          IRectangle(),
                          0,
                          0,
                          IWindow::onTopOfSiblings );

     if (frameMenuBar)
        WinDestroyWindow(frameMenuBar);

     owner->postEvent(WM_UPDATEFRAME,            // Cause frame to update
                      IEventData(FCF_MENU, 0),   // to include menu.
                      IEventData());

     startHandlingEventsFor(whMenu);
#endif
   }
   else    // it's a wrapper
   {
     if (frameMenuBar == 0) {
       ITHROWLIBRARYERROR(IC_NO_FRAME_MENUBAR,
                          IBaseErrorInfo::invalidRequest,
                          IException::recoverable);
     } /* endif */

#ifdef IC_WIN
     // Save the menu handle and the owner in private data
     fhMenu = frameMenuBar;
     fowner = owner->handle();
     fmenuFrame = owner;
#else
     setAutoDestroyWindow(false);
     startHandlingEventsFor(frameMenuBar);
#endif

   }
}

/*------------------------------------------------------------------------------
| IMenuBar::IMenuBar                                                           |
|  Create the menu from resource file                                          |
------------------------------------------------------------------------------*/
IMenuBar::IMenuBar( const IResourceId& menuResId,
                    IFrameWindow*      owner )
                  : fMenuBarData( 0 )
{
   IMODTRACE_DEVELOP("IMenuBar::ctor(2)");
   IASSERTPARM(owner != 0);

#ifdef IC_WIN
   // Query for an existing menu and get the handle of it
   HMENU frameMenuBar = GetMenu( owner->handle() );
   fhMenu = (void*)
#else
   HWND frameMenuBar = handleWithParent(FID_MENU, owner->handle());
   IWindowHandle whMenu =
#endif
     (menuResId.resourceLibrary()).loadMenu(menuResId, owner);

#ifdef IC_WIN
   // set owner handle
   fowner = owner->handle();
   fmenuFrame = owner;

   // Set this new menu as the menu for the window
   SetMenu( fowner, fhMenu );

   // If there was a previous menu, destroy it
   if (frameMenuBar)
     DestroyMenu( frameMenuBar );

   // Build the lookup table for this if old menu style
   if (!(IPlatform::isWin95() || IPlatform::isNTNewShell()))
   {
     if (frameMenuBar)
       IMenuPrivate::deleteLookUpTable( fmenuFrame );
     IMenuPrivate::buildLookUpTableFromResourceLoad( fmenuFrame,
                                                     fhMenu, menuResId );
   }

   // Update frame window to display with new menubar
   DrawMenuBar( fowner );
#else
   if(frameMenuBar)                     //Destroy the old FID_MENU
      WinDestroyWindow(frameMenuBar);

   owner->postEvent(WM_UPDATEFRAME,            // Cause frame to update
                    IEventData(FCF_MENU, 0),   // to include menu.
                    IEventData());

   startHandlingEventsFor(whMenu);
#endif
}


/*------------------------------------------------------------------------------
| IMenuBar::~IMenuBar                                                          |
------------------------------------------------------------------------------*/
IMenuBar::~IMenuBar()
{ }

/*------------------------------------------------------------------------------
| IMenuBar::setMenu                                                            |
|  Set a new menu resource ID. The previous menu-bar is destroyed.             |
------------------------------------------------------------------------------*/
IMenuBar& IMenuBar::setMenu(const IResourceId& menuResId)
{
#ifdef IC_WIN
   // Query for an existing menu and get the handle of it
   HMENU oldMenuBar = fhMenu;
   fhMenu = (void*)
#else
   IWindowHandle hwndMenu =
#endif
     (menuResId.resourceLibrary()).loadMenu(menuResId, owner());

#ifdef IC_WIN
   // Set this new menu as the menu for the window
   SetMenu( fowner, fhMenu );

   // If there was a previous menu, destroy it
   if (oldMenuBar)
     DestroyMenu( oldMenuBar );

   // If old menu style, get rid of old lookup table and then build new one
   if (!(IPlatform::isWin95() || IPlatform::isNTNewShell()))
   {
     if (oldMenuBar)
       IMenuPrivate::deleteLookUpTable( fmenuFrame );
     IMenuPrivate::buildLookUpTableFromResourceLoad( fmenuFrame,
                                                     fhMenu, menuResId );
   }

   // Update frame window to display with new menubar
   DrawMenuBar( fowner );
#else
   if (id() != FID_MENU) {
      WinSetWindowUShort(hwndMenu,  //Menu window ID is create with
                         QWS_ID,    //FID_MENU. Change it.
                         (unsigned short)id());
   } else {
     IWindow* powner = owner();
     if (isValid())
         WinDestroyWindow(handle());
      powner->postEvent(WM_UPDATEFRAME,              // Cause frame to update
                       IEventData(FCF_MENU, 0),     // to include menu.
                       IEventData());
   } /* endif */
   startHandlingEventsFor(hwndMenu);
#endif
   return *this;
}

/*------------------------------------------------------------------------------
| IMenuBar::defaultStyle                                                       |
|                                                                              |
| Return the default style for new menu bar objects.                           |
------------------------------------------------------------------------------*/
IMenuBar::Style IMenuBar::defaultStyle()
{
  return currentDefaultStyle;
}

#ifdef IC_WIN
/*------------------------------------------------------------------------------
| IMenuBar::setDefaultStyle                                                       |
|                                                                              |
| Set the default style for new menu bar objects.                              |
------------------------------------------------------------------------------*/
void IMenuBar::setDefaultStyle( const IMenuBar::Style& newDefault )
{
  currentDefaultStyle = newDefault;
}
#endif

/*------------------------------------------------------------------------------
| IMenuBar::convertToGUIStyle                                                  |
|                                                                              |
| Returns base style for the control by default, or extended style if          |
| extended flag (bExtOnly) is set.                                             |
------------------------------------------------------------------------------*/
unsigned long IMenuBar::convertToGUIStyle( const IBitFlag& guiStyle,
                                           Boolean         bExtOnly ) const
{
  // Obtain the style from the class (IMenu) that we inherit from
  unsigned long ulStyle = Inherited::convertToGUIStyle( guiStyle, bExtOnly );

  if (bExtOnly)
  {
    // Use mask to return extended styles in the user defined range and
    // extended styles for IMenuBar, IMBS_, which have a one-to-one
    // correspondence to our style bits
    ulStyle |= guiStyle.asExtendedUnsignedLong() & IMBS_EXTMASK;
  }
  else
  {
    // No base styles for IMenuBar, but use mask for future expandability
    ulStyle |= guiStyle.asUnsignedLong() & IMBS_MASK;
  }

  return( ulStyle );
}

