/*******************************************************************************
* FILE NAME: inotehdr.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Implementation of the class(es):                                           *
*     INotebookHandler                                                         *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C" {
  #define INCL_WINFRAMEMGR
  #define INCL_WINSTDBOOK
  #define INCL_WINDIALOGS           // Needed for PDLGTEMPLATE in BOOKPAGEINFO
  #include <iwindefs.h>

#ifdef IC_WIN
  #ifndef _WIN32
    #define _WIN32
  #endif
  #include <commctrl.h>
#endif
}

#include <inotehdr.hpp>

#ifndef _IWINDOW_
  #include <iwindow.hpp>
#endif

#ifdef IC_WIN
  #ifndef _ICLNOTEBOOKW_
    #include <iclnbw.h>
  #endif
  #ifndef _IWCNAME_
    #include <iwcname.hpp>
  #endif
  #ifndef _IKEYEVT_
    #include <ikeyevt.hpp>
  #endif
#endif

#ifndef _INOTEBK_
  #include <inotebk.hpp>
#endif
#ifndef _IPLATFRM_
  #include <iplatfrm.hpp>
#endif

// Segment definitions
#ifdef IC_PAGETUNE
  #define _INOTEHDR_CPP_
  #include <ipagetun.h>
#endif

static unsigned long ulCurrentPageId = 0;


/*------------------------------------------------------------------------------
| INotebookHandler::dispatchHandlerEvent                                       |
|                                                                              |
| Dispatch command functions for this handler.                                 |
------------------------------------------------------------------------------*/
Boolean INotebookHandler :: dispatchHandlerEvent ( IEvent &event )
{
  Boolean stopProcessingEvent = false;

  switch ( event.eventId() )
  {
#ifndef IC_WIN
    case WM_HELP:
      {
        if ( !event.window()->hasFocus() )
        {          // Focus is not on the notebook (and hence, not a tab).
          IWindow* parent = event.window()->parent();
          if (parent)
          {
            /******************************************************************/
            /* The user must have requested contextual help from a            */
            /* page window.  We will avoid the notebook's processing          */
            /* of WM_HELP, which is geared for its tabs.  If the help         */
            /* request gets here, the page window must not be a frame         */
            /* (it is probably a canvas).  By routing the requests to         */
            /* the notebook's parent, these contextual helps can be           */
            /* resolved through the help subtable of the notebook's           */
            /* frame window.                                                  */
            /******************************************************************/
            parent->handle().postEvent( event.eventId(),
                                        event.parameter1(),
                                        event.parameter2() );
            stopProcessingEvent = true;
          }
        }
      }
      break;
#endif //not IC_WIN

#ifdef IC_WIN
    case WM_HELP:
      {
        /**********************************************************************/
        /* Get a pointer to the notebook object.                              */
        /**********************************************************************/
        INotebook* pNotebk = (INotebook *)event.controlWindow();
        IWindow* parent = event.window()->parent();
        HELPINFO helpInfo=*((LPHELPINFO)(void *)(event.parameter2()));

        /**********************************************************************/
        /* If focus is on a tab control's tab, (native and/or ccl)            */
        /*   then post BKN_HELP (processed by IPageHandler::help().           */
        /**********************************************************************/
        if (( pNotebk ) &&
            ( event.handle()==helpInfo.hItemHandle ) )
        {
          /********************************************************************/
          /* Windows tab control processing.  Get the top page handle.        */
          /********************************************************************/
          IPageHandle topPageHandle = pNotebk->topPage();

          if (parent)
          {
            /******************************************************************/
            /* The user must have requested contextual help on a tab          */
            /* control's tab.  Simulate the tab help notification for the     */
            /* CUA '91 notebook control by posting the page help event        */
            /* to the parent of the notebook control.                         */
            /******************************************************************/
            parent->handle().postEvent(
                               WM_COMMAND,
                               IEventParameter1( (unsigned short)pNotebk->id(),
                                                 BKN_HELP ),
                               IEventParameter2( topPageHandle ) );

            stopProcessingEvent = true;
          }
        }
        else
        {
          if (parent)
          {
            /******************************************************************/
            /* The user must have requested contextual help from a            */
            /* page window.  We will avoid the notebook's processing          */
            /* of WM_HELP, which is geared for its tabs.  If the help         */
            /* request gets here, the page window must not be a frame         */
            /* (it is probably a canvas).  By routing the requests to         */
            /* the notebook's parent, these contextual helps can be           */
            /* resolved through the help subtable of the notebook's           */
            /* frame window.  Please note that we must send the help message  */
            /* rather than post it since parameter2 is a pointer to the       */
            /* HELPINFO structure.                                            */
            /******************************************************************/
            parent->handle().sendEvent( event.eventId(),
                                        event.parameter1(),
                                        event.parameter2() );

            stopProcessingEvent = true;
          }
        }
      } //WM_HELP
      break;

    case WM_SIZE:
    case IC_UM_TCM_SIZE:
      {
        /**********************************************************************/
        /* Only process Windows tab control sizing event.                     */
        /**********************************************************************/
        IWindowHandle hwndControl( event.controlHandle() );
        if ( !(hwndControl &&
             ( IWindowClassName( hwndControl ) == WC_TABCONTROL )) )
          break;

        /**********************************************************************/
        /* Obtain pointer to notebook object.                                 */
        /**********************************************************************/
        INotebook*
          pNotebk = (INotebook*)IWindow::windowWithHandle( hwndControl );

        /**********************************************************************/
        /* If all tabs are to be shown (i.e. multi rows), we have to size the */
        /* tab control prior to calculating the display rectangle, to prevent */
        /* the application page window from clipping the bottom row of tabs.  */
        /**********************************************************************/
        ISize newSize( event.parameter2().lowNumber(),
                       event.parameter2().highNumber() );

        if ( pNotebk->areAllTabsVisible() )
        {
          if ( event.eventId() ==  WM_SIZE )
          {
            SetWindowPos( pNotebk->handle(),
                          NULL, 0, 0,
                          (int)newSize.width(),
                          (int)newSize.height(),
                          SWP_NOMOVE | SWP_NOZORDER );

            event.controlHandle().postEvent( IC_UM_TCM_SIZE,
                                             event.parameter1(),
                                             event.parameter2() );
          }
          else
          {
            pNotebk->tabControlResize( newSize );

            stopProcessingEvent = true;
          }
        }
        else
        {
          pNotebk->tabControlResize( newSize );
        }
      } //WM_SIZE & IC_UM_TCM_SIZE
      break;

    case WM_SYSKEYDOWN:
      {
        /**********************************************************************/
        /* Only process Windows tab control keyboard events of interest.      */
        /**********************************************************************/
        IWindowHandle hwndControl( event.controlHandle() );
        if ( !(hwndControl &&
             ( IWindowClassName( hwndControl ) == WC_TABCONTROL )) )
          break;

        /**********************************************************************/
        /* Process key events for the Windows tab control.                    */
        /**********************************************************************/
        IKeyboardEvent keyevt( event );
        if ( !keyevt.isVirtual() || !keyevt.isAltDown() ||
             keyevt.isShiftDown() || keyevt.isCtrlDown() )
        {
          break;
        }

        /**********************************************************************/
        /* Obtain pointer to notebook object.                                 */
        /**********************************************************************/
        INotebook*
          pNotebk = (INotebook*)IWindow::windowWithHandle( hwndControl );

        /**********************************************************************/
        /* Alt+Down:  set the focus to the application's page window          */
        /**********************************************************************/
        if ( keyevt.virtualKey() == IKeyboardEvent::down )
        {
          /********************************************************************/
          /* Get the top page handle if it exists.                            */
          /********************************************************************/
          IPageHandle topPageHandle = pNotebk->topPage();
          if (topPageHandle)
          {
            /******************************************************************/
            /* Get the application page window if it exists and set the       */
            /* focus to it.                                                   */
            /******************************************************************/
            IWindow* pageWindow = pNotebk->window( topPageHandle );
            if (pageWindow)
            {
              pageWindow->setFocus();
              stopProcessingEvent = true;
            }
          }
        }
        /**********************************************************************/
        /* Alt+PageUp:  bring the previous page to the top                    */
        /**********************************************************************/
        else if ( keyevt.virtualKey() == IKeyboardEvent::pageUp )
        {
          IPageHandle topPageHandle = pNotebk->topPage();
          if (topPageHandle)
          {
            IPageHandle prevPageHandle = pNotebk->previousPage( topPageHandle );
            if (prevPageHandle)
            {
              pNotebk->turnToPage( prevPageHandle );
              stopProcessingEvent = true;
            }
          }
        }
        /**********************************************************************/
        /* Alt+PageDown:  bring the next page to the top                      */
        /**********************************************************************/
        else if ( keyevt.virtualKey() == IKeyboardEvent::pageDown )
        {
          IPageHandle topPageHandle = pNotebk->topPage();
          if (topPageHandle)
          {
            IPageHandle nextPageHandle = pNotebk->nextPage( topPageHandle );
            if (nextPageHandle)
            {
              pNotebk->turnToPage( nextPageHandle );
              stopProcessingEvent = true;
            }
          }
        }
      } //WM_SYSKEYDOWN
      break;

    case WM_NOTIFY:
      {
        /**********************************************************************/
        /* Process the Windows tab ctrl's notifications for selection events  */
        /**********************************************************************/
        LPNMHDR lpnmhdr = (LPNMHDR)event.parameter2().asUnsignedLong();

        IWindowHandle hwndControl( lpnmhdr->hwndFrom );

        /**********************************************************************/
        /* Verify that we are processing a Windows tab control                */
        /**********************************************************************/
        if ( hwndControl  &&
             (IWindowClassName( hwndControl ) == WC_TABCONTROL) )
        {
          if ( (lpnmhdr->code == TCN_SELCHANGE) ||
               (lpnmhdr->code == TCN_SELCHANGING) )
          {
            INotebook*
              pNotebk = (INotebook*)IWindow::windowWithHandle( hwndControl );

            /******************************************************************/
            /* Fill in the PAGESELECTNOTIFY structure that we use in Windows  */
            /******************************************************************/
            PAGESELECTNOTIFY pageSelectNotify;
            pageSelectNotify.hwndBook = hwndControl;

            if ( lpnmhdr->code == TCN_SELCHANGE )
            {
              /****************************************************************/
              /* Use value for previously selected page that we stored during */
              /* processing of TCN_SELCHANGING.                               */
              /****************************************************************/
              pageSelectNotify.ulPageIdNew =
                   (DWORD)pNotebk->topPage().asUnsigned();
              pageSelectNotify.ulPageIdCur = (DWORD)ulCurrentPageId;

              /****************************************************************/
              /* The tab selection has already occurred and the new tab is    */
              /* now the top page in the tab control.  Do our required        */
              /* processing and issue the selection notification.             */
              /****************************************************************/
              stopProcessingEvent =
                pNotebk->processTabSelect( &pageSelectNotify, false );
            }
            else
            {
              /****************************************************************/
              /* The tab selection for a new tab is pending.                  */
              /* ------------------------------------------------------------ */
              /* We have no easy way of determining which new tab is selected.*/
              /* Therefore, we set both values to the top tab page.           */
              /****************************************************************/
              pageSelectNotify.ulPageIdNew =
                   (DWORD)pNotebk->topPage().asUnsigned();
              pageSelectNotify.ulPageIdCur = pageSelectNotify.ulPageIdNew;
              ulCurrentPageId = pageSelectNotify.ulPageIdCur;

              /****************************************************************/
              /* The tab selection is now in progress (pending).  Do our      */
              /* required processing and issue the selection pending          */
              /* notification.                                                */
              /****************************************************************/
              if ( pNotebk->processTabSelect( &pageSelectNotify, true ) )
                event.setResult( true );

              stopProcessingEvent = true;
            }
          }
        }
        break;
      } //WM_NOTIFY

    case WM_SYSCOMMAND:
      {
        /**********************************************************************/
        /* Process the Windows tab control's WM_SYSCOMMAND/SC_CLOSE event     */
        /* for Windows95 to work around a Windows95 bug.                      */
        /**********************************************************************/
        IWindowHandle hwndControl( event.controlHandle() );
        if ( !(hwndControl &&
             ( IWindowClassName( hwndControl ) == WC_TABCONTROL )  &&
               IPlatform::isWin95()) )
        {
          break;
        }

        if ( (((unsigned long)event.parameter1()) & 0xFFF0) == SC_CLOSE )
        {
          INotebook*
            pNotebk = (INotebook*)IWindow::windowWithHandle( hwndControl );

          /********************************************************************/
          /* Return true if a tab control to circumvent Windows95 bug.        */
          /********************************************************************/
          if ( pNotebk && !pNotebk->isPMCompatible() )
          {
            return( true );
          }
        }
      } //WM_SYSCOMMAND
      break;
#endif //IC_WIN

    /******************************************************************/
    /* Store the colors that are being set and then let this message  */
    /* pass on to the notebook control.                               */
    /******************************************************************/
    case BKM_SETNOTEBOOKCOLORS:
      ((INotebook*)event.window())
          ->storeNotebookColors( event.parameter1(), event.parameter2() );
      break;

    default:
      break;
  } //end switch

  return stopProcessingEvent;
}


/*------------------------------------------------------------------------------
| INotebookHandler::INotebookHandler                                           |
------------------------------------------------------------------------------*/
INotebookHandler :: INotebookHandler ( )
{
}

/*------------------------------------------------------------------------------
| INotebookHandler::~INotebookHandler                                          |
------------------------------------------------------------------------------*/
INotebookHandler :: ~INotebookHandler ( )
{
}
