/*******************************************************************************
* FILE NAME: iprofile.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Implementation of the class(es):                                           *
*     IProfile                                                                 *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C" {
  #define INCL_WINSHELLDATA
  #define INCL_SHLERRORS
  #include <iwindefs.h>
}

#ifdef IC_PMWIN
  #include <stdio.h>
#endif

#include <iprofile.hpp>
#include <iprofprv.hpp>
#include <iexcept.hpp>
#include <iplatfrm.hpp>
#include <istring.hpp>
#include <istparse.hpp>
#include <ithread.hpp>
#include <itrace.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IPROFILE_CPP_
  #include <ipagetun.h>
#endif

#ifdef IC_PMWIN
// New static (not thread safe) for determining when to translate
// profile, application and key names and when not to.  Default is
// true to remain compatible with GA release.
static Boolean fgMapKey = true;
#endif
#ifdef IC_WIN
/*------------------------------------------------------------------------------
| Modify a key name to meet Win32 registry requirements                        |
------------------------------------------------------------------------------*/
static IString IRegFixKey( const char* name )
{
  IString keyName(name);

  // Convert invalid key name characters for compatibility with OS/2
  // Also converts hierarchical characters to flatten tree.
  if (IProfile::isKeyMappingEnabled())
  {
    // Convert slashes first, to allow us to differentiate later
    // between an existing combo and one we created.
    keyName.change("/","//");

    // Convert illegal characters for Win32 Registry.
    keyName.change("*","/1");
    keyName.change("?","/0");

    // Convert backslash to flatten hierarchy.
    keyName.change("\\","/_");

    // On Win32s, spaces are illegal so convert them.
    if (IPlatform::isWin32S())
      keyName.change(" ","/-");
  }
  return keyName;
}

/*------------------------------------------------------------------------------
| Check if a Win32 registry contains the specified key                         |
------------------------------------------------------------------------------*/
static Boolean IRegTestKey( HKEY parentKey, const char* name )
{
  HKEY          openKey;
  IString       keyName = IRegFixKey( name );
  unsigned long rc;
  if (IPlatform::isWin32S())
    rc = RegOpenKey ( parentKey, keyName, &openKey );
  else
    rc = RegOpenKeyEx ( parentKey, keyName, 0, KEY_READ, &openKey );

  if ( rc == ERROR_SUCCESS )
  {
    RegCloseKey(openKey);
    return true;
  }
  return false;
}

/*------------------------------------------------------------------------------
| Open a key in the Win32 registry                                             |
------------------------------------------------------------------------------*/
static HKEY IRegOpenKey( HKEY parentKey, const char* name )
{
  HKEY    openKey;
  IString keyName = IRegFixKey( name );
  if (IPlatform::isWin32S())
  {
    if ( RegCreateKey ( parentKey, keyName, &openKey ) != ERROR_SUCCESS )
      ITHROWGUIERROR("RegCreateKey");
  }
  else
  {
    unsigned long disposition;
    if ( RegCreateKeyEx ( parentKey, keyName, 0, "",
                          REG_OPTION_NON_VOLATILE, (KEY_READ | KEY_WRITE),
                          NULL, &openKey, &disposition ) != ERROR_SUCCESS )
    {
      ITHROWGUIERROR("RegCreateKeyEx");
    }
  }
  return openKey;
}

/*------------------------------------------------------------------------------
| Delete a key from the Win32 registry                                         |
------------------------------------------------------------------------------*/
static void IRegDeleteKey ( HKEY parentKey, const char* name )
{
  IString keyName = IRegFixKey( name );
  RegDeleteKey( parentKey, keyName );
}

/*------------------------------------------------------------------------------
| Open the profile name passed, returning the key for it.  Also set flag       |
| to denote whether this is a system profile or not                            |
------------------------------------------------------------------------------*/
static HKEY IRegOpenProfile( const char *profName, Boolean* isSystemProfile )
{
  HKEY openKey, parentKey = 0;
  IString token1, token2;
  IString token3 = profName;

  token3 >> token1 >> "\\" >> token3;

  if (token1 == "HKEY_CURRENT_USER")
    parentKey = HKEY_CURRENT_USER;

  if (token1 == "HKEY_LOCAL_MACHINE")
    parentKey = HKEY_LOCAL_MACHINE;

  if (token1 == "HKEY_USERS")
    parentKey = HKEY_USERS;

  if (token1 == "HKEY_CLASSES_ROOT")
    parentKey = HKEY_CLASSES_ROOT;

  // If the first token matches a system defined keyword and
  // has additional key information as well
  if ((parentKey) && (token3 != ""))
  {
    token3 >> token2 >> "\\" >> token3;

    // If only one subkey provided, open that subkey
    if (token3 == "")
    {
      if (token2 == "Software")
        (*isSystemProfile) = TRUE;
      openKey = IRegOpenKey( parentKey, token2 );
    }
    else
    {
      // Otherwise open the parent, then the subkey, and close parent
      HKEY tempKey = IRegOpenKey( parentKey, token2 );
      openKey = IRegOpenKey( tempKey, token3 );
      RegCloseKey( tempKey );
    }
  }
  else
  {
    // If not system profile, then open app profile under local area.
    HKEY tempKey = IRegOpenKey( HKEY_LOCAL_MACHINE, "Software" );
    openKey = IRegOpenKey( tempKey, profName );
    RegCloseKey( tempKey );
  }
  return openKey;
}

/*------------------------------------------------------------------------------
| Get the data type for a key from the Win32 registry                          |
------------------------------------------------------------------------------*/
static unsigned long IRegDataType ( HKEY openKey )
{
  char          buffer[10];
  unsigned long rc;
  unsigned long dataType = REG_NONE;
  long          bufferSize = 10;

  // Query the operating system for the type of data stored at key
  if (IPlatform::isWin32S())
    rc = RegQueryValue( openKey, "DataType", buffer, &bufferSize );
  else
    rc = RegQueryValueEx( openKey, "DataType", 0, 0,
                          (LPBYTE) buffer, (LPDWORD) &bufferSize );

  // If successful, then set data type flag for future operations
  if ( rc == ERROR_SUCCESS )
  {
     buffer[bufferSize] = '\0';
     IString dType(buffer);
     if ( dType == "Binary" )
       dataType = REG_BINARY;
     if ( dType == "Integer" )
       dataType = REG_DWORD;
     if ( dType == "String" )
       dataType = REG_SZ;
  }
  return dataType;
}

/*------------------------------------------------------------------------------
| Get the string for a key from the Win32 registry                             |
------------------------------------------------------------------------------*/
static IString IRegGetString ( HKEY openKey )
{
  IString       str;
  char*         buffer;
  long          bufferSize;
  unsigned long rc;

  // First get the size of the buffer needed to retrieve the string.
  if (IPlatform::isWin32S())
    rc = RegQueryValue( openKey, "Data", 0, &bufferSize );
  else
    rc = RegQueryValueEx( openKey, "Data", 0, 0, 0, (LPDWORD)&bufferSize );

  // If it fails, throw an exception
  if ( rc != ERROR_SUCCESS )
    ITHROWGUIERROR("RegQueryValueEx");

  // Next, allocate a buffer and retrieve the string data
  buffer = new char[bufferSize+1];
  if (IPlatform::isWin32S())
    rc = RegQueryValue( openKey, "Data", buffer, &bufferSize );
  else
    rc = RegQueryValueEx( openKey, "Data", 0, 0, (LPBYTE)buffer,
                          (LPDWORD)&bufferSize );

  // Finally terminate the buffer, create a string object,
  // free the character buffer and return the string.
  buffer[bufferSize] = '\0';
  str = IString(buffer);
  delete [] buffer;
  return str;
}

/*------------------------------------------------------------------------------
| Get the integer for a key from the Win32 registry                            |
------------------------------------------------------------------------------*/
static long IRegGetInteger ( HKEY openKey )
{
  long          value;
  long          bufferSize;
  if (IPlatform::isWin32S())
  {
    // If the platform is Win32s, then all data is stored as a string
    // so first, get the length of the string data
    char *buffer;
    if ( RegQueryValue( openKey, "Data", 0, &bufferSize ) != ERROR_SUCCESS )
      ITHROWGUIERROR("RegQueryValue");

    // If key valid, then allocate buffer for receiving character data
    // and request string data
    buffer = new char[bufferSize+1];
    RegQueryValue( openKey, "Data", buffer, &bufferSize );

    // Terminate buffer, create a string object and convert to integer,
    // then free the character buffer
    buffer[bufferSize] = '\0';
    value = IString(buffer).asInt();
    delete [] buffer;
  }
  else
  {
    // If regular 32-bit platform, then data can be stored as an integer
    // so query the value directly
    bufferSize = sizeof(value);
    if ( RegQueryValueEx( openKey, "Data", 0, 0, (LPBYTE)&value,
                          (LPDWORD)&bufferSize ) != ERROR_SUCCESS )
    {
      ITHROWGUIERROR("RegQueryValueEx");
    }
  }
  return value;
}

/*------------------------------------------------------------------------------
| Get the binary data for a key from the Win32 registry                        |
------------------------------------------------------------------------------*/
static IString IRegGetBinary ( HKEY openKey )
{
  IString       str;
  unsigned long cnt;
  unsigned long rc = ERROR_SUCCESS;
  long          bufferSize;
  for ( cnt = 0; rc == ERROR_SUCCESS; cnt++ )
  {
    IString valueName(cnt);
    if (IPlatform::isWin32S())
      rc = RegQueryValue( openKey, valueName, 0, &bufferSize );
    else
      rc = RegQueryValueEx( openKey, valueName, 0, 0, 0,
                            (LPDWORD)&bufferSize );

    if ( rc == ERROR_SUCCESS )
    {
      char* buffer = new char[bufferSize+1];
      if (IPlatform::isWin32S())
      {
        RegQueryValue( openKey, valueName, buffer, &bufferSize );
        buffer[bufferSize] = '\0';
        IString tmp(buffer);
        tmp.x2c();
        str += tmp;
      }
      else
      {
        rc = RegQueryValueEx( openKey, valueName, 0, 0,
                              (LPBYTE)buffer, (LPDWORD)&bufferSize );
        IString tmp(buffer,(UINT)bufferSize);
        str += tmp;
      }
      delete [] buffer;
    }
  }
  return str;
}

/*------------------------------------------------------------------------------
| Write a string in the Win32 registry                                         |
------------------------------------------------------------------------------*/
static void IRegWriteString ( HKEY openKey, IString data )
{
  long          bufferSize;
  unsigned long rc;
  if (IPlatform::isWin32S())
  {
    bufferSize = 6;
    rc = RegSetValue( openKey, "DataType", REG_SZ, "String",
                      bufferSize );
    if ( rc != ERROR_SUCCESS )
      ITHROWGUIERROR("RegSetValue");
    bufferSize = data.length();
    rc = RegSetValue( openKey, "Data", REG_SZ, data, bufferSize );
    if ( rc != ERROR_SUCCESS )
      ITHROWGUIERROR("RegSetValue");
  }
  else
  {
    bufferSize = 6;
    rc = RegSetValueEx( openKey, "DataType", 0, REG_SZ,
                        (const BYTE*)"String", bufferSize );
    if ( rc != ERROR_SUCCESS )
      ITHROWGUIERROR("RegSetValueEx");
    bufferSize = data.length();
    rc = RegSetValueEx( openKey, "Data", 0, REG_SZ, data, bufferSize );
    if ( rc != ERROR_SUCCESS )
      ITHROWGUIERROR("RegSetValueEx");
  }
}

/*------------------------------------------------------------------------------
| Get the integer for a key from the Win32 registry                            |
------------------------------------------------------------------------------*/
static void IRegWriteInteger ( HKEY openKey, long data )
{
  long          bufferSize;
  unsigned long rc;
  if (IPlatform::isWin32S())
  {
    bufferSize = 7;
    rc = RegSetValue( openKey, "DataType", REG_SZ, "Integer",
                      bufferSize );
    if ( rc != ERROR_SUCCESS )
      ITHROWGUIERROR("RegSetValue");
    IString buffer(data);
    bufferSize = buffer.length();
    rc = RegSetValue( openKey, "Data", REG_SZ, buffer, bufferSize );
    if ( rc != ERROR_SUCCESS )
      ITHROWGUIERROR("RegSetValue");
  }
  else
  {
    bufferSize = 7;
    rc = RegSetValueEx ( openKey, "DataType", 0, REG_SZ,
                         (const BYTE*) "Integer", bufferSize );
    if ( rc != ERROR_SUCCESS )
      ITHROWGUIERROR("RegSetValueEx");
    bufferSize = sizeof(data);
    rc = RegSetValueEx( openKey, "Data", 0, REG_DWORD,
                        (const BYTE*) &data, bufferSize );
    if ( rc != ERROR_SUCCESS )
      ITHROWGUIERROR("RegSetValueEx");
  }
}

/*------------------------------------------------------------------------------
| Get the binary data for a key from the Win32 registry                        |
------------------------------------------------------------------------------*/
static void IRegWriteBinary ( HKEY openKey, IString data )
{
  long          bufferSize;
  unsigned long rc;
  unsigned long cnt;
  unsigned long written = 0;
  if (IPlatform::isWin32S())
  {
    bufferSize = 6;
    rc = RegSetValue( openKey, "DataType", REG_SZ, "Binary",
                      bufferSize );
    if ( rc != ERROR_SUCCESS )
      ITHROWGUIERROR("RegSetValue");
  }
  else
  {
    bufferSize = 6;
    rc = RegSetValueEx( openKey, "DataType", 0, REG_SZ,
                        (const BYTE*)"Binary", bufferSize );
    if ( rc != ERROR_SUCCESS )
      ITHROWGUIERROR("RegSetValueEx");
  }
  for ( cnt = 0; written < data.length(); cnt++ )
  {
    IString tmp;
    IString valueName(cnt);
    if (IPlatform::isWin32S())
    {
      if (( data.length() - written ) > 3500 )
        tmp = data.subString( (UINT)written+1,3500);
      else
        tmp = data.subString((UINT)written+1);
      written += tmp.length();
      tmp.c2x();
      bufferSize = tmp.length();
      rc = RegSetValue( openKey, valueName, REG_SZ, tmp, bufferSize );
      if ( rc != ERROR_SUCCESS )
        ITHROWGUIERROR("RegSetValue");
    }
    else
    {
      if (( data.length() - written ) > 16000 )
        tmp = data.subString((UINT)written+1,16000);
      else
        tmp = data.subString((UINT)written+1);
      written += tmp.length();
      bufferSize = tmp.length();
      rc = RegSetValueEx( openKey, valueName, 0, REG_BINARY, tmp,
                           bufferSize );
      if ( rc != ERROR_SUCCESS )
        ITHROWGUIERROR("RegSetValueEx");
    }
  }
}
#endif

/*------------------------------------------------------------------------------
| IProfile::IProfile                                                           |
------------------------------------------------------------------------------*/
IProfile::IProfile ( const char *profName )
  : profileName(profName), defApplName(), fProfileData(new IProfileData())
{
  IMODTRACE_DEVELOP("IProfile::ctor (char *)");
  IASSERTPARM(profName != 0);

#ifndef IC_WIN
  //Note:  Create a temporary IString for the profile name to
  //       workaround a PMSHAPI bug that attempts to write to
  //       read-only storage
  hIni = PrfOpenProfile( IThread::current().anchorBlock(),
                         IString( profName ) );
  if (!hIni)
    ITHROWGUIERROR("PrfOpenProfile");
#endif
#ifdef IC_WIN
  fProfileData->hKey = IRegOpenProfile( profName,
                                        &(fProfileData->isSystemProfile) );
#endif
}


/*------------------------------------------------------------------------------
| IProfile::IProfile                                                           |
| Copy constructor                                                             |
------------------------------------------------------------------------------*/
IProfile::IProfile ( const IProfile& aProfile )
  : fProfileData(new IProfileData())
{
#ifndef IC_WIN
  if ((aProfile.hIni != HINI_PROFILE) &&
      (aProfile.hIni != HINI_USERPROFILE) &&
      (aProfile.hIni != HINI_SYSTEMPROFILE))
  {
     hIni = PrfOpenProfile(IThread::current().anchorBlock(),
                            (PSZ)aProfile.profileName);
     if (!hIni)
        ITHROWGUIERROR("PrfOpenProfile");
  }
  else
     this->hIni = aProfile.hIni;
#endif

  this->defApplName = aProfile.defApplName;
  this->profileName = aProfile.profileName;

#ifdef IC_WIN
  this->fProfileData->hKey =
    IRegOpenProfile( this->profileName, &(this->fProfileData->isSystemProfile));
#endif
}

#ifndef IC_WIN
/*------------------------------------------------------------------------------
| IProfile::IProfile                                                           |
|  Private Ctor. to create from Profile Handle                                 |
------------------------------------------------------------------------------*/
IProfile::IProfile ( const IProfileHandle& profile )
  : hIni(profile), defApplName(), fProfileData(new IProfileData())
{
  HAB hab = WinQueryAnchorBlock(HWND_DESKTOP);
  PRFPROFILE Profile;
  Profile.cchUserName = 0;
  Profile.pszUserName = 0;
  Profile.cchSysName = 0;
  Profile.pszSysName = 0;
  char Buffer[100];

  if ( PrfQueryProfile( hab, &Profile ))
  {
    if (hIni == HINI_USERPROFILE)
    {
      char* pszData = new char[Profile.cchUserName+1];
      Profile.pszUserName = (PSZ)pszData;
      Profile.cchSysName = 100;
      Profile.pszSysName = (PSZ)Buffer;
      PrfQueryProfile(hab, &Profile);
      profileName = IString(pszData);
    }
    else
    {
      char* pszData = new char[Profile.cchSysName];
      Profile.cchUserName = 100;
      Profile.pszUserName = (PSZ)Buffer;
      Profile.pszSysName = (PSZ)pszData;
      PrfQueryProfile(hab, &Profile);
      profileName = IString(pszData);
    } /* endif */
  } /* endif */
}
#endif  // not IC_WIN

/*------------------------------------------------------------------------------
| IProfile::systemProfile                                                      |
------------------------------------------------------------------------------*/
IProfile IProfile::systemProfile ( )
{
  IMODTRACE_DEVELOP("IProfile::systemProfile");

#ifndef IC_WIN
  return IProfile(HINI_SYSTEMPROFILE);
#endif
#ifdef IC_WIN
  return IProfile("HKEY_LOCAL_MACHINE\\Software");
#endif
}

/*------------------------------------------------------------------------------
| IProfile::userProfile                                                        |
------------------------------------------------------------------------------*/
IProfile IProfile::userProfile ( )
{
  IMODTRACE_DEVELOP("IProfile::userProfie");

#ifndef IC_WIN
  return IProfile(HINI_USERPROFILE);
#endif
#ifdef IC_WIN
  return IProfile("HKEY_CURRENT_USER\\Software");
#endif
}

/*------------------------------------------------------------------------------
| IProfile::operator =                                                         |
------------------------------------------------------------------------------*/
IProfile& IProfile::operator =(const IProfile& aProfile)
{
#ifndef IC_WIN
  /* Note: Ensure self assignment works */
  if (this->hIni != aProfile.hIni)
  {
     if ((hIni != HINI_PROFILE) &&
         (hIni != HINI_USERPROFILE) &&
         (hIni != HINI_SYSTEMPROFILE))
     {
        // close file if not system & user
        PrfCloseProfile(hIni);

        hIni = PrfOpenProfile(IThread::current().anchorBlock(),
                              (PSZ)aProfile.profileName);
        if (!hIni)
           ITHROWGUIERROR("PrfOpenProfile");
     }
     else
        this->hIni = aProfile.hIni;

     this->defApplName = aProfile.defApplName;
     this->profileName = aProfile.profileName;
  } /* endif */
#endif

#ifdef IC_WIN
  this->defApplName = aProfile.defApplName;
  this->profileName = aProfile.profileName;
  this->fProfileData->hKey =
    IRegOpenProfile( this->profileName, &(this->fProfileData->isSystemProfile));
#endif
  return *this;
}


/*------------------------------------------------------------------------------
| IProfile::name                                                               |
| Return profile name                                                          |
------------------------------------------------------------------------------*/
IString IProfile::name ( ) const
{
  return profileName;
}

#ifndef IC_WIN
/*------------------------------------------------------------------------------
| IProfile::handle                                                             |
| Return the profile handle                                                    |
------------------------------------------------------------------------------*/
IProfileHandle IProfile::handle() const
{
  return hIni;
}
#endif  // not IC_WIN

/*------------------------------------------------------------------------------
| IProfile::defaultApplicationName                                             |
| Return default application name                                              |
------------------------------------------------------------------------------*/
IString IProfile::defaultApplicationName ( ) const
{
  return defApplName;
}

/*------------------------------------------------------------------------------
| IProfile::setDefaultApplicationName                                          |
| Sets the default application name                                            |
------------------------------------------------------------------------------*/
IProfile& IProfile::setDefaultApplicationName ( const char *applName )
{
  defApplName = IString(applName);
  return *this;
}

/*------------------------------------------------------------------------------
| IProfile::~IProfile                                                          |
------------------------------------------------------------------------------*/
IProfile::~IProfile ( )
{
#ifndef IC_WIN
  if ((hIni != HINI_PROFILE) &&
      (hIni != HINI_USERPROFILE) &&
      (hIni != HINI_SYSTEMPROFILE))
  {
    PrfCloseProfile(hIni);           // close file if not system & user
  } /* endif */
#endif

#ifdef IC_WIN
  RegCloseKey(fProfileData->hKey);
#endif

  delete fProfileData;
}

/*------------------------------------------------------------------------------
| IProfile::elementWithKey                                                     |
------------------------------------------------------------------------------*/
IString IProfile::elementWithKey ( const char *key, const char *applName ) const
{
  IMODTRACE_DEVELOP("IProfile::elementWithKey");
  IASSERTPARM(key != 0);

  if (!applName)
    IASSERTPARM(defApplName.length() > 0);

#ifndef IC_WIN
  unsigned long ulLen = length(key, applName);

  if (ulLen > 0)
  {
    char* buffer = new char[ulLen + 1];
    /**************************/
    /* Query string first.    */
    /**************************/
    IGUIErrorInfo clear("Clear Last Error");
    PrfQueryProfileString( hIni,
           (applName) ? (PSZ)applName : (PSZ)defApplName,
           (PSZ)key, (PSZ)"", (PSZ)buffer, ulLen + 1);

    IGUIErrorInfo err("PrfQueryProfileString");
    if (err.errorId() != 0)
    {
      if (err.errorId() == PMERR_INVALID_ASCIIZ )
      {
        /**************************/
        /* Query binary data.     */
        /**************************/
        if (!PrfQueryProfileData( hIni,
               (applName) ? (PSZ)applName : (PSZ)defApplName,
               (PSZ)key, buffer, &ulLen + 1) )
        {
          ITHROWGUIERROR("PrfQueryProfileData");
        }
      }
      else
      {
        IInvalidRequest ex(err.text(), err.errorId(), unrecoverable);
        ITHROW(ex);
      }
    }

    IString str(buffer, ulLen);
    delete [] buffer;
    return str;

  } /* endif */
  return IString();
#endif

#ifdef IC_WIN
  IString str;

  IString appName(applName);
  if (!applName)
    appName = defApplName;

  // Make sure the application name is ok
  if (!IRegTestKey( fProfileData->hKey, appName ))
  {
    ITHROWGUIERROR("RegOpenKey");
  }

  HKEY appKey = IRegOpenKey( fProfileData->hKey, appName );

  // Make sure the key name is ok
  if (!IRegTestKey( appKey, key ))
  {
    RegCloseKey(appKey);
    ITHROWGUIERROR("RegOpenKey");
  }

  HKEY dataKey = IRegOpenKey(appKey,key);

  // Close the application key
  RegCloseKey(appKey);

  // Get the value type
  unsigned long dataType = IRegDataType(dataKey);

  // Check if data is not in supported format
  if ( dataType == REG_NONE )
  {
    RegCloseKey(dataKey);
    ITHROWGUIERROR("RegQueryKey");
  }

  // Get data
  switch ( dataType )
  {
    case REG_SZ:
      str = IRegGetString(dataKey);
      break;
    case REG_DWORD:
    {
      long tmp = IRegGetInteger(dataKey);
      str = IString((char*)&tmp,sizeof(tmp));
      break;
    }
    case REG_BINARY:
      str = IRegGetBinary(dataKey);
      break;
  }

  // Close the key and return
  RegCloseKey(dataKey);
  return str;
#endif
}


/*------------------------------------------------------------------------------
| IProfile::integerWithKey                                                     |
------------------------------------------------------------------------------*/
long IProfile::integerWithKey ( const char *key, const char *applName ) const
{
  IMODTRACE_DEVELOP("IProfile::integerWithKey");
  IASSERTPARM( key != 0 );
  if (!applName)
    IASSERTPARM( defApplName.length() > 0 );

  long lResult;

#ifndef IC_WIN
  unsigned long resultSize = sizeof(lResult);
  if (!PrfQueryProfileData( hIni,
             (applName) ? (PSZ)applName : (PSZ)defApplName,
             (PSZ)key, (PVOID)&lResult, &resultSize))
  {
    ITHROWGUIERROR("PrfQueryProfileData");
  }
#endif

#ifdef IC_WIN
  IString appName(applName);
  if (!applName)
    appName = defApplName;

  // Make sure the application name is ok
  if (!IRegTestKey( fProfileData->hKey, appName ))
  {
    ITHROWGUIERROR("RegOpenKey");
  }

  HKEY appKey = IRegOpenKey( fProfileData->hKey, appName );

  // Make sure the key name is ok
  if (!IRegTestKey( appKey, key ))
  {
    RegCloseKey(appKey);
    ITHROWGUIERROR("RegOpenKey");
  }

  HKEY dataKey = IRegOpenKey(appKey,key);

  // Close the application key
  RegCloseKey(appKey);

  // Get the value type
  unsigned long dataType = IRegDataType(dataKey);

  // Check if data is not in supported format
  if ( dataType != REG_DWORD )
  {
    RegCloseKey(dataKey);
    ITHROWGUIERROR("RegQueryKey");
  }

  // Get data
  lResult = IRegGetInteger(dataKey);

  // Close the key
  RegCloseKey(dataKey);
#endif

  return lResult;
}


/*------------------------------------------------------------------------------
| IProfile::addOrReplaceElementWithKey                                         |
|   Write string data to file.                                                 |
------------------------------------------------------------------------------*/
IProfile& IProfile::addOrReplaceElementWithKey( const char *key,
                                                const IString& data,
                                                const char *applName)
{
  IMODTRACE_DEVELOP("IProfile::addOrReplaceElementWithKey");
  IASSERTPARM(key != 0);
  if (!applName) {
     IASSERTPARM(defApplName.length() > 0);
  } /* endif */

#ifndef IC_WIN
  if (!PrfWriteProfileData(hIni,
                           (applName) ? (PSZ)applName : (PSZ)defApplName,
                           (PSZ)key, (PSZ)data, data.length() )) {
     ITHROWGUIERROR("PrfWriteProfileString");
  } /* endif */
#endif

#ifdef IC_WIN
  IString appName(applName);
  if (!applName)
    appName = defApplName;

  // Open the application key
  HKEY appKey = IRegOpenKey( fProfileData->hKey, appName );

  // If the value already has data then delete it
  if (IRegTestKey( appKey, key ))
  {
    RegCloseKey(appKey);
    deleteElementWithKey( key, appName );
    appKey = IRegOpenKey( fProfileData->hKey, appName );
  }

  HKEY dataKey = IRegOpenKey(appKey,key);
  RegCloseKey(appKey);

  // Set the key value
  if (data.isPrintable())
  {
    IRegWriteString(dataKey,data);
  }
  else
  {
    IRegWriteBinary(dataKey,data);
  }

  RegCloseKey(dataKey);
#endif

  return *this;

}

/*------------------------------------------------------------------------------
| IProfile::addOrReplaceElementWithKey                                         |
|   Write binary data to file.                                                 |
------------------------------------------------------------------------------*/
IProfile& IProfile::addOrReplaceElementWithKey
                            ( const char *key, long data, const char *applName)
{
  IMODTRACE_DEVELOP("IProfile::addOrReplaceElementWithKey");
  IASSERTPARM( key != 0 );
  if (!applName)
    IASSERTPARM( defApplName.length() > 0 );

#ifndef IC_WIN
  if (!PrfWriteProfileData( hIni, (applName) ? (PSZ)applName :
          (PSZ)defApplName, (PSZ)key, (void*)&data, sizeof(data)))
  {
    ITHROWGUIERROR("PrfWriteProfileData");
  } /* endif */
#endif

#ifdef IC_WIN
  IString appName(applName);
  if (!applName)
    appName = defApplName;

  // Open the application key
  HKEY appKey = IRegOpenKey( fProfileData->hKey, appName );

  // If the value already has data then delete it
  if (IRegTestKey( appKey, key ))
  {
    RegCloseKey(appKey);
    deleteElementWithKey( key, appName );
    appKey = IRegOpenKey( fProfileData->hKey, appName );
  }

  HKEY dataKey = IRegOpenKey(appKey,key);
  RegCloseKey(appKey);

  // Set the key value
  IRegWriteInteger( dataKey, data );

  RegCloseKey(dataKey);
#endif

  return *this;
}

/*------------------------------------------------------------------------------
| IProfile::deleteElementWithKey                                               |
------------------------------------------------------------------------------*/
IProfile& IProfile::deleteElementWithKey ( const char *key,
                                           const char *applName)
{
  IMODTRACE_DEVELOP("IProfile::deleteElementWithKey");
  IASSERTPARM(key != 0);
  if (!applName) {
     IASSERTPARM(defApplName.length() > 0);
  } /* endif */

#ifndef IC_WIN
  if (!PrfWriteProfileString( hIni,
       (applName) ? (PSZ)applName : (PSZ)defApplName, (PSZ)key, 0) )
    ITHROWGUIERROR("PrfWriteProfileString");
#endif

#ifdef IC_WIN
  IString appName(applName);
  if (!applName)
    appName = defApplName;

  // Make sure the application name is ok
  if (!IRegTestKey( fProfileData->hKey, appName ))
    ITHROWGUIERROR("RegOpenKey");

  // Open the application
  HKEY appKey = IRegOpenKey( fProfileData->hKey, appName );

  // Delete the key value
  IRegDeleteKey( appKey, key );

  // Check if there is another subkey
  char          buffer[MAX_PATH] = "";
  unsigned long keys = 0;
  if ( RegEnumKey ( appKey, 0, buffer, MAX_PATH ) == ERROR_SUCCESS )
    keys = 1;

  // Close open application key
  RegCloseKey( appKey );

  // If no other data then delete the application
  if ( !keys )
    IRegDeleteKey( fProfileData->hKey, appName );

#endif
  return *this;
}


/*------------------------------------------------------------------------------
| IProfile::deleteElementWithApplication                                       |
------------------------------------------------------------------------------*/
IProfile& IProfile::deleteElementWithApplication ( const char *applName)
{
  IMODTRACE_DEVELOP("IProfile::deleteElementWithApplication");
  if (!applName) {
     IASSERTPARM(defApplName.length() > 0);
  } /* endif */

#ifndef IC_WIN
  if (!PrfWriteProfileString( hIni,
                 (applName) ? (PSZ)applName : (PSZ)defApplName, 0, 0 ))
  {
    ITHROWGUIERROR("PrfWriteProfileString");
  } /* endif */
#endif

#ifdef IC_WIN
  IProfile::Cursor cursor(*this,applName);
  for ( cursor.setToFirst(); cursor.isValid(); cursor.setToNext())
  {
    IString subkey = applicationOrKeyAt(cursor);
    deleteElementWithKey(subkey,applName);
  }
#endif
  return *this;
}

#ifdef IC_PMWIN
/*------------------------------------------------------------------------------
| IProfile::deleteProfile                                                      |
------------------------------------------------------------------------------*/
void IProfile::deleteProfile ( const char* profileName )
{
  // Do not allow the user or system profiles to be removed
  if (( IProfile::systemProfile().name() == profileName ) ||
      ( IProfile::userProfile().name()   == profileName ))
  {
    return;
  }

#ifdef IC_WIN
  // First delete any applications
  {
    IString subKey;
    IProfile tmpProfile( profileName );
    IProfile::Cursor cursor(tmpProfile);
    for ( cursor.setToFirst(); cursor.isValid(); cursor.setToNext() )
    {
      subKey = tmpProfile.applicationOrKeyAt(cursor);
      tmpProfile.deleteElementWithApplication(subKey);
    }
  }
  // Now we can delete the profile
  HKEY softwareKey = IRegOpenKey(HKEY_LOCAL_MACHINE,"Software");
  IRegDeleteKey( softwareKey, profileName );
  RegCloseKey( softwareKey );
#endif

#ifndef IC_WIN
  // Delete the profile file
  remove(profileName);
#endif
}

/*------------------------------------------------------------------------------
| IProfile::enableKeyMapping - Enables the mapping of specific characters      |
|   that are invalid for a Win32 registry entry as a key to valid chars.       |
------------------------------------------------------------------------------*/
void IProfile::enableKeyMapping( Boolean mapKey )
{
  fgMapKey = mapKey;
}

/*------------------------------------------------------------------------------
| IProfile::disableKeyMapping - Disables the mapping of specific characters    |
|   that are invalid for a Win32 registry entry as a key to valid chars.       |
------------------------------------------------------------------------------*/
void IProfile::disableKeyMapping( )
{
  enableKeyMapping( false );
}

/*------------------------------------------------------------------------------
| IProfile::isKeyMappingEnabled - Returns the enablement state of mapping of   |
|   specific characters that are invalid for a Win32 registry entry as a key.  |
------------------------------------------------------------------------------*/
Boolean IProfile::isKeyMappingEnabled( )
{
  return ( fgMapKey );
}
#endif // IC_PMWIN

/*------------------------------------------------------------------------------
| IProfile::numberOfApplications                                               |
------------------------------------------------------------------------------*/
unsigned long IProfile::numberOfApplications () const
{
  IMODTRACE_DEVELOP("IProfile::numberOfApplications");

  unsigned long
    result = 0;

  IProfile::Cursor cursor((IProfile&)*this);

  for ( cursor.setToFirst(); cursor.isValid(); cursor.setToNext() )
    result++;

  return result;
}

/*------------------------------------------------------------------------------
| IProfile::numberOfKeys                                                       |
------------------------------------------------------------------------------*/
unsigned long IProfile::numberOfKeys ( const char *applName) const
{
  IMODTRACE_DEVELOP("IProfile::numberOfKeys");
  if (!applName)
    IASSERTPARM( defApplName.length() > 0 );

  IProfile::Cursor cursor((IProfile&)*this,
                          (applName) ? (PSZ)applName : (PSZ)defApplName );

  unsigned long
    result = 0;

  for ( cursor.setToFirst(); cursor.isValid(); cursor.setToNext() )
    result++;

  return result;
}

/*------------------------------------------------------------------------------
| IProfile::containsApplication                                                |
------------------------------------------------------------------------------*/
Boolean IProfile::containsApplication ( const char *applName ) const
{
  IMODTRACE_DEVELOP("IProfile::containsApplication");
  IASSERTPARM(applName != 0);

  IProfile::Cursor cursor((IProfile&)*this);

  for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext() )
  {
#ifdef IC_WIN
    IString currentApp(applicationOrKeyAt(cursor));
    IString applNameU = IRegFixKey(applName);
    if (currentApp.upperCase() == applNameU.upperCase())
      return true;
#endif
#ifndef IC_WIN
    if (applicationOrKeyAt(cursor) == applName)
      return true;
#endif
  } /* endfor */
  return false;
}

/*------------------------------------------------------------------------------
| IProfile::containsKeyName                                                    |
------------------------------------------------------------------------------*/
Boolean IProfile::containsKeyName( const char *key,
                                   const char *applName ) const
{
  IMODTRACE_DEVELOP("IProfile::containsKeyName");

  IProfile::Cursor cursor((IProfile&)*this,
                          (applName) ? applName : (const char*)defApplName );

  for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext() )
  {
#ifdef IC_WIN
    IString currentKey(applicationOrKeyAt(cursor));
    IString keyNameU = IRegFixKey(key);
    if (currentKey.upperCase() == keyNameU.upperCase())
      return true;
#endif
#ifndef IC_WIN
    if (applicationOrKeyAt(cursor) == key)
      return true;
#endif
  } /* endfor */
  return false;
}

#ifndef IC_WIN
/*------------------------------------------------------------------------------
| IProfile::length                                                             |
|   Private function to return the length of the profile data                  |
------------------------------------------------------------------------------*/
unsigned long IProfile::length(const char *keyName, const char *applName) const
{
  unsigned long ulSize;
  if ( PrfQueryProfileSize( hIni,
             (applName) ? (PSZ)applName : (PSZ)defApplName,
             (PSZ)keyName, (unsigned long *) &ulSize))
  {
    return ulSize;
  }
  else
    ITHROWGUIERROR("PrfQueryProfileSize");

  return 0;
}
#endif // not IC_WIN

/*------------------------------------------------------------------------------
| IProfile::applicationOrKeyAt                                                 |
------------------------------------------------------------------------------*/
IString IProfile::applicationOrKeyAt ( const Cursor &cursor ) const
{
  IMODTRACE_DEVELOP("IProfile::applicationOrKeyAt");
  IASSERTPARM(cursor.isValid());

  // cursor pos marks start of app/key name
  return IString( ((char*)cursor.strCl) + cursor.pos - 1 );
}

/*------------------------------------------------------------------------------
| IProfile::Cursor::~Cursor                                                    |
------------------------------------------------------------------------------*/
IProfile::Cursor::~Cursor ( )
{
  delete fCursorData;
}

/*------------------------------------------------------------------------------
| IProfile::Cursor::Cursor                                                     |
------------------------------------------------------------------------------*/
IProfile::Cursor::Cursor ( IProfile  &profile )
  : pos(1), strCl(), applNameCl(), fCursorData(new IProfileCursorData())
{
  IMODTRACE_DEVELOP("IProfile::Cursor::ctor1");

#ifndef IC_WIN
  initCursor( profile.handle(), 0 );
#endif
#ifdef IC_WIN
  initCursor( profile.name(), NULL );
#endif
}

/*------------------------------------------------------------------------------
| IProfile::Cursor:: Cursor                                                    |
------------------------------------------------------------------------------*/
IProfile::Cursor:: Cursor ( IProfile  &profile, const char *applName)
  : pos(1), strCl(), fCursorData(new IProfileCursorData())
{
  IMODTRACE_DEVELOP("IProfile::Cursor::ctor2");

  if (applName)
    applNameCl = applName;
  else
    applNameCl = profile.defaultApplicationName();

#ifndef IC_WIN
  initCursor( profile.handle(), applNameCl );
#endif

#ifdef IC_WIN
  initCursor( profile.name(), applNameCl );
#endif
}

/*------------------------------------------------------------------------------
| IProfile::Cursor::setToFirst                                                 |
| Have the cursor point to the first element                                   |
------------------------------------------------------------------------------*/
IBase::Boolean IProfile::Cursor :: setToFirst ()
{
  this->pos = 1;
  // Valid if not off end...
  return this->pos < this->strCl.length();
}

/*------------------------------------------------------------------------------
| IProfile::Cursor::setToNext                                                  |
| Have the cursor point to the next element                                    |
------------------------------------------------------------------------------*/
IBase::Boolean IProfile::Cursor :: setToNext ()
{
  // Skip to next null...
  this->pos = this->strCl.indexOf( '\0', this->pos );
  // If not found, set position off end...
  if ( this->pos == 0 )
    this->pos = this->strCl.length() + 1;
  // Else, point to next byte...
  else
    this->pos++;
  // Valid if not off end...
  return this->pos < this->strCl.length();
}

/*------------------------------------------------------------------------------
| IProfile::Cursor::setToPrevious                                              |
| Have the cursor point to the previous element                                |
------------------------------------------------------------------------------*/
IBase::Boolean IProfile::Cursor :: setToPrevious ()
{
  // If not at start, back up to previous null...
  if ( this->pos > 1 && this->pos < this->strCl.length() )
    this->pos = this->strCl.lastIndexOf( '\0', this->pos - 2 );
  // Else, leave invalid...
  else
    this->pos = this->strCl.length();
  // If not found, then we must now be at start...
  if ( this->pos == 0 )
    this->pos = 1;
  // Else, position at byte past the null...
  else
    this->pos++;

  // Valid if not off end...
  return this->pos < this->strCl.length();
}

/*------------------------------------------------------------------------------
| IProfile::Cursor::setToLast                                                  |
| Have the cursor point to the last element                                    |
------------------------------------------------------------------------------*/
IBase::Boolean IProfile::Cursor :: setToLast ()
{
   // Position at last null (before the terminating one).
   this->pos = this->strCl.lastIndexOf( '\0', this->strCl.length() - 2 );
   // Advance to next byte.
   this->pos++;

   // Valid if not off end...
   return this->pos < this->strCl.length();
}

/*------------------------------------------------------------------------------
| IProfile::Cursor :: isValid                                                  |
------------------------------------------------------------------------------*/
Boolean IProfile::Cursor :: isValid (Boolean bCheckFile) const
{
  if ( this->pos >= this->strCl.length() )
    return false;
  else
  {
    if (bCheckFile)
    {
      const char* p = ( applNameCl.length() == 0 ) ?
                              (char*)0 : (char*)applNameCl;
#ifndef IC_WIN
      unsigned long ulLen = 0;
      PrfQueryProfileSize(hProfileCl, p, 0, &ulLen);
      IString current( 0, ulLen );
      if (ulLen)
      {
        PrfQueryProfileString( hProfileCl, p, 0, "",
                               (char *)current, ulLen+1 );
      }
      return (current == strCl);
#endif
#ifdef IC_WIN
      // Create another cursor and see if the data strings are the same
      IProfile tmpProfile( fCursorData->profileName );
      IProfile::Cursor cursor(tmpProfile,p);
      return (strCl == cursor.strCl);
#endif
    }
    else
      return true;
  }
}

/*------------------------------------------------------------------------------
| IProfile::Cursor::invalidate                                                 |
| Invalidate the cursor                                                        |
------------------------------------------------------------------------------*/
void IProfile::Cursor :: invalidate ()
{
  this->pos = this->strCl.length();
}

#ifndef IC_WIN
/*------------------------------------------------------------------------------
| IProfile::Cursor::initCursor                                                 |
|   Read profile for all application names or key names                        |
|                                                                              |
|   On exit, strCl will contain:                                               |
|      | entry1 | 0 | entry2 | 0 | ... | entry n | 0 | 0 |                     |
|                                                      A                       |
|                  Terminating null of the IString -----                       |
------------------------------------------------------------------------------*/
void IProfile::Cursor::initCursor( const IProfileHandle& hProfile,
                                   const char *applName )
{
  IMODTRACE_DEVELOP("IProfile::Cursor::initCursor");

  hProfileCl = hProfile;
  unsigned long ulLen;

  PrfQueryProfileSize( hProfile, applName, 0, &ulLen );

  this->strCl = IString( 0, ulLen );

  if (ulLen)
    PrfQueryProfileString( hProfile, applName, 0, "",
                           (char*)strCl, ulLen+1 );
  else
    strCl = "";
}
#endif  // not IC_WIN

#ifdef IC_WIN
/*------------------------------------------------------------------------------
| IProfile::Cursor::initCursor                                                 |
|   Read profile for all application names or key names                        |
|                                                                              |
|   On exit, strCl will contain:                                               |
|      | entry1 | 0 | entry2 | 0 | ... | entry n | 0 | 0 |                     |
|                                                      A                       |
|                  Terminating null of the IString -----                       |
------------------------------------------------------------------------------*/
void IProfile::Cursor::initCursor( const char * profileName,
                                   const char * applName )
{
  IMODTRACE_DEVELOP("IProfile::Cursor::initCursor");

  fCursorData->profileName = profileName;
  strCl = "";

  // Open the profile
  HKEY hKeyProfile = IRegOpenProfile( profileName,
                                      &(fCursorData->isSystemProfile) );
  HKEY enumKey = 0;
  if ( applName != NULL )
  {
    if ( IRegTestKey ( hKeyProfile, applName ))
      enumKey = IRegOpenKey( hKeyProfile, applName );
  }
  else
    enumKey = hKeyProfile;

  if ( enumKey )
  {
    char          buffer[MAX_PATH] = "";
    unsigned long key;
    for ( key = 0;
          RegEnumKey ( enumKey, key, buffer, MAX_PATH ) == ERROR_SUCCESS;
          key++ )
    {
      // Construct a temporary string for the current key
      Boolean invalid = false;
      IString keyName(buffer);

      // If translate characters on, restore characters converted
      if (IProfile::isKeyMappingEnabled())
      {
        unsigned long tmp = keyName.indexOf("/");
        if ( tmp != 0 )
        {
          do
          {
            tmp++;
            switch ( keyName[tmp] )
            {
              case '_':
                keyName[tmp] = '\\';
                break;
              case '-':
                if (IPlatform::isWin32S())
                  keyName[tmp] = ' ';
                break;
              case '0':
                keyName[tmp] = '?';
                break;
              case '1':
                keyName[tmp] = '*';
                break;
              case '/':
                break;
              default:
                invalid = true;
            }
            tmp = keyName.indexOf("/",(UINT)tmp+1);
          }
          while ( tmp != 0 );
          keyName.change("/\\","\\");
          keyName.change("/*","*");
          keyName.change("/?","?");
          if (IPlatform::isWin32S())
            keyName.change("/ "," ");
          keyName.change("//","/");
        }
      }
      // If application name passed
      if ( applName != NULL )
      {
        // Verify that the name is valid before adding to string
        HKEY dataKey = IRegOpenKey( enumKey, keyName );
        unsigned long dataType = IRegDataType(dataKey);
        RegCloseKey(dataKey);
        if ( dataType == REG_NONE )
          invalid = true;
      }
      if ( !invalid )
      {
        strCl += keyName;
        strCl += IString('\0');
      }
    }
  }

  if ( applName != NULL )
  {
    if ( enumKey )
      RegCloseKey(enumKey);
  }
  // Close the key for the profile
  RegCloseKey(hKeyProfile);

}
#endif  // IC_WIN

/*------------------------------------------------------------------------------
| IProfileData::IProfileData                                                   |
|                                                                              |
| This class contains private data used by IProfile                            |
------------------------------------------------------------------------------*/
IProfileData::IProfileData()
#ifdef IC_WIN
  : isSystemProfile(false), defaultString("IProfile-default string"), hKey(0)
#endif
{
  IMODTRACE_DEVELOP("IProfileData::IProfileData::ctor");
}

/*------------------------------------------------------------------------------
| IProfileData::~IProfileData                                                  |
------------------------------------------------------------------------------*/
IProfileData::~IProfileData()
{}

/*------------------------------------------------------------------------------
| IProfileCursorData::IProfileCursorData                                       |
|                                                                              |
| This class contains private data used by IProfile::Cursor                    |
------------------------------------------------------------------------------*/
IProfileCursorData::IProfileCursorData()
#ifdef IC_WIN
  : profileName(), isSystemProfile(false),
    defaultString("IProfile-default string")
#endif
{
  IMODTRACE_DEVELOP("IProfileCursorData::IProfileCursorData::ctor");
}

/*------------------------------------------------------------------------------
| IProfileCursorData::~IProfileCursorData                                      |
------------------------------------------------------------------------------*/
IProfileCursorData::~IProfileCursorData()
{}

