/*******************************************************************************
* FILE NAME: ireslib1.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of IDynamicLinkLibrary and it's      *
*   related classes.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C" {
  #define INCL_DOSMODULEMGR
  #define INCL_DOSRESOURCES    /* Resource types */
  #define INCL_DOSINFOSEG
  #define INCL_DOSPROCESS
  #define INCL_WINERRORS
  #include <iwindefs.h>
}

#include <ireslib.hpp>
#include <iresstat.hpp>
#include <istring.hpp>
#include <irect.hpp>
#include <itrace.hpp>
#include <iset.h>
#include <iexcept.hpp>
#include <ireslock.hpp>


// Segment definitions
#ifdef IC_PAGETUNE
  #define _IRESLIB1_CPP_
  #include <ipagetun.h>
#endif


/* Used to manage shared Handles */
class IDLLModule : public IBase
{
  public:
   IDLLModule(const char* pszModuleName);
   IDLLModule(const char* pszModuleName,const  IModuleHandle& handle);
   ~IDLLModule();


       IString name() const      { return strClName;  }

 Boolean  operator ==(const IDLLModule& hmod) const;
 unsigned long useCount() const { return ulClUseCount;    }
 unsigned long addRef();
 unsigned long removeRef();
IModuleHandle  handle() const    { return hmodCl; }
      IString asDebugInfo() const;
 private:
   IString       strClName;
   IModuleHandle hmodCl;
   unsigned long ulClUseCount;
};

/*------------------------------------------------------------------------------
| IDLLModule::IDLLModule                                                       |
------------------------------------------------------------------------------*/
IDLLModule :: IDLLModule( const char* pszModuleName )
        : strClName(pszModuleName), ulClUseCount(0)
{}

/*------------------------------------------------------------------------------
| IDLLModule::IDLLModule                                                       |
------------------------------------------------------------------------------*/
IDLLModule ::   IDLLModule( const char* pszModuleName,
                            const IModuleHandle& handle )
        : strClName(pszModuleName), ulClUseCount(1), hmodCl(handle)
{}

/*------------------------------------------------------------------------------
| IDLLModule::~IDLLModule                                                      |
------------------------------------------------------------------------------*/
IDLLModule ::   ~IDLLModule() {}

/*------------------------------------------------------------------------------
| IDLLModule::operator ==                                                      |
------------------------------------------------------------------------------*/
Boolean IDLLModule :: operator ==(const IDLLModule& hmod) const
{
  return( (hmod.name() == name()) ? true : false );
}


class IDLLSet : public ISet<IDLLModule*>
{
public:
   IDLLSet ( );
  ~IDLLSet ( );
};

/*------------------------------------------------------------------------------
| IDLLSet::IDLLSet() : ISet<IDLLModule*>                                       |
------------------------------------------------------------------------------*/
IDLLSet :: IDLLSet() : ISet<IDLLModule*>(5) {}

/*------------------------------------------------------------------------------
| IDLLSet::~IDLLSet                                                            |
------------------------------------------------------------------------------*/
IDLLSet :: ~IDLLSet() {}


/*------------------------------------------------------------------------------
| IResLibStaticPtr::deleteDLLSet                                               |
------------------------------------------------------------------------------*/
void IResLibStaticPtr :: deleteDLLSet()
{
  if(pDLLSet)
    delete pDLLSet;
}

/*------------------------------------------------------------------------------
| IResLibStaticPtr::dllSet                                                     |
------------------------------------------------------------------------------*/
IDLLSet&  IResLibStaticPtr ::   dllSet()
{
  if(pDLLSet==0)
    pDLLSet = new IDLLSet;
  return *pDLLSet;
}

/***************************************************************/
/*                                                             */
/* IDLLModule Strategy:                                        */
/*  Calling dllSet() will create IDLLSet if it doesn't exist   */
/*                                                             */
/*  IDynamicLinkLibrary ctor adds IDLLModule object if it      */
/*     doesn't already exist                                   */
/*  addRef() by IDynamicLinkLibrary::open()                    */
/*  removeRef() by IDynamicLinkLibrary::close()                */
/*  IDllModule deleteted by IDynamicLinkLibrary dtor           */
/*    if 0 useCount.                                           */
/*                                                             */
/*                                                             */
/***************************************************************/


/*------------------------------------------------------------------------------
| IDynamicLinkLibrary::IDynamicLinkLibrary                                     |
------------------------------------------------------------------------------*/
IDynamicLinkLibrary :: IDynamicLinkLibrary(const char* pszResourceFile)
{
  IMODTRACE_DEVELOP("IDynamicLinkLibrary::IDynamicLinkLibrary(char*)");

  if (pszResourceFile != 0)
  {
    ITRACE_ALL(IString("DLL: ")+pszResourceFile);

    /* Lock Resource Table */
    IResourceLock reslibLock( IResLibStaticPtr::libraryKey() );
    Boolean fFound = false;

    IDLLModule* pdllNew = new IDLLModule( pszResourceFile );
    IDLLModule* pdllCollection;
    IDLLSet::Cursor rescur( IResLibStaticPtr::dllSet() );
    for (rescur.setToFirst(); rescur.isValid(); rescur.setToNext())
    {
      pdllCollection = IResLibStaticPtr::dllSet().elementAt( rescur );
      ITRACE_ALL(pdllCollection->asDebugInfo());
      if (*pdllCollection == *pdllNew)
      {
        pmodClResource = pdllCollection;
        fFound = true;
        break;
      }
    }
    if( !fFound )
    {
      pmodClResource = pdllNew;
      IResLibStaticPtr::dllSet().add( pmodClResource );
    }
    else
      delete pdllNew;

  open();
  }
}


/*------------------------------------------------------------------------------
| IDynamicLinkLibrary::IDynamicLinkLibrary                                     |
|   Copy Constructor                                                           |
------------------------------------------------------------------------------*/
IDynamicLinkLibrary :: IDynamicLinkLibrary(const IDynamicLinkLibrary& reslib)
              : pmodClResource(reslib.pmodClResource)
{
  IMODTRACE_DEVELOP("IDynamicLinkLibrary::IDynamicLinkLibrary(const &)");

  if (pmodClResource != 0)
  {
    /* Lock Resource Table */
    IResourceLock reslibLock( IResLibStaticPtr::libraryKey() );
    pmodClResource->addRef();
  }
}

/*------------------------------------------------------------------------------
| IDynamicLinkLibrary::IDynamicLinkLibrary                                     |
------------------------------------------------------------------------------*/
IDynamicLinkLibrary :: IDynamicLinkLibrary   (const IModuleHandle& modH)
{
  IMODTRACE_DEVELOP("IDynamicLinkLibrary::IDynamicLinkLibrary(IModuleHandle&)");

  /* Lock Resource Table */
  IResourceLock reslibLock( IResLibStaticPtr::libraryKey() );
  Boolean fFound = false;

  IDLLModule* pdll;
  IDLLSet::Cursor rescur( IResLibStaticPtr::dllSet() );
  for (rescur.setToFirst(); rescur.isValid(); rescur.setToNext())
  {
    pdll = IResLibStaticPtr::dllSet().elementAt( rescur );
    if ( (pdll->handle() != 0) && (pdll->handle() == modH) )
    {
      fFound = true;
      pmodClResource = pdll;
      break;
    }
  }
  if ( !fFound )
  {
    char name[512];
    unsigned long ulErrorCode;

#ifdef IC_PM
    ulErrorCode = DosQueryModuleName( modH, sizeof(name), (char*)name );
    if (ulErrorCode != 0)
    {
      ITHROWSYSTEMERROR( ulErrorCode,
                         "DosQueryModuleName",
                         IBaseErrorInfo::accessError,
                         IException::recoverable );
    }
#endif
#ifdef IC_WIN
    ulErrorCode = GetModuleFileName( modH, (char*)name, sizeof(name) );
    if (ulErrorCode == 0)
    {
      ITHROWGUIERROR2( "GetModuleFileName",
                       IBaseErrorInfo::accessError,
                       IException::recoverable );
    }
#endif

    // Special constructor for existing hmod
    pmodClResource = new IDLLModule( name, modH );
    IResLibStaticPtr::dllSet().add( pmodClResource );
  }

  pmodClResource->addRef();
}


/*------------------------------------------------------------------------------
| IDynamicLinkLibrary::operator=                                               |
------------------------------------------------------------------------------*/
IDynamicLinkLibrary& IDynamicLinkLibrary ::
                        operator=(const IDynamicLinkLibrary& reslib)
{
  IMODTRACE_DEVELOP("IDynamicLinkLibrary::operator=");

  if (this != &reslib)
  {
    IResourceLock reslibLock( IResLibStaticPtr::libraryKey() );
    if (reslib.pmodClResource != 0)
      reslib.pmodClResource->addRef();

    if (pmodClResource != 0)
    {
      pmodClResource->removeRef();
      if (pmodClResource->useCount() == 0)
      {
        IResLibStaticPtr::dllSet().remove( pmodClResource );
        delete pmodClResource;
        pmodClResource = 0;
      }
    }
    pmodClResource = reslib.pmodClResource;
  }

  return( *this);
}



/*------------------------------------------------------------------------------
| IDynamicLinkLibrary::~IDynamicLinkLibrary                                    |
------------------------------------------------------------------------------*/
IDynamicLinkLibrary :: ~IDynamicLinkLibrary()
{
  IMODTRACE_DEVELOP("IDynamicLinkLibrary::~IDynamicLinkLibrary");

  if (handle())
    close();

  if ( (pmodClResource != 0) && (pmodClResource->useCount() == 0) )
  {
    if ( IResLibStaticPtr::dllSet().contains( pmodClResource ) )
    {
      IResLibStaticPtr::dllSet().remove( pmodClResource );
      delete pmodClResource;
    }
    pmodClResource = 0;
  }
}

/*------------------------------------------------------------------------------
| IDynamicLinkLibrary::open                                                    |
------------------------------------------------------------------------------*/
IDynamicLinkLibrary& IDynamicLinkLibrary :: open()
{
  IMODTRACE_DEVELOP("IDynamicLinkLibrary::open");

  if (pmodClResource != 0)
    pmodClResource->addRef();

  return( *this );
}

/*------------------------------------------------------------------------------
| IDynamicLinkLibrary::close                                                   |
------------------------------------------------------------------------------*/
IDynamicLinkLibrary& IDynamicLinkLibrary :: close()
{
  IMODTRACE_DEVELOP("IDynamicLinkLibrary::close");

  /* free the reference to the DLL */
  if (pmodClResource)
    pmodClResource->removeRef();

  return( *this );
}


/*------------------------------------------------------------------------------
| IDynamicLinkLibrary::fileName                                                |
------------------------------------------------------------------------------*/
IString IDynamicLinkLibrary::fileName() const
{
  IMODTRACE_DEVELOP("IDynamicLinkLibrary::fileName");

  char temp[512];
  unsigned long ulErrorCode;
#ifdef IC_PM
  ulErrorCode = DosQueryModuleName( handle(), sizeof(temp), (char*)temp );
  if (ulErrorCode != 0)
  {
    ITHROWSYSTEMERROR( ulErrorCode,
                       "DosQueryModuleName",
                       IBaseErrorInfo::accessError,
                       IException::recoverable );
  }
#endif
#ifdef IC_WIN
  ulErrorCode = GetModuleFileName( handle(), (char*)temp, sizeof(temp) );
  if (ulErrorCode == 0)
  {
    ITHROWGUIERROR2( "GetModuleFileName",
                     IBaseErrorInfo::accessError,
                     IException::recoverable );
  }
#endif

  ITRACE_ALL(temp);
  return( IString( temp ) );
}


/*------------------------------------------------------------------------------
| IDynamicLinkLibrary::isOpen                                                  |
------------------------------------------------------------------------------*/
Boolean IDynamicLinkLibrary::isOpen() const
{
  IMODTRACE_DEVELOP("IDynamicLinkLibrary::isOpen");

  if ( (pmodClResource != 0) && (pmodClResource->handle() != 0) )
    return true;
  else
    return false;
}


/*------------------------------------------------------------------------------
| IDynamicLinkLibrary::addRef                                                  |
------------------------------------------------------------------------------*/
unsigned long IDynamicLinkLibrary::addRef(const IModuleHandle& modh)
{
  IMODTRACE_DEVELOP("IDynamicLinkLibrary::addRef");

  IResourceLock reslibLock( IResLibStaticPtr::libraryKey() );
  unsigned long ulUseCount = 0;
  IDLLModule* pdllCollection;

  IDLLSet::Cursor rescur( IResLibStaticPtr::dllSet() );
  for (rescur.setToFirst(); rescur.isValid(); rescur.setToNext())
  {
    pdllCollection = IResLibStaticPtr::dllSet().elementAt( rescur );
#ifdef IC_PM
    if ((unsigned long)(pdllCollection->handle()) == (unsigned long)modh)
#endif
#ifdef IC_WIN
    if (pdllCollection->handle() == modh)
#endif
    {
      ulUseCount = pdllCollection->addRef();
      break;
    }
  }
  return( ulUseCount );
}


/*------------------------------------------------------------------------------
| IDynamicLinkLibrary::removeRef                                               |
------------------------------------------------------------------------------*/
unsigned long IDynamicLinkLibrary::removeRef(const IModuleHandle& modh)
{
  IMODTRACE_DEVELOP("IDynamicLinkLibrary::removeRef");

  IResourceLock reslibLock( IResLibStaticPtr::libraryKey() );
  unsigned long ulUseCount =0;
  IDLLModule* pdllCollection;

  IDLLSet::Cursor rescur( IResLibStaticPtr::dllSet() );
  for (rescur.setToFirst(); rescur.isValid(); rescur.setToNext())
  {
     pdllCollection = IResLibStaticPtr::dllSet().elementAt( rescur );
#ifdef IC_PM
     if ((unsigned long)(pdllCollection->handle()) == (unsigned long)modh)
#endif
#ifdef IC_WIN
     if (pdllCollection->handle() == modh)
#endif
     {
       ulUseCount = pdllCollection->removeRef();
       break;
     }
  }
  return( ulUseCount );
}


/*------------------------------------------------------------------------------
| IDynamicLinkLibrary::procAddress                                             |
|   Get proceudure address by name                                             |
------------------------------------------------------------------------------*/
void* IDynamicLinkLibrary :: procAddress(const char* pszProcName) const
{
  IMODTRACE_DEVELOP("IDynamicLinkLibrary::procAddress(char*)");
  IASSERT(handle() != 0);

#ifdef IC_PM
  unsigned long ulErrorCode = 0;
  int (_System *pfnTemp)() = 0;
  ulErrorCode = DosQueryProcAddr( handle(), 0, (char*)pszProcName, &pfnTemp );
  if (ulErrorCode != 0)
  {
    // If call failed, upper case string and try again
    IString tryUpper( pszProcName );
    tryUpper.upperCase();
    ulErrorCode = DosQueryProcAddr( handle(), 0, (char*)tryUpper, &pfnTemp );
  }
  if (ulErrorCode != 0)
  {
    ITHROWSYSTEMERROR( ulErrorCode,
                       IString("DosQueryProcAddr: ") +
                                pszProcName,
                       IBaseErrorInfo::accessError,
                       IException::recoverable );
  }
#endif
#ifdef IC_WIN
  FARPROC pfnTemp;
  pfnTemp = GetProcAddress( handle(), pszProcName );
  if ( !pfnTemp )
  {
    ITHROWGUIERROR2( IString("GetProcAddress: ") +
                              pszProcName,
                     IBaseErrorInfo::accessError,
                     IException::recoverable );

  }
#endif

  return( pfnTemp );
}

/*------------------------------------------------------------------------------
| IDynamicLinkLibrary::procAddress                                             |
|  Get procedure address from Ordinal                                          |
------------------------------------------------------------------------------*/
void* IDynamicLinkLibrary :: procAddress(unsigned long ulOrdinal) const
{
  IMODTRACE_DEVELOP("IDynamicLinkLibrary::procAddress(unsigned long)");
  IASSERT(handle() != 0);

#ifdef IC_PM
  unsigned long ulErrorCode = 0;
  int (_System *pfnTemp)() = 0;
  ulErrorCode = DosQueryProcAddr( handle(), ulOrdinal, (char*)0, &pfnTemp );
  if (ulErrorCode != 0)
  {
    ITHROWSYSTEMERROR( ulErrorCode,
                       IString("DosQueryProcAddr #: ") +
                                IString(ulOrdinal),
                       IBaseErrorInfo::accessError,
                       IException::recoverable );
  }
#endif
#ifdef IC_WIN
  FARPROC pfnTemp;
  long ordinal1 =  MAKELONG(LOWORD(ulOrdinal),0);
  char *ordinal = (char *)ordinal1;
  pfnTemp = GetProcAddress( handle(), ordinal );
  if ( !pfnTemp )
  {
    ITHROWGUIERROR2( IString("GetProcAddress #: ") +
                              IString(ulOrdinal),
                     IBaseErrorInfo::accessError,
                     IException::recoverable );

  }
#endif

  return( pfnTemp );
}

/*------------------------------------------------------------------------------
| IDynamicLinkLibrary::isEntryPoint32Bit                                       |
------------------------------------------------------------------------------*/
Boolean IDynamicLinkLibrary :: isEntryPoint32Bit
                                 ( const char *procedureName ) const
{
  IMODTRACE_DEVELOP("IDynamicLinkLibrary::isEntryPoint32Bit(char*)");

  unsigned long
    ulResult = 1;
#ifdef IC_PM
  unsigned long
    ulErrorCode = 1;
  ulErrorCode = DosQueryProcType( this->handle(),
                                  0,
                                  (PSZ)procedureName,
                                  &ulResult );

  if ( ulErrorCode )
  {
    ITHROWSYSTEMERROR( ulErrorCode,
                       IString("DosQueryProcType: ") +
                                procedureName,
                       IBaseErrorInfo::accessError,
                       IException::recoverable);
  }
#endif

  return( (Boolean)ulResult );
}

/*------------------------------------------------------------------------------
| IDynamicLinkLibrary::isEntryPoint32Bit                                       |
------------------------------------------------------------------------------*/
Boolean IDynamicLinkLibrary :: isEntryPoint32Bit
                                 ( unsigned long procedureOrdinal ) const
{
  IMODTRACE_DEVELOP("IDynamicLinkLibrary::isEntryPoint32Bit(unsigned long)");

  unsigned long
    ulResult = 1;
#ifdef IC_PM
  unsigned long
    ulErrorCode = 1;
  ulErrorCode = DosQueryProcType( this->handle(),
                                  procedureOrdinal,
                                  0,
                                  &ulResult );

  if ( ulErrorCode )
  {
    ITHROWSYSTEMERROR( ulErrorCode,
                       IString("DosQueryProcType #: ") +
                                IString(procedureOrdinal),
                       IBaseErrorInfo::accessError,
                       IException::recoverable);
  }
#endif

  return( (Boolean)ulResult );
}

/*------------------------------------------------------------------------------
| IDynamicLinkLibrary::handle                                                  |
|  Return the Handle used to load Resources                                    |
------------------------------------------------------------------------------*/
IModuleHandle IDynamicLinkLibrary::handle() const
{
  IMODTRACE_DEVELOP("IDynamicLinkLibrary::handle");

  if (pmodClResource != 0)
    return pmodClResource->handle();
  else
    return IModuleHandle(0);
}

/*------------------------------------------------------------------------------
| IDLLModule::addRef                                                           |
|   Add a reference to a DLL Module. Create if 0.                              |
------------------------------------------------------------------------------*/
unsigned long IDLLModule :: addRef()
{
  IMODTRACE_ALL("IDLLModule::addRef");

  if (useCount() == 0)
  {
    HMODULE hmod;
    unsigned long ulErrorCode = 0;
#ifdef IC_PM
    char szObjectNameBuf[100];
    ulErrorCode = DosLoadModule( (char*)szObjectNameBuf,
                                 sizeof(szObjectNameBuf),
                                 (char*)name(), &hmod );
    if ( (ulErrorCode != 0) || (hmod == 0) )
    {
      ITHROWSYSTEMERROR( ulErrorCode,
                         IString("DosLoadModule : ") +
                                  name(),
                         IBaseErrorInfo::accessError,
                         IException::recoverable );
    }
#endif
#ifdef IC_WIN
    // Suppress the error message box that appears when the dll load fails
    unsigned int
      uiErrorMode = SetErrorMode( SEM_FAILCRITICALERRORS );

    hmod = LoadLibrary( name() );
    if (hmod == 0)
    {
      ulErrorCode = GetLastError();
      if ( ulErrorCode != 0 )
      {
        SetErrorMode( uiErrorMode );
        ITHROWGUIERROR2( IString("LoadLibrary : ") +
                                  name(),
                         IBaseErrorInfo::accessError,
                         IException::recoverable );
      }
    }

    SetErrorMode( uiErrorMode );
#endif
    hmodCl = hmod;
  }

  ulClUseCount++;
  return( ulClUseCount );
}


/*------------------------------------------------------------------------------
| IDLLModule::removeRef                                                        |
|  Remove a reference. Delete if 0.                                            |
------------------------------------------------------------------------------*/
unsigned long IDLLModule :: removeRef()
{
  IMODTRACE_ALL("IDLLModule::removeRef");

  ulClUseCount--;
  if (ulClUseCount == 0)
  {
    unsigned long ulErrorCode;
#ifdef IC_PM
    ulErrorCode = DosFreeModule( handle() );
    if (ulErrorCode != 0)
    {
      ITHROWSYSTEMERROR( ulErrorCode,
                         IString("DosFreeModule : ") +
                                  name(),
                         IBaseErrorInfo::accessError,
                         IException::recoverable );
    }
#endif
#ifdef IC_WIN
    ulErrorCode = (! FreeLibrary( handle() ) );
    if (ulErrorCode != 0)
    {
      ITHROWGUIERROR2( IString("FreeLibrary : ") +
                                name(),
                       IBaseErrorInfo::accessError,
                       IException::recoverable );
    }
#endif
    hmodCl = 0;
  }

  return( ulClUseCount );
}

/*------------------------------------------------------------------------------
| IDLLModule::asDebugInfo                                                      |
------------------------------------------------------------------------------*/
IString IDLLModule :: asDebugInfo() const
{
  IString str="IDLLModule(Name(";
  str += strClName;
  str += "),Handle(";
  str += IString(hmodCl.asString());
  str += "),UseCount(";
  str += IString(ulClUseCount);
  str += ")";
  return( str );
}
