/*******************************************************************************
* FILE NAME: isubmenu.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in isubmenu.cpp                                                            *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

extern "C"
{
  #define INCL_WINMENUS
  #define INCL_WINWINDOWMGR
  #include <iwindefs.h>
}

#include <icconst.h>
#include <iexcept.hpp>
#include <iframe.hpp>
#include <iframprv.hpp>
#include <imenuprv.hpp>
#include <imnitem.hpp>
#include <ithread.hpp>
#include <ireslib.hpp>
#include <isubmenu.hpp>
#include <iplatfrm.hpp>
#include <iwindow.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _ISUBMENU_CPP_
  #include <ipagetun.h>
#endif

// enum and ISubmenuList used for maintaining changes made
// so that they can be undone

enum RecordType { removeRecord, addRecord, changeRecord };

class ISubmenuList {
public:
 RecordType     recordType;
 IMenuItem*     pMenuItem;
 ISubmenuList*  pNext;
} ;


#ifdef IC_MOTIF
/*------------------------------------------------------------------------------
  Constructor.
------------------------------------------------------------------------------*/
ISubmenu :: ISubmenu(const IWindowHandle& submenuHandle)
  : hwndHandle(submenuHandle),
    fSubmenuData(0),
    fItemList(0)
{
   initialize(submenuHandle);
}
#endif

#ifdef IC_PMWIN
/*------------------------------------------------------------------------------
  Constructor.
------------------------------------------------------------------------------*/
ISubmenu :: ISubmenu ( const IMenuHandle& submenuHandle )
  : hwndHandle(submenuHandle),
    fSubmenuData(0),
    fItemList(0)
{
#ifdef IC_WIN
  /*------------------------------------------------------------------*/
  /* For either menu type, need to find owner of menu                 */
  /*------------------------------------------------------------------*/
  FRAMEINFO frameInfo;
  frameInfo.hmenu  = submenuHandle;
  frameInfo.owner  = 0;
  frameInfo.frame  = 0;

  /*------------------------------------------------------------------*/
  /* Enumerate all of the top level windows (to find frame w/ table)  */
  /*------------------------------------------------------------------*/
  EnumWindows( (WNDENUMPROC)IMenuPrivate::EnumAllFrames,
               (LPARAM)&frameInfo );

  /*------------------------------------------------------------------*/
  /* If owner parameter set (meaning found), then use it              */
  /*------------------------------------------------------------------*/
  if (frameInfo.owner)
  {
    fowner     = frameInfo.owner;
    fmenuFrame = frameInfo.frame;
  }
#endif
}
#endif

/*------------------------------------------------------------------------------
| ISubmenu::~ISubmenu                                                          |
------------------------------------------------------------------------------*/
ISubmenu :: ~ISubmenu()
{
  if (fItemList) {
     ISubmenuList* pList = fItemList;
     ISubmenuList* pRec  = pList->pNext;
     while (pRec) {
        ISubmenuList* pTemp = pRec;
        pRec = pRec->pNext;
        delete pTemp;
     } /* endwhile */
     delete pList;
  } /* endif */
  // delete fSubmenuData;
}

/*------------------------------------------------------------------------------
| ISubmenu::handle                                                             |
------------------------------------------------------------------------------*/
IWindowHandle ISubmenu::handle ( ) const
{
#ifdef IC_MOTIF
   return hwndHandle;
#else
   return (IWindowHandle)hwndHandle;
#endif
}

#ifndef IC_MOTIF
/*------------------------------------------------------------------------------
| ISubmenu::menuHandle                                                         |
------------------------------------------------------------------------------*/
IMenuHandle ISubmenu::menuHandle ( ) const
{
   return hwndHandle;
}
#endif

//
// For all ISubmenu member functions the basic approach is:
// Call the corresponding function in IMenu to visually change the
// Menu, then journal the change so that it can be undone.
//

/*------------------------------------------------------------------------------
| ISubmenu::checkItem                                                          |
------------------------------------------------------------------------------*/
ISubmenu& ISubmenu :: checkItem(unsigned long itemId, Boolean checked)
{
  IMenuItem origItem = Inherited::menuItem(itemId);
  Inherited::checkItem(itemId, checked);
  setUndoChangeItem(origItem);
  return *this;
}

/*------------------------------------------------------------------------------
| ISubmenu::enableItem                                                         |
------------------------------------------------------------------------------*/
ISubmenu& ISubmenu :: enableItem(unsigned long itemId, Boolean enabled)
{
  IMenuItem origItem = Inherited::menuItem(itemId);
  Inherited::enableItem(itemId, enabled);
  setUndoChangeItem(origItem);
  return *this;
}

/*------------------------------------------------------------------------------
| ISubmenu::uncheckItem                                                        |
------------------------------------------------------------------------------*/
ISubmenu& ISubmenu :: uncheckItem(unsigned long itemId)
{
  IMenuItem origItem = Inherited::menuItem(itemId);
  Inherited::checkItem(itemId, false);
  setUndoChangeItem(origItem);
  return *this;
}

/*------------------------------------------------------------------------------
| ISubmenu::disableItem                                                        |
------------------------------------------------------------------------------*/
ISubmenu& ISubmenu :: disableItem(unsigned long itemId)
{
  IMenuItem origItem = Inherited::menuItem(itemId);
  Inherited::enableItem(itemId, false);
  setUndoChangeItem(origItem);
  return *this;
}

/*------------------------------------------------------------------------------
| ISubmenu::setItem                                                            |
------------------------------------------------------------------------------*/
ISubmenu& ISubmenu :: setItem   ( const IMenuItem      &item )
{
  IMenuItem origItem = Inherited::menuItem(item.id());
  Inherited::setItem(item);
  setUndoChangeItem(origItem);
  return *this;
}

/*------------------------------------------------------------------------------
| ISubmenu::setText                                                            |
------------------------------------------------------------------------------*/
ISubmenu& ISubmenu :: setText( unsigned long        itemId,
                               const char*          newText )
{
  IMenuItem origItem = Inherited::menuItem(itemId);
  Inherited::setText(itemId, newText);
  setUndoChangeItem(origItem);
  return *this;
}


/*------------------------------------------------------------------------------
| ISubmenu::setText                                                            |
------------------------------------------------------------------------------*/
ISubmenu& ISubmenu :: setText   ( unsigned long itemId,
                                  const IResourceId    &newTextResId )
{
  IMenuItem origItem = Inherited::menuItem(itemId);
  Inherited::setText(itemId, newTextResId);
  setUndoChangeItem(origItem);
  return *this;
}


/*------------------------------------------------------------------------------
| ISubmenu::setBitmap                                                          |
------------------------------------------------------------------------------*/
ISubmenu& ISubmenu :: setBitmap ( unsigned long        itemId,
                                  const IResourceId    &newBitmapResId )
{
  IMenuItem origItem = Inherited::menuItem(itemId);
  Inherited::setBitmap(itemId, newBitmapResId);
  setUndoChangeItem(origItem);
  return *this;
}

/*------------------------------------------------------------------------------
| ISubmenu::setBitmap                                                          |
------------------------------------------------------------------------------*/
ISubmenu& ISubmenu :: setBitmap ( unsigned long        itemId,
                                  unsigned long        newBitmapResId )
{
  IMenuItem origItem = Inherited::menuItem(itemId);
  Inherited::setBitmap(itemId, IResourceId(newBitmapResId));
  setUndoChangeItem(origItem);
  return *this;
}

/*------------------------------------------------------------------------------
| ISubmenu::setBitmap                                                          |
------------------------------------------------------------------------------*/
ISubmenu& ISubmenu :: setBitmap ( unsigned long        itemId,
                                 const IBitmapHandle  &bitmapHandle )
{
  IMenuItem origItem = Inherited::menuItem(itemId);
  Inherited::setBitmap(itemId, bitmapHandle);
  setUndoChangeItem(origItem);
  return *this;
}

/*------------------------------------------------------------------------------
| ISubmenu::deleteItem                                                         |
------------------------------------------------------------------------------*/
ISubmenu& ISubmenu :: deleteItem(unsigned long itemId)
{
#ifdef IC_WIN
  IMenuItem* pNewItem = 0;
  /********************************************************************/
  /* Get the menu state. If invalid, query will throw exception       */
  /********************************************************************/
  if (IPlatform::isWin95() || IPlatform::isNTNewShell())
  {
    MENUITEMINFO miItem;
    miItem.cbSize = sizeof(MENUITEMINFO);
    miItem.fMask = MIIM_STATE | MIIM_ID;

    // See if the item exists in the menu, if not throw exception.
    if (!(GetMenuItemInfo( menuHandle(), (UINT)itemId, FALSE, &miItem )))
    {
      ITHROWLIBRARYERROR( IC_INVALID_MENUITEM,
                          IBaseErrorInfo::invalidRequest,
                          IException::recoverable );
    }

    // Save the item while we still got it
    // We're using the default copy constructor
    pNewItem = new IMenuItem(Inherited::menuItem(itemId));

    // Remove menu (not DeleteMenu) since we may want to restore this
    RemoveMenu( menuHandle(), (UINT)itemId, MF_BYCOMMAND );
  }
  else
  {
    IMenuHandle tempHMenu;
    long itemPos=0;
    Boolean located = IMenuPrivate::locateMenuItem( fmenuFrame,
                            menuHandle(), itemId, &itemPos, &tempHMenu );

    // See if the item exists in the menu, if not throw exception.
    if (!located || (GetMenuState( tempHMenu, (UINT)itemPos,
                                            MF_BYPOSITION ) == 0xFFFFFFFF ))
    {
      ITHROWLIBRARYERROR( IC_INVALID_MENUITEM,
                          IBaseErrorInfo::invalidRequest,
                          IException::recoverable );
    }

    // Save the item while we still got it
    // We're using the default copy constructor
    pNewItem = new IMenuItem(Inherited::menuItem(itemId));

    // Remove menu (not DeleteMenu) since we may want to restore this
    RemoveMenu( tempHMenu, (UINT)itemPos, MF_BYPOSITION );
  }
#else
  MENUITEM miItem;

  if(!(Boolean)sendEvent(MM_QUERYITEM,
                         IEventParameter1(itemId, false),
                         IEventParameter2(&miItem)))
  {
     ITHROWLIBRARYERROR(IC_INVALID_MENUITEM,
                        IBaseErrorInfo::invalidRequest,
                        IException::recoverable);
  } /* endif */

  // Save the item while we still got it
  // We're using the default copy constructor
  IMenuItem* pNewItem = new IMenuItem(Inherited::menuItem(itemId));

  // Must REMOVEITEM rather than DELETEITEM as in IMenu::deleteItem().
  // DELETEITEM could destroy a bitmap that we want to be able to restore.
  sendEvent(MM_REMOVEITEM, IEventData(itemId, false), 0);
#endif

  // Save the item in the list
  if (fItemList) {
     ISubmenuList* pNew = new ISubmenuList;
     pNew->recordType   = removeRecord;
     pNew->pMenuItem    = pNewItem;
     pNew->pNext        = fItemList;
     fItemList          = pNew;
  } else {
     ISubmenuList* pList = fItemList;
     pList              = new ISubmenuList;
     pList->recordType  = removeRecord;
     pList->pMenuItem   = pNewItem;
     pList->pNext       = 0;
     fItemList          = pList;
  } /* endif */

  return *this;
}

/*------------------------------------------------------------------------------
| ISubmenu::removeSubmenu                                                      |
------------------------------------------------------------------------------*/
ISubmenu& ISubmenu::removeSubmenu ( unsigned long itemWithSubmenuId )
{
   // Create two copies of the item, one for archiving, one for changing
   IMenuItem origItem = Inherited::menuItem(itemWithSubmenuId);
   IMenuItem mn  = Inherited::menuItem(itemWithSubmenuId);

   // Reset the submenu handle in one item and then change the menu
   mn.setSubmenuHandle(0);
   Inherited::setItem(mn);

   // Finally archive the original item removed and exit
   setUndoChangeItem(origItem);
   return *this;
}

/*------------------------------------------------------------------------------
| ISubmenu::Cursor::Cursor                                                     |
| Cursor ctor.                                                                 |
------------------------------------------------------------------------------*/
ISubmenu::Cursor::Cursor(const ISubmenu&  menu)
  : lCurrent(0), pSubmenu((ISubmenu*)&menu), fCursorData(0)
{
#ifdef IC_WIN
  IASSERTSTATE( menu.menuHandle() != 0 );
#else
  IASSERTSTATE(WinIsWindow(IThread::current().anchorBlock(), menu.handle() ));
#endif
}

/*------------------------------------------------------------------------------
| ISubmenu::Cursor::Cursor                                                     |
| Copy ctor.                                                                   |
------------------------------------------------------------------------------*/
ISubmenu::Cursor::Cursor(const ISubmenu::Cursor& cursor )
  : lCurrent(cursor.lCurrent), pSubmenu(cursor.pSubmenu),
    fCursorData(cursor.fCursorData)
{
}

/*------------------------------------------------------------------------------
| ISubmenu::Cursor :: ~Cursor                                                  |
|                                                                              |
| Empty destructor here for page tuning.                                       |
------------------------------------------------------------------------------*/
ISubmenu::Cursor :: ~Cursor()
{
  // delete fCursorData;
}

/*------------------------------------------------------------------------------
| ISubmenu::menuItemId                                                         |
|  Return menu-item ID of item at cursor position                              |
------------------------------------------------------------------------------*/
unsigned long ISubmenu::menuItemId ( const Cursor& cursor ) const
{
  IASSERTSTATE( cursor.isValid());
  unsigned long ulId;
#ifdef IC_WIN
  Boolean rc = false;
  if (!(IPlatform::isWin95() || IPlatform::isNTNewShell()))
  {
    // Query for the id of the item at the cursor position
    ulId = GetMenuItemID( menuHandle(), (int)cursor.lCurrent );

    // If error, the item might be a submenu so query for a menu handle
    if ( ulId == 0xFFFFFFFFul )
    {
      HMENU hPop = GetSubMenu( menuHandle(), (int)cursor.lCurrent );

      // If a menu handle is returned, then get the id from the lookup table
      if (hPop)
        ulId = IMenuPrivate::getItemIdFromLookUpTable( fmenuFrame, hPop );
    }

    // If -1, then either it wasn't a popup, or it wasn't in the lookup table
    rc = ( ulId != 0xFFFFFFFFul );
  }
  else
  {
    MENUITEMINFO miItem;
    miItem.cbSize = sizeof(MENUITEMINFO);
    miItem.fMask = MIIM_STATE | MIIM_ID;

    rc = GetMenuItemInfo( menuHandle(), (int)cursor.lCurrent, TRUE,
                          &miItem );
    ulId = (unsigned long)miItem.wID;
  }

  if (!rc)
#else
  IWindowHandle hwnd = (cursor.pSubmenu->handle());
  ulId = (unsigned long)hwnd.sendEvent( MM_ITEMIDFROMPOSITION,
                                        (int)cursor.lCurrent, 0);
  if (ulId == MIT_ERROR)
#endif
  {
     ITHROWLIBRARYERROR(IC_INVALID_MENUITEM,
                        IBaseErrorInfo::invalidRequest,
                        IException::recoverable);
  } /* endif */
  return ulId;
}

/*------------------------------------------------------------------------------
| ISubmenu::Cursor ISubmenu::cursor                                            |
------------------------------------------------------------------------------*/
ISubmenu::Cursor ISubmenu::cursor ( unsigned long itemId ) const
{
#ifdef IC_WIN
   long ulId;
   if (!IMenuPrivate::locateMenuItem( fmenuFrame, menuHandle(), itemId,
                                      &ulId, NULL ))
#else
   unsigned long ulId = handle().sendEvent(MM_ITEMPOSITIONFROMID,
                                         IEventData(itemId, false), 0);
   if (ulId == MIT_NONE)
#endif
   {
     ITHROWLIBRARYERROR(IC_INVALID_MENUITEM,
                        IBaseErrorInfo::invalidRequest,
                        IException::recoverable);
   } /* endif */
   ISubmenu::Cursor cursor(*this);
   cursor.lCurrent = ulId;
   return cursor;
}

/*------------------------------------------------------------------------------
| ISubmenu::undo                                                               |
| Undo changes. Restores previous menu-items and state                         |
------------------------------------------------------------------------------*/
ISubmenu&  ISubmenu :: undo()
{
  if (fItemList) {
     ISubmenuList* pList = fItemList;
     ISubmenuList* pNextRec = pList;
     while(pNextRec) {

        switch (pNextRec->recordType) {
        case addRecord:
           Inherited::deleteItem(pNextRec->pMenuItem->id());
           break;
        case changeRecord:
           Inherited::setItem(*pNextRec->pMenuItem);
           break;
        case removeRecord:
           Inherited::addItem(*pNextRec->pMenuItem);
        } /* endswitch */

        ISubmenuList* pTemp = pNextRec;
        pNextRec = pNextRec->pNext;
        delete pTemp->pMenuItem;
        delete pTemp;
     } /* endwhile */
     fItemList = 0;
  } /* endif */
  return *this;
}

/*------------------------------------------------------------------------------
| ISubmenu::setUndoChangeItem                                                  |
|  Save a change so that it can be undone                                      |
------------------------------------------------------------------------------*/
void ISubmenu::setUndoChangeItem(const IMenuItem& origItem)
{
  if (fItemList) {
     ISubmenuList* pList = fItemList;
     ISubmenuList* pNextRec = pList;
     while(pNextRec) {
        if (pNextRec->pMenuItem->id() == origItem.id() ) {
           if (pNextRec->recordType == changeRecord || pNextRec->recordType == addRecord)
               return;
        } /* endif */
        pNextRec = pNextRec->pNext;
     } /* endif */

     IMenuItem* pNewItem = new IMenuItem(origItem);
     ISubmenuList* pNew = new ISubmenuList;
     pNew->recordType  = changeRecord;
     pNew->pMenuItem   = pNewItem;
     pNew->pNext       = fItemList;
     fItemList         = pNew;
  } else {
     IMenuItem* pNewItem = new IMenuItem(origItem);
     ISubmenuList* pList    = new ISubmenuList;
     pList->recordType  = changeRecord;
     pList->pMenuItem   = pNewItem;
     pList->pNext       = 0;
     fItemList          = pList;
  } /* endif */
}

/*------------------------------------------------------------------------------
| ISubmenu::setUndoAddItem                                                     |
|  Save an add so that it can be undone                                        |
------------------------------------------------------------------------------*/
void ISubmenu::setUndoAddItem(unsigned long itemId)
{
  IMenuItem* pNewItem = new IMenuItem(itemId);

  if (fItemList) {
     ISubmenuList* pNew  = new ISubmenuList;
     pNew->recordType  = addRecord;
     pNew->pMenuItem   = pNewItem;
     pNew->pNext       = fItemList;
     fItemList         = pNew;
  } else {
     ISubmenuList* pList  = new ISubmenuList;
     pList->recordType  = addRecord;
     pList->pMenuItem   = pNewItem;
     pList->pNext       = 0;
     fItemList          = pList;
  } /* endif */
}

/*------------------------------------------------------------------------------
| ISubmenu::addItem                                                            |
------------------------------------------------------------------------------*/
ISubmenu& ISubmenu :: addItem(IMenuItem&          menuItem,
                              unsigned long       intoSubmenuId )
{
  Inherited::addItem(menuItem);
  setUndoAddItem(menuItem.id());
  return *this;
}

/*------------------------------------------------------------------------------
| ISubmenu::addSeparator                                                       |
------------------------------------------------------------------------------*/
ISubmenu& ISubmenu :: addSeparator(unsigned long itemId,
                                   unsigned long intoSubmenuId)
{
  Inherited::addSeparator(itemId, 0);
  return *this;
}

/*------------------------------------------------------------------------------
| ISubmenu::addSeparator                                                       |
------------------------------------------------------------------------------*/
ISubmenu& ISubmenu :: addSeparator(unsigned long itemId)
{
  // This derived function changes the meaning of the single argument
  // from the base class IMenu.  In IMenu, its the intoSubmenuId.  In
  // this class, its the itemId.   This is to preserve compatibility
  // with prior releases.
  Inherited::addSeparator(itemId, 0);
  return *this;
}

/*------------------------------------------------------------------------------
| ISubmenu::addText                                                            |
------------------------------------------------------------------------------*/
ISubmenu& ISubmenu :: addText ( unsigned long       itemId,
                                const char*         itemText,
                                unsigned long       intoSubmenuId )
{
  Inherited::addText(itemId, itemText);
  return *this;
}

/*------------------------------------------------------------------------------
| ISubmenu::addText                                                            |
------------------------------------------------------------------------------*/
ISubmenu& ISubmenu :: addText( unsigned long       itemId,
                               const IResourceId  &newItemTextResId ,
                               unsigned long       intoSubmenuId )
{
  Inherited::addText(itemId, newItemTextResId);
  return *this;
}

/*------------------------------------------------------------------------------
| ISubmenu::addBitmap                                                          |
------------------------------------------------------------------------------*/
ISubmenu& ISubmenu :: addBitmap    ( unsigned long       itemId,
                                     const IResourceId   &newItemBitmapResId,
                                     unsigned long       intoSubmenuId )
{
  Inherited::addBitmap(itemId, newItemBitmapResId);
  return *this;
}

/*------------------------------------------------------------------------------
  Add bitmap menu-item from .rc
------------------------------------------------------------------------------*/
ISubmenu& ISubmenu :: addBitmap    ( unsigned long itemId,
                                     unsigned long newItemBitmapResId,
                                     unsigned long intoSubmenuId )
{
  return addBitmap(itemId, IResourceId(newItemBitmapResId));
}


/*------------------------------------------------------------------------------
| ISubmenu::addBitmap                                                          |
------------------------------------------------------------------------------*/
ISubmenu& ISubmenu :: addBitmap    ( unsigned long       itemId,
                                     const IBitmapHandle &itemBitmap ,
                                     unsigned long       intoSubmenuId )
{
  Inherited::addBitmap(itemId, itemBitmap);
  return *this;
}

/*------------------------------------------------------------------------------
| IBase::Boolean ISubmenu::Cursor :: setToNext                                 |
------------------------------------------------------------------------------*/
IBase::Boolean ISubmenu::Cursor :: setToNext ()
{
  if (lCurrent >= 0 && lCurrent < pSubmenu->numberOfItems() - 1 ) {
     ++lCurrent;
     return true;
  } else {
     lCurrent = -1L;
     return false;
  } /* endif */
}

/*------------------------------------------------------------------------------
| ISubmenu::add                                                                |
------------------------------------------------------------------------------*/
ISubmenu& ISubmenu::add( IMenuItem&  menuItem, Cursor&   cursor)
{
   menuItem.setIndex(cursor.lCurrent);
   return cursor.pSubmenu->addItem(menuItem);
}

/*------------------------------------------------------------------------------
| ISubmenu::addAsNext                                                          |
------------------------------------------------------------------------------*/
ISubmenu& ISubmenu::addAsNext ( IMenuItem&  menuItem, Cursor&   cursor)
{
   menuItem.setIndex(cursor.lCurrent + 1);
   return cursor.pSubmenu->addItem(menuItem);
}

/*------------------------------------------------------------------------------
| ISubmenu::elementAt                                                          |
------------------------------------------------------------------------------*/
IMenuItem ISubmenu::elementAt ( const IMenu::Cursor&   cursor) const
{
   ITHROWLIBRARYERROR(IC_MEMBER_ACCESS_ERROR,
                      IBaseErrorInfo::invalidRequest,
                      IException::recoverable);
   IMenuItem mn(0);
   return mn;
}

/*------------------------------------------------------------------------------
| ISubmenu::deleteAt                                                           |
------------------------------------------------------------------------------*/
ISubmenu& ISubmenu::deleteAt ( IMenu::Cursor&   cursor)
{
   ITHROWLIBRARYERROR(IC_MEMBER_ACCESS_ERROR,
                      IBaseErrorInfo::invalidRequest,
                      IException::recoverable);
   return *this;
}

/*------------------------------------------------------------------------------
| ISubmenu::removeSubmenuAt                                                    |
------------------------------------------------------------------------------*/
ISubmenu& ISubmenu::removeSubmenuAt ( IMenu::Cursor&   cursor)
{
   ITHROWLIBRARYERROR(IC_MEMBER_ACCESS_ERROR,
                      IBaseErrorInfo::invalidRequest,
                      IException::recoverable);
   return *this;
}

/*------------------------------------------------------------------------------
| ISubmenu::add  (hidden IMenu::Cursor version)                                |
------------------------------------------------------------------------------*/
ISubmenu& ISubmenu::add ( IMenuItem&  menuItem, IMenu::Cursor&   cursor)
{
   ITHROWLIBRARYERROR1(IC_MEMBER_ACCESS_ERROR,
                       IBaseErrorInfo::invalidRequest,
                       IException::recoverable,
                       "ISubmenu");
   return *this;
}

/*------------------------------------------------------------------------------
| ISubmenu::addAsNext (hidden IMenu::Cursor version)                           |
------------------------------------------------------------------------------*/
ISubmenu& ISubmenu::addAsNext ( IMenuItem&  menuItem, IMenu::Cursor&   cursor)
{
   ITHROWLIBRARYERROR1(IC_MEMBER_ACCESS_ERROR,
                       IBaseErrorInfo::invalidRequest,
                       IException::recoverable,
                       "ISubmenu");
   return *this;
}

/*------------------------------------------------------------------------------
  Have the cursor point to the first element
------------------------------------------------------------------------------*/
IBase::Boolean ISubmenu::Cursor :: setToFirst ()
{
  lCurrent = 0;
  return true;
}

/*------------------------------------------------------------------------------
  Have the cursor point to the previous element
------------------------------------------------------------------------------*/
IBase::Boolean ISubmenu::Cursor :: setToPrevious ()
{
  --lCurrent;
  return (lCurrent < 0) ? false : true;
}

/*------------------------------------------------------------------------------
  Have the cursor point to the last element
------------------------------------------------------------------------------*/
IBase::Boolean ISubmenu::Cursor :: setToLast ()
{
  lCurrent = pSubmenu->numberOfItems() - 1;
  return (lCurrent < 0) ? false : true;
}

/*------------------------------------------------------------------------------
  Invalidate the cursor
------------------------------------------------------------------------------*/
IBase::Boolean ISubmenu::Cursor :: isValid () const
{
  return (lCurrent < 0) ? false : true;
}

/*------------------------------------------------------------------------------
  Invalidate the cursor
------------------------------------------------------------------------------*/
void ISubmenu::Cursor :: invalidate ()
{
  lCurrent = -1L;
}

/*------------------------------------------------------------------------------
  Get menu item at cursor position.
------------------------------------------------------------------------------*/
IMenuItem ISubmenu::elementAt ( const Cursor& cursor ) const
{
   return cursor.pSubmenu->menuItem(cursor.pSubmenu->menuItemId(cursor));
}

/*------------------------------------------------------------------------------
  Delete menu item at cursor position.
------------------------------------------------------------------------------*/
ISubmenu& ISubmenu::deleteAt      ( Cursor&           cursor   )
{
   return cursor.pSubmenu->deleteItem(cursor.pSubmenu->menuItemId(cursor));
}

/*------------------------------------------------------------------------------
  Remove submenu at cursor position.
------------------------------------------------------------------------------*/
ISubmenu& ISubmenu::removeSubmenuAt ( Cursor&           cursor   )
{
   return cursor.pSubmenu->removeSubmenu(menuItemId(cursor));
}

