/*******************************************************************************
* FILE NAME: ithread.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Implementation of the class(es):                                           *
*     IThread                                                                  *
*     IThread::Cursor                                                          *
*     IThreadFn                                                                *
*     IThreadMemberFn                                                          *
*     ICurrentThread                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

extern "C" {
  #include <stdlib.h>
  #define INCL_DOSPROCESS
  #define INCL_DOSSESMGR
  #define INCL_DOSERRORS
  #define INCL_WINMESSAGEMGR
  #define INCL_WINFRAMEMGR
  #define INCL_WINHOOKS
  #include <iwindefs.h>
}

#include <ithread.hpp>
#include <icconst.h>
#include <icritsec.hpp>
#include <iexcept.hpp>
#include <iplatfrm.hpp>
#include <ireslock.hpp>
#include <istring.hpp>
#include <itrace.hpp>
#include <iwinlsts.hpp>
#include <isrtset.h>

#ifdef IC_PERF
#ifdef IC_WIN
  #include <iperf.h>
#endif
#ifdef IC_PM
  #include <icsperf.h>
#endif
#endif

#ifdef IC_WIN
   // Adjust a 0 centered value (-15,-2..2,15) to a 0..31 range
   #define IC_THREAD_PRIORITY_ADJUST 15
#endif


// Segment definitions.
#ifdef IC_PAGETUNE
  #define _ITHREAD_CPP_
  #include <ipagetun.h>
#endif

#ifdef IC_WIN
typedef ISortedSet<unsigned int>
   IWMQuitStack;

void _setProcessDetachInProgress(void);
extern "C"  void _Optlink __IInitTerm(unsigned long ulflag);

#endif

/*------------------------------------------------------------------------------
| Objects of the class IStartThread maintain information about started threads.|
| Instances of IStartedThread are managed by class IThread.                    |
------------------------------------------------------------------------------*/
class IStartedThread : public IRefCounted {
typedef IRefCounted
  Inherited;
public:
  IStartedThread ( const IThreadId &tid = 0 );
 ~IStartedThread ( );

IString
  asDebugInfo ( ) const;

Boolean
  isStarted ( );

void
  end ( );

#ifdef IC_PM
tib_t
 *tib ( );
#endif

IString
  variable ( const IString &key ) const;

IStartedThread
 &setVariable ( const IString &key,
                const IString &value );

static void
  add        ( IStartedThread &aThread ),
  remove     ( IStartedThread &aThread ),
  initThread ( IStartedThread *pThread );

static IStartedThread
 *locate ( const IThreadId &tid );

static Boolean
  isTIDValid ( const IThreadId &tid );

static void
  flushCache ( );

static IStartedThread
 *checkCacheFor ( const IThreadId &tid );

static void
  cache ( const IThreadId &tid,
          IStartedThread  *pStarted );

static void
  watch ( );

static IString
  classDebugInfo ( );

#ifdef IC_WIN
static IThreadId
  waitForThreadEnd(Boolean waitForAll,
                   const IThreadId& excludedThread,
                   Boolean throwOnError=true );

static IThreadHandle
  currentHandle( Boolean create=true );    // creates a handle for current thread
#endif

static IThreadId
  idWithHandle   (const IThreadHandle& handle );
static IThreadHandle
  handleWithId   (const IThreadId& tid );

#ifdef IC_WIN
IThreadHandle
  handle( ) const { return fhandle; }
#else
IThreadHandle
  handle( ) const { return id; }
#endif

IThreadId
  id;            // TID.

IMessageQueueHandle
  hmq;           // HMQ.

IAnchorBlockHandle
  hab;           // HAB.

Boolean
  autoInitPM,    // Auto-init PM flag.
  started,       // Started flag.
  startedHere,   // Started-by-initThread flag.
  pmStartedHere, // PM-initialized-by-IThread flag.
  inArray;       // In-threads-array flag.

long
  queueSize;     // PM message queue size.

unsigned long
  stkSize,       // Stack size.
  fCount,        // Creation order.
  inMsgLoop;     // In-msg-polling-loop count (increment when polling loop
                 // begins and decrement when polling loop ends to effectively
                 // track threads that are calling processMsgs(), and issue
                 // the quit message to each thread when it terminates.

IApplication::PriorityClass
  priorityClass; // Priority class.

int
  priorityLevel; // Priority level.

IThreadFn
 *fnObj;         // Function object reference.

unsigned
  slot;          // Slot in threads-array.

IWindowList
 *windowList;    // Pointer to info list.

#ifdef IC_WIN
IDialogControls
 *fdialogControls;   // Pointer to accelerator list
IThreadHandle
  fhandle;           // thread handle.  == id in PM
IWMQuitStack
  *fquitStack;       // stack of WM_QUIT requests from modal dialogs
#endif

#ifdef IC_PM
tib_t
 *pTIB;          // Pointer to TIB structure.
#endif

enum { varSize = 16 };

unsigned
  numVars;

// Class for storage of data that is associated with the started thread.
// The data is stored as IStrings and can be set and retrieved through the
// functions setVariable and variable of IStartedThread and IThread.
struct Assoc
  {
  Assoc ( );
 ~Assoc ( );
  IString
    key,
    value;
  } vars[ varSize ];

static IStartedThread
 **threads;      // Array of IStartedThread pointers.

static unsigned
  arraySize;     // Size of threads array.

static IStartedThread
 *cachedThread;  // Cached ptr.

static IThreadId
  cachedTID;     // Cached id.

static Boolean
  watching;      // Watching for non-IThread terminations.

static unsigned long
  fStartedThreadCount,
  cacheHits,
  cacheMisses;

static IPrivateResource
 *pResource,
 &lock ( );

static IThreadId
  fgPrimaryTID;    // primary thread ID

#ifdef IC_WIN
static HANDLE
  threadListChangeEvent,
  threadInitEvent,
  threadCreateSem;
static unsigned long
  tlsIndex;
static unsigned long
  fsProcessDetachInProgress;
                     // flag for when we are in DLL_PROCESS_DETACH
#endif
}; // Class IStartedThread


#ifdef IC_WIN
/*------------------------------------------------------------------------------
| ICurrentThreadData                                                           |
| Private data and implementation class for ICurrentThread.                    |
------------------------------------------------------------------------------*/
class ICurrentThreadData : public IBase {
typedef IBase
  Inherited;
public:
static Boolean
  translateAccelerator ( MSG& message );
#ifdef IC_WIN
static Boolean
  fsetMessageQueue;  // Flag to set message queue only once on Win32s
#endif
private:
  ICurrentThreadData ( );    // No need for real object yet.
 ~ICurrentThreadData ( );
}; // ICurrentThreadData

#endif


/*------------------------------------------------------------------------------
                        S T A T I C    O B J E C T S
------------------------------------------------------------------------------*/
#pragma data_seg(ICLNonConst)

#ifdef IC_WIN
Boolean
  fsetMessageQueue = false;  // Flag to set message queue only once on Win32s  //asmm.
#endif

unsigned
  IStartedThread::arraySize = 0;

unsigned long
  IThread::dfltStackSize = 32768;

Boolean
  IThread::dfltAutoInitGUI = -1;

long
  IThread::dfltQueueSize = 30;

ICurrentThread
 *IThread::pCurrent = 0;

IStartedThread
 **IStartedThread::threads = 0;

IStartedThread
 *IStartedThread::cachedThread = 0;

Boolean
  IStartedThread::watching = false;

IThreadId
  IStartedThread::cachedTID = 0;

unsigned long
  IStartedThread::fStartedThreadCount = 0,
  IStartedThread::cacheHits   = 0,
  IStartedThread::cacheMisses = 0;

IPrivateResource
 *IStartedThread::pResource = 0;

#ifdef IC_PM
IThreadId
  IStartedThread::fgPrimaryTID(1);
#endif

#ifdef IC_WIN
// We rely on static initialization of non thread local statics
// occurring in the context of the primary thread to initialize the following:
IThreadId
  IStartedThread::fgPrimaryTID( (IHandle::Value)GetCurrentThreadId() );

// event posted when a thread is added to our list to recycle any
// waitForAll events
HANDLE
  IStartedThread::threadListChangeEvent = 0;
// event posted by initThread to release the creating thread
HANDLE
  IStartedThread::threadInitEvent = ( (IPlatform::isWin32S()) ? 0 :
                             CreateEvent( 0,  // default security
                                          0,  // auto reset
                                          0,  // initially off
                                          0 ) ); // unnamed

HANDLE
  IStartedThread::threadCreateSem = ( (IPlatform::isWin32S()) ? 0 :
                             CreateSemaphore( 0,  // default security
                                              1,  // initially signaled
                                              1,  // maximum count
                                              0 ) ); // unnamed
unsigned long
  IStartedThread::tlsIndex = (unsigned long)-1;

unsigned long
  IStartedThread::fsProcessDetachInProgress(0);
                     // flag for when we are in DLL_PROCESS_DETACH

unsigned long dummyStack = ( (IPlatform::isWin32S()) ? 0 :
                              IThread::current().stackSize() );


#endif


#ifdef IC_WIN
/*------------------------------------------------------------------------------
| _setProcessDetachInProgress(void)                                            |
|                                                                              |
| There are ways in Win32 by which you can cause DLL_PROCESS_DETACH            |
| to occur on a different thread than the one which called DLL_PROCESS_ATTACH. |
| The application can cause this by calling ExitThread in the primary thread   |
| while other threads are active.                                              |
| In Win95, it appears that some of the system calls used by the C runtime     |
| cause this to occur.                                                         |
| We have defined a function _setProcessDetachInProgress in ithread.cpp        |
| which sets a flag to indicate that we are in DLL_PROCESS_DETACH.             |
| This function is called by _DLL_InitTerm when DLL_PROCESS_DETACH             |
| processing begins.                                                           |
|                                                                              |
------------------------------------------------------------------------------*/
void _setProcessDetachInProgress(void)
{
  IStartedThread::fsProcessDetachInProgress = 1;
  return;
}


/*------------------------------------------------------------------------------
| __IInitTerm                                                                  |
|   This function is used to process thread attach and detach events.          |
|   It is called by our DLL_InitTerm function when we are in a DLL             |
|   and by the C runtime's TLS callback in file tls.c when we are              |
|   in an .exe.                                                                |
|   In both cases "ulflag" takes the values                                    |
|     DLL_PROCESS_DETACH       0                                               |
|     DLL_PROCESS_ATTACH       1                                               |
|     DLL_THREAD_ATTACH        2                                               |
|     DLL_THREAD_DETACH        3                                               |
|                                                                              |
|   Note these are the same flag values that Dll_InitTerm takes.               |
|                                                                              |
|   This function must be packaged in ITHREAD.OBJ, so that it will             |
|   be linked in whenever this workaround is required.                         |
|                                                                              |
|   The PROCESS_ATTACH call occurs before any static initialization            |
|   and before main() is entered.                                              |
|                                                                              |
|   The PROCESS_DETACH call occurs after main() returns and after              |
|   all static destruction.                                                    |
|                                                                              |
------------------------------------------------------------------------------*/
void _Optlink __IInitTerm(unsigned long ulflag)
{
   switch (ulflag)
      {
      case  DLL_PROCESS_ATTACH:
         // We are called before static init.  Do nothing here...it is
         // handled in static constructors.
         break;
      case  DLL_PROCESS_DETACH:
         // We are called after static destructors.  Do nothing here...it is
         // handled in the destructors.
         break;
      case  DLL_THREAD_ATTACH:
#ifndef _WIN32S
         {
         IMODTRACE_DEVELOP("__IInitTerm - DLL_THREAD_ATTACH");
         if ( IStartedThread::currentHandle( false ).asUnsigned() == 0)
            {
            // have not seen this one yet.
            ITRACE_DEVELOP("__IInitTerm - new thread");
            // Force an object to be allocated in the list of threads if it is
            // not already.   This will set the currentHandle value and
            // initialize the id member of IStartedThread.
            IThread newThread(
               IThreadId( (IHandle::Value) GetCurrentThreadId() ) );
            }
         }
#endif
         break;
      case  DLL_THREAD_DETACH:
#ifndef _WIN32S
         {
         IMODTRACE_DEVELOP("__IInitTerm - DLL_THREAD_DETACH");
         // check if we have seen this thread
         if ( IStartedThread::currentHandle( false ).asUnsigned() != 0)
            {
            // We have seen it.  Check if entry exists and remove it.  We do not
            // want to do this to the primary thread or remove any threads
            // during process termination.  Doing so may cause the thread to
            // be reallocated during IOC static destructors.
            IHandle::Value id = (IHandle::Value)GetCurrentThreadId();
            if ( (IStartedThread::handleWithId( id ).asUnsigned() ) &&
                 (id != IStartedThread::fgPrimaryTID ) &&
                 (!IStartedThread::fsProcessDetachInProgress ) )
               {
               ITRACE_DEVELOP("__IInitTerm - removing thread");
               // entry exists
               IStartedThread* thread = IStartedThread::locate( id );
               if (thread)
                  {
                  thread->end();
                  thread->removeRef();
                  }
               }
            }
         }
#endif
         break;
       } // switch
    return;
}
#endif

/*------------------------------------------------------------------------------
| IThread::Cursor::timeStamp - Incremented each time a thread is started or    |
|                              stopped.  A cursor is valid if its created data |
|                              member is equal to this value.                  |
------------------------------------------------------------------------------*/
unsigned
  IThread::Cursor::timeStamp = 1;
#pragma data_seg()


/*------------------------------------------------------------------------------
| IStartedThread::IStartedThread                                               |
|                                                                              |
| Initialize all the members.  If a thread ID is given, it is used.            |
------------------------------------------------------------------------------*/
IStartedThread :: IStartedThread ( const IThreadId &tid )
  : id            ( tid ),
    hmq           ( 0 ),
    hab           ( 0 ),
    inArray       ( false ),
    started       ( ( tid.asUnsigned() ) ? true : false ),
    startedHere   ( false ),
    pmStartedHere ( false ),
    slot          ( 0 ),
#ifdef IC_PM
    pTIB          ( 0 ),
#endif
    autoInitPM    ( IThread::defaultAutoInitGUI() ),
    queueSize     ( IThread::defaultQueueSize() ),
    stkSize       ( IThread::defaultStackSize() ),
    fCount        ( ++IStartedThread::fStartedThreadCount ),
    inMsgLoop     ( 0 ),
    priorityClass ( IApplication::noChange ),
    priorityLevel ( -1 ),
    fnObj         ( 0 ),
    windowList    ( 0 ),
#ifdef IC_WIN
    fdialogControls ( 0 ),
    fhandle         ( 0 ),
    fquitStack      ( 0 ),
#endif
    numVars       ( 0 )
  {
  IMODTRACE_DEVELOP( IString( "IStartedThread::IStartedThread(" ) +
                     IString( tid.asUnsigned() ) +
                     IString( ")" ) );
  }

/*------------------------------------------------------------------------------
| IStartedThread::Assoc::Assoc                                                 |
|                                                                              |
| Constructor for Assoc class. This class allows storing data associated with  |
| the started thread.                                                          |
|                                                                              |
| Empty ctor to prevent generated static version.                              |
------------------------------------------------------------------------------*/
IStartedThread::Assoc :: Assoc ( )
  {
  }

/*------------------------------------------------------------------------------
| IStartedThread::Assoc::~Assoc                                                |
|                                                                              |
| Empty dtor to prevent generated static version.                              |
------------------------------------------------------------------------------*/
IStartedThread::Assoc :: ~Assoc ( )
  {
  }

/*------------------------------------------------------------------------------
| IStartedThread::~IStartedThread                                              |
|                                                                              |
| Remove this object from the static array.                                    |
------------------------------------------------------------------------------*/
IStartedThread :: ~IStartedThread ( )
  {
  IMODTRACE_DEVELOP( IString( "IStartedThread::~IStartedThread(" ) +
                     IString( this->id.asUnsigned() ) +
                     IString( ")" ) );
  delete windowList;
#ifdef IC_WIN
  delete fdialogControls;
  if (fquitStack)
     delete fquitStack;
#endif
  }

/*------------------------------------------------------------------------------
| IStartedThread::isStarted                                                    |
|                                                                              |
| Return whether the thread represented by this object is currently started.   |
|                                                                              |
| If the "started here" flag is set, then the thread was started via           |
| IStartedThread::initThread and we know it hasn't ended.                      |
|                                                                              |
| Otherwise, we check the thread ID via a call to DosWaitThread and if         |
| not OK, return false (and set the "started" flag appropriately).             |
------------------------------------------------------------------------------*/
Boolean IStartedThread :: isStarted ( )
  {
  Boolean result = true;
  if ( !this->startedHere )
    if ( this->started )
      // Verify it, though!
      result = this->started = IStartedThread::isTIDValid( this->id );
    else
      result = false;
  return result;
  }

/*------------------------------------------------------------------------------
| IStartedThread::isTIDValid                                                   |
|                                                                              |
| Return whether the thread represented by this object is currently started.   |
------------------------------------------------------------------------------*/
Boolean IStartedThread :: isTIDValid ( const IThreadId &id )
{
  IMODTRACE_DEVELOP( IString( "IStartedThread::isTIDValid" ));

#ifdef IC_PM
  IThreadId tid = id;
#endif

  Boolean
    result =
#ifdef IC_PM
             ( id == IStartedThread::fgPrimaryTID  ||
               id == IThread::currentId()          ||
               DosWaitThread( (PTID)&tid,
                              DCWW_NOWAIT ) == ERROR_THREAD_NOT_TERMINATED );
#endif
#ifdef IC_WIN
             ( id == IStartedThread::fgPrimaryTID  ||
               id == IThread::currentId()          ||
               ( !IPlatform::isWin32S()            &&
                 WaitForSingleObject( IStartedThread::handleWithId( id ),
                                      0 ) == WAIT_TIMEOUT) );
#endif
  return( result );
}

/*------------------------------------------------------------------------------
| IStartedThread::asDebugInfo                                                  |
------------------------------------------------------------------------------*/
IString IStartedThread :: asDebugInfo ( ) const
{
  IResourceLock
    lock( IStartedThread::lock() );
  IString result( "IStartedThread(" );
  result += this->Inherited::asDebugInfo();
  result += ",useCount=";
  result += IString( this->useCount() );
  result += ",id=";
  result += this->id.asDebugInfo();
  result += ",hmq=";
  result += this->hmq.asDebugInfo();
  result += ",hab=";
  result += this->hab.asDebugInfo();
  result += ",inArray=";
  result += IString( this->inArray );
  result += ",started=";
  result += IString( this->started );
  result += ",startedHere=";
  result += IString( this->startedHere );
  result += ",pmStartedHere=";
  result += IString( this->pmStartedHere );
  result += ",autoInitPM=";
  result += IString( this->autoInitPM );
  result += ",slot=";
  result += IString( this->slot );
#ifdef IC_PM
  result += ",pTIB=";
  result += IString( (unsigned long) this->pTIB );
#endif
  result += ",queueSize=";
  result += IString( this->queueSize );
  result += ",stkSize=";
  result += IString( this->stkSize );
  result += ",fCount=";
  result += IString( this->fCount );
  result += ",priorityClass=";
  result += IString( (int) this->priorityClass );
  result += ",priorityLevel=";
  result += IString( this->priorityLevel );
  result += ",fnObj=";
  result += IString( (unsigned long) this->fnObj );
  result += ",windowList=";
  result += IString( (unsigned long) this->windowList );
  result += ")";
  return( result );
}

/*------------------------------------------------------------------------------
| IStartedThread::classDebugInfo                                               |
------------------------------------------------------------------------------*/
IString IStartedThread :: classDebugInfo ( )
  {
  IResourceLock
    lock( IStartedThread::lock() );
  IString result( "IStartedThread::(cacheHits=" );
  result += IString( IStartedThread::cacheHits );
  result += ",cacheMisses=";
  result += IString( IStartedThread::cacheMisses );
  result += ",arraySize=";
  result += IString( IStartedThread::arraySize );
  result += ",threads=";
  result += IString( (unsigned long) IStartedThread::threads );
  result += "->[";
  IString comma;
  for ( unsigned i = 0;
        i < IStartedThread::arraySize;
        i++ )
    {
    result += comma;
    comma = ",";
    if ( IStartedThread::threads[i] )
      result += IStartedThread::threads[i]->asDebugInfo();
    else
      result += "0";
    }
  result += "])";
  return result;
  }

/*------------------------------------------------------------------------------
| IStartedThread::flushCache                                                   |
|                                                                              |
| Reset cache.                                                                 |
|                                                                              |
| Notes:                                                                       |
|   1. Must be called with resources protected.                                |
------------------------------------------------------------------------------*/
void IStartedThread :: flushCache ( )
  {
  ICritSec
    lock;
  IStartedThread::cachedTID    = 0;
  IStartedThread::cachedThread = 0;
  IThread::Cursor::timeStamp++; // Invalidate any existing cursors.
  }

/*------------------------------------------------------------------------------
| IStartedThread::cache                                                        |
|                                                                              |
| Set cached thread id/ptr.                                                    |
|                                                                              |
| Notes:                                                                       |
|   1. Must be called with resources protected.                                |
------------------------------------------------------------------------------*/
inline void IStartedThread :: cache ( const IThreadId &tid,
                                      IStartedThread  *thread )
  {
  IStartedThread::cachedTID    = tid;
  IStartedThread::cachedThread = thread;
  }

/*------------------------------------------------------------------------------
| IStartedThread::checkCacheFor                                                |
|                                                                              |
| Check cached thread id/ptr.                                                  |
|                                                                              |
| Notes:                                                                       |
|   1. Must be called with resources protected.                                |
------------------------------------------------------------------------------*/
inline IStartedThread *IStartedThread :: checkCacheFor ( const IThreadId &tid )
  {
  IStartedThread *result = 0;
  if ( tid == IStartedThread::cachedTID )
    {
    result = IStartedThread::cachedThread;
    IStartedThread::cacheHits++;
    }
  else
    IStartedThread::cacheMisses++;
  return result;
  }

/*------------------------------------------------------------------------------
| IStartedThread::add                                                          |
|                                                                              |
| Search the static array for a free slot and add the argument IStartedThread  |
| object there.  If there is no room, allocate a larger array.                 |                                             |
|                                                                              |
| Notes:                                                                       |
|   1. Access to the static data is protected via a critical section.  Most    |
|      other IStartedThread methods that utilize this data do likewise.        |
|   2. This function is intended to be called by IThread::start().             |
------------------------------------------------------------------------------*/
void IStartedThread :: add ( IStartedThread &aThread )
  {
  IMODTRACE_DEVELOP( IString( "IStartedThread::add(" ) +
                     IString( aThread.id.asUnsigned() ) +
                     IString( ")" ) );
  // Lock static info.
  IResourceLock
    lock( IStartedThread::lock() );
  // If flag set, simply return.
  if ( aThread.inArray )
    return;
  // Search for free slot.
  for ( unsigned i = 0;
        i < IStartedThread::arraySize
        &&
        IStartedThread::threads[i];
        i++ ) {}
  if ( i == IStartedThread::arraySize )
    {
    // Extend array size.
    unsigned n = 2 * IStartedThread::arraySize + 1;
    // Allocate new array.
    IStartedThread **t = new IStartedThread*[ n ];
    // Copy old portion.
    while ( i-- )
      t[i] = IStartedThread::threads[ i ];
    // Zero remainder.
    i = IStartedThread::arraySize;
    while ( i < n )
      t[i++] = 0;
    // Free old array.
    delete [] IStartedThread::threads;
    // Update static info.
    IStartedThread::threads   = t;
    i                         = IStartedThread::arraySize;
    IStartedThread::arraySize = n;
    }
  // Set i-th slot to point to aThread.
  IStartedThread::threads[i] = &aThread;
  aThread.inArray = true;
  aThread.slot    = i;
  aThread.started = true;
  aThread.addRef();
#ifdef IC_WIN
  // Signal we have modified the thread list to the waitForThreadEnd function
  // Skip for win32s (as a precuation: IStartThread::add not called in 32s)
  if ( (IStartedThread::threadListChangeEvent != 0) &&
       (!IPlatform::isWin32S()) )
     SetEvent( IStartedThread::threadListChangeEvent );
#endif
  }

/*------------------------------------------------------------------------------
| IStartedThread::remove                                                       |
|                                                                              |
| Search the static array for the argument IStartedThread object and reset the |
| slot to zero if found.                                                       |
|                                                                              |
| Notes:                                                                       |
|   1. Access to the static data is protected via a critical section.  Most    |
|      other IStartedThread methods that utilize this data do likewise.        |
------------------------------------------------------------------------------*/
void IStartedThread :: remove ( IStartedThread &aThread )
  {
  IMODTRACE_DEVELOP( IString( "IStartedThread::remove(" ) +
                     IString( aThread.id.asUnsigned() ) +
                     IString( " - " ) +
                     IString::d2x( IString( (int)&aThread )) +
                     IString( ")" ) );
  IResourceLock
    lock( IStartedThread::lock() );
  if ( aThread.inArray )
    {
    IStartedThread::threads[ aThread.slot ] = 0;
    aThread.inArray = false;
    aThread.slot    = 0;
    aThread.removeRef();
    IStartedThread::flushCache();
    }
  }

/*------------------------------------------------------------------------------
| IStartedThread::locate                                                       |
|                                                                              |
| Search through the static array of started threads and return a pointer to   |
| the one with the argument threadID.                                          |
|                                                                              |
| If not found, then determine if the thread ID is valid and if so, allocate a |
| new IStartedThread and insert it into the array.                             |
|                                                                              |
| Notes:                                                                       |
|   1. The use count of the returned IStartedThread will be incremented        |
|      by one (or initialized to one if the object is allocated here).         |
|      The caller must ensure it gets decremented (IThread ensure this         |
|      by calling removeRef() at destruction).                                 |
|   2. Care is taken to ensure the static thread info is stable while          |
|      being examined/updated.                                                 |
|   3. This function throws an exception if the argument threadID is           |
|      not a valid OS/2 thread.                                                |
------------------------------------------------------------------------------*/
IStartedThread *IStartedThread :: locate ( const IThreadId &id )
  {
  IMODTRACE_ALL( IString( "IStartedThread::locate(" ) +
                     IString( id.asUnsigned() ) +
                     IString( ")" ) );
  IThreadId tid = id;
  IStartedThread
   *result = 0;
  {
  // Check cache first:
  ICritSec
    lock;
  result = IStartedThread::checkCacheFor( id );
  if ( result )
    result -> addRef();
  }
  // If not found, search array for this thread...
  if ( !result )
    {
    // Lock static info.
    IResourceLock
      lock( IStartedThread::lock() );
    for ( unsigned i = 0;

          i < IStartedThread::arraySize
          &&
          ( IStartedThread::threads[i] == 0
            ||
            IStartedThread::threads[i]->id != id );

          i++ ) {}

    if ( i == IStartedThread::arraySize )
      // Thread not found, see if it is a real thread.
      // In Windows, the thread must be the current thread for
      // this logic to work correctly.  We need to get the handle.
      // To avoid the exception that will result if this is not the
      // case, use the second argument on the IThreadId constructor.
#ifdef IC_WIN
      if ( tid.asUnsigned() == GetCurrentThreadId() )
#else
      if ( IStartedThread::isTIDValid( tid ) )
#endif
        // This thread looks OK.
        {
        // Allocate new IStartedThread for this TID.
        result = new IStartedThread( id );
#ifdef IC_WIN
        result->fhandle = IStartedThread::currentHandle( true );
#endif
        // Insert into array.
        IStartedThread::add( *result );
        // Cache this thread.
        IStartedThread::cache( id, result );
        // Watch for when it ends...
        if ( tid != IStartedThread::fgPrimaryTID )
          IStartedThread::watch();
        }
      else
        {
        ITHROWLIBRARYERROR( IC_INVALID_THREAD_ID,
                            IBaseErrorInfo::invalidParameter,
                            IException::recoverable );
        }
    else
      // Found a match.
      {
      result = IStartedThread::threads[i];
      result -> addRef();
      IStartedThread::cache( id, result );
      }
    }
  return result;
  }

#ifdef IC_WIN
/*------------------------------------------------------------------------------
| IStartedThread::waitForThreadEnd                                             |
|   Waits for one or all known threads to end.                                 |
|   A single thread may be specified to be EXCLUDED from the list...           |
|   typically this would be the current thread.                                |
|   throwOnError true means an exception is thrown if something wrong          |
|   happens.  Otherwise,  IThreadId 0 is returned in an error case.            |
------------------------------------------------------------------------------*/
IThreadId IStartedThread::waitForThreadEnd(Boolean          waitForAll,
                                           const IThreadId& excludedThread,
                                           Boolean          throwOnError )
{
   IMODTRACE_ALL("IStartedThread::waitForThreadEnd");

   if (IPlatform::isWin32S())
   {
    ITRACE_DEVELOP("Win32s: IStartedThread::waitForThreadEnd has no effect, returns 0");
    return IThreadId(0);
   }

   if (IStartedThread::threadListChangeEvent == 0)
      {
      IStartedThread::threadListChangeEvent = CreateEvent(
                              0,                // default security
                              1,                // manual reset
                              0,                // initially off
                              0 );              // unnamed.
      // could not create event.
      if (IStartedThread::threadListChangeEvent == 0)
         {
         ITHROWGUIERROR2( "CreateEvent",
                           IBaseErrorInfo::accessError,
                           IException::recoverable );
         }
      }


   HANDLE     postedHandle = IStartedThread::threadListChangeEvent;
   IThreadId  result = 0;

   // We loop until an event other than the threadListChangeEvent is
   // posted indicating termination.  The threadListChangeEvent tells us
   // that the thread list has changed and we need to inspect it.
   while (postedHandle == IStartedThread::threadListChangeEvent)
      {
      ITRACE_DEVELOP("Setting up wait");
      HANDLE*       handleList = 0;
      IThreadId*    idList = 0;
      unsigned long handleCount = 0;

      // lock the thread list for a snapshot of it
         {
         IResourceLock
           lock( IStartedThread::lock() );

         handleList = new HANDLE[ IStartedThread::arraySize + 1];
         idList = new IThreadId[ IStartedThread::arraySize + 1];


         // reset the thread list changed event.  We only care about these
         // if they occur while we are asleep.
         ResetEvent( IStartedThread::threadListChangeEvent );

         // copy thread IDs to the output list.
         for ( unsigned i = 0;
               i < IStartedThread::arraySize;
               i++ )
            {
#ifdef IC_TRACE_DEVELOP
            if (IStartedThread::threads[i] != 0)
              {
              ITRACE_DEVELOP(IString("IStartedThread ") + IString(i) +
                         IString(" id=") +
                         IString(IStartedThread::threads[i]->id.asUnsigned()) +
                         IString(" handle=") +
                         IString(IStartedThread::threads[i]->handle().asUnsigned()) );
              }
#endif
            if ((IStartedThread::threads[i] != 0) &&
                (IStartedThread::threads[i]->id != excludedThread ) )
               {
               handleList[handleCount] = IStartedThread::threads[i]->handle();
               idList[handleCount] = IStartedThread::threads[i]->id;
               ITRACE_DEVELOP(IString("IStartedThread handleList ") +
                              IString(handleCount) +
                              IString(" id=") +
                              IString(idList[handleCount].asUnsigned()) +
                              IString(" handle=") +
                              IString((unsigned long)handleList[handleCount]) );
               handleCount++;
               }
            }
         } // unlock list

      // The last list entry is the thread list changed event.
      // That means that we will only see it if others are NOT posted.
      // We don't use this if we are waiting for all threads to end,
      // since it may not be posted in this case and we would wait forever
      // on it.
      if (!waitForAll)
         {
         handleList[handleCount] = IStartedThread::threadListChangeEvent;
         idList[handleCount] = 0;
         ITRACE_DEVELOP(IString("IStartedThread threadListChangeEvent") +
                        IString(handleCount) +
                        IString(" id=") +
                        IString(idList[handleCount].asUnsigned()) +
                        IString(" handle=") +
                        IString((unsigned long)handleList[handleCount]) );
         handleCount++;
         }

      // If there are objects to wait on, wait on them.  The objects
      // are either thread objects which are posted on termination or
      // the thread list changed event.
      if (handleCount)
         {
         // Wait for something to happen.
         unsigned long
            rc = WaitForMultipleObjects(
                    handleCount,               // count of handles
                    handleList,                // array of handles
                    waitForAll,                // wait for all flag
                    (DWORD)INFINITE ) ;        // no timeout

         if (rc == WAIT_FAILED)
            {
            if (throwOnError)
               ITHROWGUIERROR2( "WaitForMultipleObjects",
                                 IBaseErrorInfo::accessError,
                                 IException::recoverable );
            postedHandle = 0;      // make sure we exit
            }
         else
            {
            postedHandle = handleList[rc - WAIT_OBJECT_0];
            result = idList[rc - WAIT_OBJECT_0];
            }
         }
      else
         {
         // No threads to wait on.  Return immediately.
         postedHandle = 0;      // make sure we exit
         }

      // clean up
      delete [] handleList;
      delete [] idList;

      // If we are waiting for all to end, check the state of the
      // threadListChangeEvent object.  If it is set new threads
      // have been created while we were waiting.
      if (waitForAll)
         {
         if (WAIT_OBJECT_0 == WaitForSingleObject(
                                 IStartedThread::threadListChangeEvent, 0) )
            {
            // Thread list changed.  Loop to check it again.
            postedHandle = IStartedThread::threadListChangeEvent;
            result = 0;
            }
         }      // if wait for all

      }  // while

   return result;
}

/*------------------------------------------------------------------------------
| IStartedThread::currentHandle                                                |
| Queries the handle for the current thread.  By default (create true),        |
| a new handle is created if one is not already allocated.                     |
------------------------------------------------------------------------------*/
IThreadHandle IStartedThread:: currentHandle( Boolean create )
{
  IMODTRACE_ALL("IStartedThread::currentHandle");
  if (IPlatform::isWin32S())
  {
    ITRACE_DEVELOP("Win32s: currentHandle always returns GetCurrentThreadId()");
    return IThreadHandle((IHandle::Value)GetCurrentThreadId());
  }
  else
  {
    if (IStartedThread::tlsIndex == (unsigned long)-1)
       {
       // Initialize the thread local storage index for the process
       IStartedThread::tlsIndex = TlsAlloc();
       if (IStartedThread::tlsIndex == (unsigned long)-1)
          {
          ITHROWGUIERROR2("TlsAlloc",
                          IBaseErrorInfo::accessError,
                          IException::recoverable );
          }
       ITRACE_DEVELOP(IString("tlsIndex=") + IString(tlsIndex));
       }

    // Get the stored handle.  Note that the uninitialized value is set
    // to 0 by the system.
    HANDLE tlsThreadHandle = TlsGetValue( IStartedThread::tlsIndex );
    ITRACE_ALL(IString("tlsThreadHandle=") + IString((unsigned long)tlsThreadHandle));

    // if the handle has not been set yet create one from the
    // pseudohandle given by GetCurrentThread().  Store in thread local
    // storage so we only duplicate the handle once per thread.
    if ( (tlsThreadHandle == 0) && create )
       {
       HANDLE prochand = GetCurrentProcess();
       DuplicateHandle(
          prochand,           //HANDLE hSourceProcessHandle,
          GetCurrentThread(), //HANDLE hSourceHandle,
          prochand,           //HANDLE hTargetProcessHandle,
          &tlsThreadHandle,   //LPHANDLE lpTargetHandle,
          0,                  //DWORD dwDesiredAccess,
          0,                  //BOOL bInheritHandle,
          DUPLICATE_SAME_ACCESS );   //DWORD dwOptions

       // Store the duplicated handle in thread local storage.  Future
       // references use the stored value
       if (!TlsSetValue( IStartedThread::tlsIndex, tlsThreadHandle ))
          {
          ITHROWGUIERROR2("TlsSetValue",
                          IBaseErrorInfo::accessError,
                          IException::recoverable );
          }
       ITRACE_DEVELOP(IString("new tlsThreadHandle=") +
                      IString((unsigned long)tlsThreadHandle) +
                      IString(" id=") + IString(GetCurrentThreadId()) );
       }
    return IThreadHandle( tlsThreadHandle );
  }
}
#endif

/*------------------------------------------------------------------------------
| IStartedThread::idWithHandle                                                 |
------------------------------------------------------------------------------*/
IThreadId  IStartedThread::idWithHandle   (const IThreadHandle& handle )
{
   // In Windows, a thread handle and a thread id are 2 different things,
   // with no API support to convert between them.   In other environments
   // a thread handle and thread id are the same.
   IMODTRACE_DEVELOP("IStartedThread::idWithHandle");

#ifdef IC_WIN
   IThreadId result = 0;

   // lock the static thread list
   IResourceLock
     lock( IStartedThread::lock() );

   // search for a matching handle
   for ( unsigned i = 0;
         (i < IStartedThread::arraySize) && (!result);
         i++ )
      {
      if ((IStartedThread::threads[i] != 0) &&
          (IStartedThread::threads[i]->handle() == handle ) )
         {
         // trickery to try to obtain the id value
         if ( IStartedThread::threads[i]->id == 0 )
            {
            IStartedThread::threads[i]->id =
                                        (IHandle::Value)GetCurrentThreadId();
            }
         result = IStartedThread::threads[i]->id;
         }
      }
   return result;
#else
   return handle;
#endif
}

/*------------------------------------------------------------------------------
| IStartedThread::handleWithId                                                 |
------------------------------------------------------------------------------*/
IThreadHandle  IStartedThread::handleWithId   (const IThreadId& tid )
{
   // In Windows, a thread handle and a thread id are 2 different things,
   // with no API support to convert between them.   In other environments
   // a thread handle and thread id are the same.
   IMODTRACE_DEVELOP("IStartedThread::handleWithId");

#ifdef IC_WIN
   IThreadHandle result = 0;

   // lock the static thread list
   IResourceLock
     lock( IStartedThread::lock() );

   // search for a matching id
   for ( unsigned i = 0;
         (i < IStartedThread::arraySize) && (!result);
         i++ )
      {
      if ((IStartedThread::threads[i] != 0) &&
          (IStartedThread::threads[i]->id == tid ) )
         {
         // trickery to try to obtain the handle value
         if ((IStartedThread::threads[i]->handle() == 0) &&
             (tid == (IHandle::Value)GetCurrentThreadId() ) )
            {
            // we are executing in the thread...one of the 2 places we
            // can get the handle.
            IStartedThread::threads[i]->fhandle =
               IStartedThread::currentHandle( true );
            }
         result = IStartedThread::threads[i]->handle();
         }
      }
   return result;

#else
   return tid;
#endif
}

/*------------------------------------------------------------------------------
| initThread                                                                   |
------------------------------------------------------------------------------*/
extern "C" void initThread ( void *aThread )
  {
  IStartedThread::initThread( (IStartedThread *)aThread );
  }

/*------------------------------------------------------------------------------
| IStartedThread::initThread                                                   |
|                                                                              |
| Implementation:                                                              |
|   Dispatch the run() virtual function of the IThreadFn object pointed        |
|   to by the argument.                                                        |
|                                                                              |
|   If the "auto init" flag is set, then we initialize PM (and terminate on    |
|   thread completion).                                                        |
|                                                                              |
| Notes:                                                                       |
|   1. The argument's use count is incremented for the duration of this        |
|      function to prevent it from being deleted due to IThread destructors    |
|      being called in other threads.                                          |
------------------------------------------------------------------------------*/
void IStartedThread :: initThread ( IStartedThread *aThread )
  {
#ifdef IC_WIN
  // The IThreadAttachPrivate constructor will have run on this thread
  // already and allocated an IStartedThread object for it since the
  // thread ID and handle are not yet initialized. That means that
  // there may be 2 IStartedThread objects for this thread.  We want
  // to use the IStartedThread object passed in instead of that one, since
  // the one passed may contain various settings we need to use.
  if ( IStartedThread::handleWithId( IThread::currentId() ).asUnsigned() )
     {
     IStartedThread* junk = IStartedThread::locate( IThread::currentId() );
     IStartedThread::remove( *junk );
     junk->removeRef();
     }
#endif

  aThread->id = IThread::currentId();

#ifdef IC_WIN

  // Initialize the handle in the IStartedThread
  aThread->fhandle = IStartedThread::currentHandle( true );
  // Now that handle is present, allow creator to proceed.
  if (IPlatform::isWin32S())
    ITRACE_DEVELOP("Win32s: IStartedThread::initThread does not SetEvent()");
  else
    SetEvent( IStartedThread::threadInitEvent );
#endif
#ifdef IC_PM
  aThread->tib(); // Set pTIB member.
#endif

  IMODTRACE_DEVELOP( IString( "IStartedThread::initThread(" ) +
                     IString( aThread->id.asUnsigned() ) +
                     IString( " - " ) +
                     IString::d2x( IString( (int)aThread )) +
                     IString( ")" ) );

  aThread->pmStartedHere = aThread->autoInitPM;
  // Initialize GUI if we need to...
  if ( aThread->pmStartedHere )
    IThread::current().initializeGUI( aThread->queueSize );
  // If priority is other than "default", set it via API...
  if ( aThread->priorityClass != IApplication::noChange
       ||
       aThread->priorityLevel != -1 )
    IThread::current().setPriority( aThread->priorityClass,
                                    aThread->priorityLevel );
  // Dispatch user function...
  aThread->fnObj->run();
  // Terminate GUI if we initialized it...
  if ( aThread->pmStartedHere )
    IThread::current().terminateGUI();
  // Terminate thread...
  aThread->end();
  }

/*------------------------------------------------------------------------------
| IStartedThread::end                                                          |
|                                                                              |
| Reset the object to an "unstarted" state.  This function can only be called  |
| if this->startedHere is true.                                                |
------------------------------------------------------------------------------*/
void IStartedThread :: end ( )
  {
  IStartedThread::flushCache();
  this->startedHere = this->started = false;
  if ( this->fnObj )
    {
    this->fnObj->removeRef();
    this->fnObj = 0;
    }
#ifdef IC_WIN
  if ( this->id.asUnsigned() == GetCurrentThreadId() )
     {
     // clear the tls stored handle value
     if (!(IPlatform::isWin32S()) &&
         !TlsSetValue( IStartedThread::tlsIndex, 0))
        {
        ITHROWGUIERROR2("TlsSetValue",
                        IBaseErrorInfo::accessError,
                        IException::recoverable );
        }
     }
  if ( !CloseHandle( this->fhandle ) )
  {
    ITHROWSYSTEMERROR( GetLastError(),
                       "CloseHandle",
                       IBaseErrorInfo::accessError,
                       IException::recoverable );
  }

  this->fhandle = 0;
#endif
#ifdef IC_PM
  this->pTIB = 0;
#endif
  IStartedThread::remove( *this );
  }

/*------------------------------------------------------------------------------
| IStartedThread::tib                                                          |
|                                                                              |
| Return pointer to system TIB structure.  If zero on entry, then this member  |
| is set.  This function can only be called for the current thread (if not,    |
| then 0 may be returned).                                                     |
------------------------------------------------------------------------------*/
#ifdef IC_PM
tib_t *IStartedThread :: tib ( )
  {
  // In non-OS/2 environments return the current value (0)
  if ( !this->pTIB )
    {
    PTIB
      ptib;
    PPIB
      ppib;
    DosGetInfoBlocks( &ptib, &ppib );
    if ( this->id == ptib->tib_ptib2->tib2_ultid )
      this->pTIB = ptib;
    }
  return this->pTIB;
  }
#endif

/*------------------------------------------------------------------------------
| IStartedThread::variable                                                     |
|                                                                              |
| Examine each element in the vars array.  If one with a matching key is found,|
| return the value associated with it, otherwise return null.                  |
------------------------------------------------------------------------------*/
IString IStartedThread :: variable ( const IString &key ) const
  {
  unsigned
    index = 0;

  while ( index < this->numVars )
    if ( vars[ index ].key == key )
      return vars[ index ].value;
    else
      index++;

  return IString();
  }

/*------------------------------------------------------------------------------
| IStartedThread::setVariable                                                  |
|                                                                              |
| Search for a matching key.  If found, set the associated value to the new    |
| one. If no matching key is found and there is room in the array, fill it in  |
| with the argument key/value pair.                                            |
------------------------------------------------------------------------------*/
IStartedThread &IStartedThread :: setVariable ( const IString &key,
                                                const IString &value )
  {
  unsigned
    index = 0;
  {
  IResourceLock
    lock( IStartedThread::lock() );
  while ( index < this->numVars )
    if ( vars[ index ].key == key )
      {
      vars[ index ].value = value;
      break;
      }
    else
      index++;
  if ( index == this->numVars
       &&
       index < varSize )
    {
    vars[ index ].key = key;
    vars[ index ].value = value;
    this->numVars++;
    }
  else if (index >= varSize)
    {
    ITHROWLIBRARYERROR( IC_THREAD_VARIABLE_LIMIT_EXCEEDED,
                        IBaseErrorInfo::invalidRequest,
                        IException::recoverable );
    }
  }
  return *this;
  }

/*------------------------------------------------------------------------------
| IStartedThread::lock                                                         |
|                                                                              |
| If the pointed-to private resource hasn't been initialized, do it.           |
|                                                                              |
| Return the private resource.                                                 |
------------------------------------------------------------------------------*/
IPrivateResource &IStartedThread :: lock ( )
  {
  if ( !IStartedThread::pResource )
    IStartedThread::pResource = new IPrivateResource;
  return *IStartedThread::pResource;
  }

/*------------------------------------------------------------------------------
| IThread::IThread                                                             |
|                                                                              |
| The role of the IThread constructors is twofold:                             |
|   1. They must locate the appropriate IStartedThread instance,               |
|      creating one if necessary, set the "thread" member to point             |
|      to this instance, and bump the IStartedThread use count.                |
|   2. Start the OS/2 thread (if necessary).                                   |
------------------------------------------------------------------------------*/
IThread :: IThread ( const IThreadId& tid,
                     const IThreadHandle& handle  )
  : thread( 0 ),
    fThreadData( 0 )
  {
  IMODTRACE_ALL("IThread::IThread");

#ifdef IC_WIN
  if  (handle != IThreadHandle::noHandle)
     {
     // Handle given.  Assume valid and use it.
     if (tid == IStartedThread::idWithHandle( handle ) )
        {
        // already known ... set pointer
        thread = IStartedThread::locate( tid );
        }
     else
        {
        // new thread to us...fully qualified.  Add to list
        thread = new IStartedThread( tid );
        thread->fhandle = handle;
        IStartedThread::add(*thread );
        // Watch for when it ends...
        if ( tid != IStartedThread::fgPrimaryTID )
          IStartedThread::watch();
        }
     }
  else
     {
     // handle is default value.  The thread must already be known to
     // us or be the current thread.  locate() checks and will throw
     // if this is not the case.
     thread = IStartedThread::locate( tid );
     }
#else
  // Dont need to worry about the handle ... ignore it.
  thread = IStartedThread::locate( tid );
#endif
  }


IThread :: IThread ( )
  : thread( IThread::newStartedThread() ),
    fThreadData( 0 )
  {
  }

IThread :: IThread ( const IThread &aThread )
  : thread( aThread.startedThread() ),
    fThreadData( 0 )
  {
  this->thread->addRef();
  }

/*------------------------------------------------------------------------------
| IThread::newStartedThread                                                    |
------------------------------------------------------------------------------*/
IStartedThread *IThread::newStartedThread ( )
  {
  return new IStartedThread;
  }

/*------------------------------------------------------------------------------
| IThread::asString                                                            |
------------------------------------------------------------------------------*/
IString IThread :: asString ( ) const
  {
  IString result( "IThread(" );
  result += IString( this->id().asUnsigned() );
  result += ")";
  return result;
  }

/*------------------------------------------------------------------------------
| IThread::asDebugInfo                                                         |
------------------------------------------------------------------------------*/
IString IThread :: asDebugInfo ( ) const
  {
  IString result( "IThread(" );
  result += Inherited::asDebugInfo();
  result += ",thread=";
  result += IString( (unsigned long) this->startedThread() );
  result += "->";
  result += this->startedThread()->asDebugInfo();
  result += ")";
  return result;
  }

/*------------------------------------------------------------------------------
| IThread::~IThread                                                            |
|                                                                              |
| The destructor must remove the reference to the associated IStartedThread.   |                                                         |
------------------------------------------------------------------------------*/
IThread :: ~IThread ( )
  {
  if ( this->thread )
     this->thread->removeRef();
//delete fThreadData; when it is defined.
  }

/*------------------------------------------------------------------------------
| IThread::operator =                                                          |
|                                                                              |
| Attach this thread to aThread's associated started thread.                   |
------------------------------------------------------------------------------*/
IThread &IThread::operator = ( const IThread &aThread )
  {
  if ( aThread.startedThread() )
     aThread.startedThread()->addRef();
  if ( this->thread )
     this->thread->removeRef();
  this->thread = aThread.startedThread();
  return *this;
  }

/*------------------------------------------------------------------------------
| IThread::startedThread                                                       |
|                                                                              |
| Returns the thread member.                                                   |
------------------------------------------------------------------------------*/
IStartedThread *IThread :: startedThread ( ) const
  {
  return this->thread;
  }

/*------------------------------------------------------------------------------
| IThread::currentId                                                           |
|                                                                              |
| Construct an IThreadId from the system info blocks.                          |
------------------------------------------------------------------------------*/
IThreadId IThread :: currentId ( )
  {
#ifdef IC_PM
  PTIB
    ptib;
  PPIB
    ppib;
  DosGetInfoBlocks( &ptib, &ppib );
  return IThreadId( ptib->tib_ptib2->tib2_ultid );
#endif
#ifdef IC_WIN
  return IThreadId( (IHandle::Value)GetCurrentThreadId());
#endif
  }



/*------------------------------------------------------------------------------
| IThread::currentHandle                                                       |
------------------------------------------------------------------------------*/
IThreadHandle IThread :: currentHandle ( )
  {
#ifdef IC_WIN
  // We need to get the handle to the thread, since that is what we
  // need for all the thread and synchronization functions.
  return IStartedThread::currentHandle( true );
#else
  // the threadid is the handle
  return IThread::currentId();
#endif

  }


/*------------------------------------------------------------------------------
| IThread::prepareToStart                                                      |
|                                                                              |
| Check that thread is not currently started (if so, throw an exception).      |
| Otherwise, initialize the IStartedThread object appropriately and            |
| return its address.                                                          |
------------------------------------------------------------------------------*/
IStartedThread *IThread :: prepareToStart (
                                    const IReference<IThreadFn> &aFnObjRef,
                                    Boolean                      autoInitGUI )
  {
  IMODTRACE_DEVELOP( "IThread::prepareToStart" );

  IStartedThread
   *p = this->startedThread();

  // If thread is already started, throw exception.
  if ( p->isStarted() )
    {
    ITHROWLIBRARYERROR( IC_THREAD_STARTED,
                        IBaseErrorInfo::invalidRequest,
                        IException::recoverable );
    }

  if ( !p->inArray )
    IStartedThread::add( *p );

  p->startedHere = true;
  p->fnObj       = &(*aFnObjRef);
  p->fnObj->addRef();
  p->autoInitPM  = autoInitGUI;
#ifdef IC_WIN
  // Acquire the threadCreateSem semaphore.
  // This semaphore is owned during the thread creation process until such
  // time as the created thread has run initThread.  It signals the creator
  // who then frees this semaphore.
  if (IPlatform::isWin32S())
    // No threads are started in Win32s, no need to acquire semaphore.
    ITRACE_DEVELOP("Win32s: IThread::prepareToStart does not acquire semaphore");
  else
    if (WAIT_FAILED == WaitForSingleObject( IStartedThread::threadCreateSem,
                                          (DWORD)INFINITE ))
     {
     ITHROWGUIERROR2( "WaitForSingleObject",
                      IBaseErrorInfo::accessError,
                      IException::recoverable );
     }
#endif

  return p;
  }

/*------------------------------------------------------------------------------
| IThread::threadStarted                                                       |
|                                                                              |
| Check the return code from beginthread (see ithreads.cpp).  If an error      |
| occurred, throw exception, else set thread id.                               |
------------------------------------------------------------------------------*/
void IThread :: threadStarted( IStartedThread *p, int rc )
{
  IMODTRACE_DEVELOP("IThread::threadStarted");

  if ( rc == -1 )
  {
    p -> end();
    ITHROWSYSTEMERROR( rc,
                       "_beginthread",
                       IBaseErrorInfo::outOfSystemResource,
                       IException::recoverable );
  }
  else
  {
    ITRACE_DEVELOP( IString("Thread ID=") + IString(rc) );
#ifdef IC_WIN
    // Wait on some initialization to occur in the started thread
    // Specifically, we want IStartedThread::initThread to run in the
    // context of the new thread and store the handle.
  if (IPlatform::isWin32S())
   // No threads are started in Win32s, no need to acquire
   //or release semaphores
   ITRACE_DEVELOP("Win32s: IThread::threadStarted does not acquire semaphore");
  else
   {
    if (WAIT_FAILED == WaitForSingleObject( IStartedThread::threadInitEvent,
                                            (DWORD)INFINITE ))
    {
      // make sure we release
      ReleaseSemaphore( IStartedThread::threadCreateSem, 1, 0 );
      ITHROWGUIERROR2( "WaitForSingleObject",
                       IBaseErrorInfo::accessError,
                       IException::recoverable );
    }
    // Now we know that thread creation is complete.  Release the
    // semaphore to allow other thread creations to occur.
    ReleaseSemaphore( IStartedThread::threadCreateSem, 1, 0 );
   }
#else
    p->id = (IHandle::Value)rc;
#endif
  }
}

/*------------------------------------------------------------------------------
| IThread::stop                                                                |
|                                                                              |
| Issue DosKillThread.  If it fails, throw an exception.                       |
------------------------------------------------------------------------------*/
void IThread :: stop ( )
{
  IMODTRACE_DEVELOP("IThread::stop");

#ifdef IC_PM
  unsigned long rc = DosKillThread( this->id() );
  if ( rc != NO_ERROR )
  {
    ITHROWSYSTEMERROR( rc,
                       "DosKillThread",
                       IBaseErrorInfo::accessError,
                       IException::recoverable );
  }
#endif
#ifdef IC_WIN
  if (IPlatform::isWin32S())
    ITRACE_DEVELOP("Win32s: stop() does not call TerminateThread");

  //0 might not be the correct exitcode in all cases
  if ( (!IPlatform::isWin32S()) && (!TerminateThread(this->handle(), 0)) )
  {
    ITHROWGUIERROR2( "TerminateThread",
                     IBaseErrorInfo::accessError,
                     IException::recoverable );
  }
#endif
  else
    this->startedThread() -> end();
}

/*------------------------------------------------------------------------------
| ICurrentThread::~ICurrentThread                                              |
|                                                                              |
| Empty dtor to prevent generated static version.                              |
------------------------------------------------------------------------------*/
ICurrentThread :: ~ICurrentThread ( )
{
}

/*------------------------------------------------------------------------------
| ICurrentThread::suspend                                                      |
|                                                                              |
| Issue DosSuspendThread.  If it fails, throw an exception.                    |
------------------------------------------------------------------------------*/
void ICurrentThread :: suspend ( )
{
  if ( this->isGUIInitialized() )
  {
    ITHROWLIBRARYERROR( IC_SUSPEND_PM_THREAD,
                        IBaseErrorInfo::invalidRequest,
                        IException::recoverable );
  }
  else
    this->IThread::suspend();
}

/*------------------------------------------------------------------------------
| IThread::suspend                                                             |
|                                                                              |
| Issue DosSuspendThread.  If it fails, throw an exception.                    |
------------------------------------------------------------------------------*/
void IThread :: suspend ( )
{
  IMODTRACE_DEVELOP("IThread::suspend");

#ifdef IC_PM
  unsigned long rc = DosSuspendThread( this->id() );
  if ( rc != NO_ERROR )
  {
    ITHROWSYSTEMERROR( rc,
                       "DosSuspendThread",
                       IBaseErrorInfo::accessError,
                       IException::recoverable );
  }
#endif
#ifdef IC_WIN
  if (IPlatform::isWin32S())
  {
    ITRACE_DEVELOP("Win32s: suspend() has no effect");
    return;
  }
  if ( WAIT_FAILED == SuspendThread( this->handle() ) )
  {
    ITHROWGUIERROR2( "SuspendThread",
                     IBaseErrorInfo::accessError,
                     IException::recoverable );
  }
#endif
}

/*------------------------------------------------------------------------------
| IThread::resume                                                              |
|                                                                              |
| Issue DosResumeThread.  If it fails, throw an exception.                     |
------------------------------------------------------------------------------*/
void IThread :: resume ( )
{
  IMODTRACE_DEVELOP("IThread::resume");

#ifdef IC_PM
  unsigned long rc = DosResumeThread( this->id() );
  if ( rc != NO_ERROR )
  {
    ITHROWSYSTEMERROR( rc,
                       "DosResumeThread",
                       IBaseErrorInfo::accessError,
                       IException::recoverable );
  }
#endif
#ifdef IC_WIN
  if (IPlatform::isWin32S())
  {
    ITRACE_DEVELOP("Win32s: resume() has no effect");
    return;
  }
  if ( WAIT_FAILED == ResumeThread( this->handle() ) )
  {
    ITHROWGUIERROR2( "ResumeThread",
                     IBaseErrorInfo::accessError,
                     IException::recoverable );
  }
#endif
}

/*------------------------------------------------------------------------------
| IThread::id                                                                  |
|                                                                              |
| Return the id from the associated started thread.                            |
------------------------------------------------------------------------------*/
IThreadId IThread :: id ( ) const
  {
  return this->startedThread()->id;
  }

/*------------------------------------------------------------------------------
| IThread::handle                                                              |
------------------------------------------------------------------------------*/
IThreadHandle IThread :: handle ( ) const
  {
  // catch cases where handle is not yet setup
  IASSERTSTATE( this->startedThread()->handle() != 0 );
  return this->startedThread()->handle();
  }

/*------------------------------------------------------------------------------
| IThread::stackSize                                                           |
|                                                                              |
| Return the stkSize member of the referenced IStartedThread.                  |
------------------------------------------------------------------------------*/
unsigned long IThread :: stackSize ( ) const
  {
#ifdef IC_PM
  tib_t *ptib = this->startedThread()->tib();
  if ( ptib )
    {
    return (char*)ptib->tib_pstacklimit - (char*)ptib->tib_pstack;
    }
  else
    return this->startedThread()->stkSize;
#endif
#ifdef IC_WIN
  // The available stack for the thread is controlled by the
  // /STACK linker directive.  The stack will grow to this limit.
  // This member is used to control the committed stack allocated
  // when the thread is created.   We do not attempt to determine
  // the committed size of the stack.
  return this->startedThread()->stkSize;
#endif
  }

/*------------------------------------------------------------------------------
| IThread::setStackSize                                                        |
|                                                                              |
| Set the stkSize member of the referenced IStartedThread.                     |
------------------------------------------------------------------------------*/
IThread &IThread :: setStackSize ( unsigned long newStkSize )
  {
  this->startedThread() -> stkSize = newStkSize;
  return *this;
  }

/*------------------------------------------------------------------------------
| IThread::autoInitGUI                                                         |
|                                                                              |
| Return the autoInitPM member of the referenced IStartedThread.               |
------------------------------------------------------------------------------*/
Boolean IThread :: autoInitGUI ( ) const
  {
  return this->startedThread() -> autoInitPM;
  }

/*------------------------------------------------------------------------------
| IThread::setAutoInitGUI                                                      |
|                                                                              |
| Set the autoInitPM member of the referenced IStartedThread.                  |
------------------------------------------------------------------------------*/
IThread &IThread :: setAutoInitGUI ( Boolean newFlag )
  {
  this->startedThread() -> autoInitPM = newFlag;
  return *this;
  }

/*------------------------------------------------------------------------------
| IThread::stopProcessingMsgs                                                  |
|                                                                              |
| If the thread is in a message polling loop, post a WM_QUIT, else, throw an   |
| invalid request exception.                                                   |
------------------------------------------------------------------------------*/
IThread &IThread :: stopProcessingMsgs ( )
  {
  IMODTRACE_DEVELOP( "ICurrentThread::stopProcessingMsgs" );

  IStartedThread
   *p = this->startedThread();

  if ( p->inMsgLoop )
    {
#if 0  //  Causes sporatic traps in Extra - HT
    // Allow some amount of clean up on the windows.  These windows
    // otherwise would not do WM_DESTROY processing, since WM_QUIT
    // would shut down any further messages.
    IWindowList* winList = this->windowList();
    if ( winList )
    {
       IWindowList::Cursor cursor( *winList );
       for ( cursor.setToFirst(); cursor.isValid(); cursor.setToNext() )
       {         // Send a message to all windows.
          IWindow* window = winList->elementAt( cursor );
          window->sendEvent( IC_UM_CLOSE );
       }
    }
#endif

    p->hmq.postEvent( WM_QUIT );
    }
  else
    {
    ITHROWLIBRARYERROR( IC_THREAD_NOT_PROCESSING_MSGS,
                        IBaseErrorInfo::invalidRequest,
                        IException::recoverable );
    }
  return *this;
  }

/*------------------------------------------------------------------------------
| IThread::priorityClass                                                       |
|                                                                              |
|  Query the actual thread class.  If not available (i.e., the thread isn't    |
|  started) return the priority class that's been set).                        |
------------------------------------------------------------------------------*/
IApplication::PriorityClass IThread :: priorityClass ( ) const
{
  IMODTRACE_DEVELOP( "IThread::priorityClass" );

  IStartedThread
   *p = this->startedThread();
  if ( p->isStarted() )
  {
    unsigned t = 0;
#ifdef IC_PM
    tib_t
     *ptib = p->tib();
    if ( ptib )
      t = ptib->tib_ptib2->tib2_ulpri / 256;
#endif
#ifdef IC_WIN
    // Note that in Win32 the priority class applies to the process,
    // not a single thread.
    if (IPlatform::isWin32S())
    {
      ITRACE_DEVELOP("Win32s: priorityClass() does not get priority");
      t = (unsigned)PRTYC_NOCHANGE;
    }
    else
      t = (unsigned)GetPriorityClass( GetCurrentProcess() );
#endif
    IApplication::PriorityClass result = IApplication::noChange;
    switch ( t ) {
      case PRTYC_IDLETIME:
        result = IApplication::idleTime;
        break;
      case PRTYC_REGULAR:
        result = IApplication::regular;
        break;
      case PRTYC_TIMECRITICAL:
        result = IApplication::timeCritical;
        break;
      case PRTYC_FOREGROUNDSERVER:
        result = IApplication::foregroundServer;
        break;
    }

    return( result );
  }   // is started

  return( p->priorityClass );
}

/*------------------------------------------------------------------------------
| IThread::priorityLevel                                                       |
|                                                                              |
| Query the actual thread priority.  If not available (i.e., the thread        |
| isn't started) return either the priority that's been set, or, 0).           |
------------------------------------------------------------------------------*/
unsigned IThread :: priorityLevel ( ) const
{
  IMODTRACE_DEVELOP( "IThread::priorityLevel" );

#ifdef IC_WIN
  if (IPlatform::isWin32S())
  {
    ITRACE_DEVELOP(IString("Win32s: priorityLevel() has no effect, return 0"));
    return 0;
  }
#endif
  IStartedThread
   *p = this->startedThread();
  if ( p->isStarted() )
  {
#ifdef IC_PM
    tib_t
     *t = p->tib();
    if ( t )
      return( t->tib_ptib2->tib2_ulpri % 256 );
#endif
#ifdef IC_WIN
    // GetThreadPriority returns a value between -15 and 15, inclusive
    int   t = GetThreadPriority( p->handle() );
    return( (unsigned)(IC_THREAD_PRIORITY_ADJUST + t) );
#endif
  }

  return( ( p->priorityLevel == -1 ) ? 0 : p->priorityLevel );
}

/*------------------------------------------------------------------------------
| IThread::adjustPriority                                                      |
|                                                                              |
| Adjust the thread priority level by the given delta.                         |
------------------------------------------------------------------------------*/
IThread &IThread :: adjustPriority ( int delta )
{
  IMODTRACE_DEVELOP( "IThread::adjustPriority" );

  IStartedThread
   *p = this->startedThread();

#ifdef IC_PM
  p->priorityLevel = delta;
  if ( p->isStarted() )
  {
    unsigned long
      rc = DosSetPriority( PRTYS_THREAD, PRTYC_NOCHANGE, delta, p->id );
    if ( rc != NO_ERROR )
    {
      ITHROWSYSTEMERROR( rc,
                         "DosSetPriority",
                         IBaseErrorInfo::accessError,
                         IException::recoverable );
    }
  }
#endif
#ifdef IC_WIN
  if (IPlatform::isWin32S())
  {
    ITRACE_DEVELOP(IString("Win32s: adjustPriority() has no effect"));
    return( *this );
  }
  if ( p->isStarted() )
  {
    int t = GetThreadPriority( p->handle() );
    t += delta;
    // clamp to allowed values
    if (t > THREAD_PRIORITY_HIGHEST)
      t = THREAD_PRIORITY_HIGHEST;

    if (t < THREAD_PRIORITY_LOWEST)
      t = THREAD_PRIORITY_LOWEST;

    // store the value exposed via interface (0..31)
    p->priorityLevel = t + IC_THREAD_PRIORITY_ADJUST;

    if ( !SetThreadPriority( p->handle(), t ) )
    {
      ITHROWGUIERROR2( "SetThreadPriority",
                        IBaseErrorInfo::accessError,
                        IException::recoverable );
    }
  }
  else
  {
    // set delta
    p->priorityLevel = delta;
  }
#endif

  return( *this );
}

/*------------------------------------------------------------------------------
| IThread::setPriority                                                         |
|                                                                              |
| Set the priority class member of the referenced IStartedThread.              |
| In addition, issue DosSetPriority (in case the thread is started).           |
------------------------------------------------------------------------------*/
IThread &IThread :: setPriority ( IApplication::PriorityClass priorityClass,
                                  unsigned                    priorityLevel )
{
  IMODTRACE_DEVELOP( "IThread::setPriority" );

#ifdef IC_WIN
  if (IPlatform::isWin32S())
  {
    ITRACE_DEVELOP(IString("Win32s: setPriority() has no effect"));
    return *this;
  }
#endif

  if ( priorityClass == IApplication::noChange )
    priorityClass = this->priorityClass();

  IStartedThread
   *p = this->startedThread();
  p->priorityClass = priorityClass;
  p->priorityLevel = priorityLevel;
  if ( p->isStarted() )
  {
    unsigned long
      pclass = 0;
    switch ( priorityClass ) {
      case IApplication::idleTime:
        pclass = PRTYC_IDLETIME;
        break;
      case IApplication::regular:
        pclass = PRTYC_REGULAR;
        break;
      case IApplication::timeCritical:
        pclass = PRTYC_TIMECRITICAL;
        break;
      case IApplication::foregroundServer:
        pclass = PRTYC_FOREGROUNDSERVER;
        break;
    }
#ifdef IC_PM
    unsigned long
      rc = DosSetPriority( PRTYS_THREAD,
                           pclass,
                           priorityLevel,
                           p->id );
    if ( rc != NO_ERROR )
    {
      ITHROWSYSTEMERROR( rc,
                         "DosSetPriority",
                         IBaseErrorInfo::accessError,
                         IException::recoverable );
    }
#endif
#ifdef IC_WIN
    if (pclass != 0)
    {
      if ( !SetPriorityClass( GetCurrentProcess(), pclass) )
      {
        ITHROWGUIERROR2( "SetPriorityClass",
                         IBaseErrorInfo::accessError,
                         IException::recoverable );
      }
    }

    int t = priorityLevel - IC_THREAD_PRIORITY_ADJUST;
    // clamp to allowed values
    if (t > THREAD_PRIORITY_HIGHEST)
      t = THREAD_PRIORITY_HIGHEST;

    if (t < THREAD_PRIORITY_LOWEST)
      t = THREAD_PRIORITY_LOWEST;

    if ( !SetThreadPriority( p->handle(), t ) )
    {
      ITHROWGUIERROR2( "SetThreadPriority",
                       IBaseErrorInfo::accessError,
                       IException::recoverable );
    }
#endif
  }

  return( *this );
}

/*------------------------------------------------------------------------------
| IThread::queueSize                                                           |
|                                                                              |
| Returns the queue size from the associated started thread                    |
------------------------------------------------------------------------------*/
long IThread :: queueSize ( ) const
  {
  return this->startedThread()->queueSize;
  }

/*------------------------------------------------------------------------------
| IThread::setQueueSize                                                        |
|                                                                              |
| Set the queue size of the associated started thread                          |
------------------------------------------------------------------------------*/
IThread &IThread :: setQueueSize ( long newQSize )
  {
  this->startedThread()->queueSize = newQSize;
  return *this;
  }

/*------------------------------------------------------------------------------
| IThread::messageQueue                                                        |
|                                                                              |
| Return the hmq. First try the stored hmq (if there is one), else return 0 .  |
| This non-const member function will adjust the stored HMQ if it is           |
| found to be invalid.                                                         |
------------------------------------------------------------------------------*/
IMessageQueueHandle IThread :: messageQueue ( )
  {
  IStartedThread
   *p = this->startedThread();

  IMessageQueueHandle hmq = p->hmq;

#ifdef IC_PM
  if ( p->hmq )
    {
    MQINFO info;
    if ( !WinQueryQueueInfo( hmq, &info, sizeof info ) )
      {
      p->hmq = 0; // No longer valid.
      hmq    = 0;
      }
    }
#endif
  return hmq;
  }

/*------------------------------------------------------------------------------
| IThread::messageQueue                                                        |
|                                                                              |
| Return the hmq. First try the stored hmq (if there is one), else return 0 .  |
------------------------------------------------------------------------------*/
IMessageQueueHandle IThread :: messageQueue ( ) const
  {
  IStartedThread
   *p = this->startedThread();

  IMessageQueueHandle hmq = p->hmq;

#ifdef IC_PM
  if ( p->hmq )
    {
    MQINFO info;
    if ( !WinQueryQueueInfo( hmq, &info, sizeof info ) )
      hmq = 0;
    }
#endif
  return hmq;
  }

/*------------------------------------------------------------------------------
| IThread::windowList                                                          |
|                                                                              |
| Returns the window list from the associated started thread                   |
------------------------------------------------------------------------------*/
IWindowList *IThread :: windowList ( ) const
{
  return this->startedThread()->windowList;
}

/*------------------------------------------------------------------------------
| IThread::setWindowList                                                       |
|                                                                              |
| Sets the window list for the associated start thread                         |
------------------------------------------------------------------------------*/
IThread &IThread :: setWindowList ( IWindowList *list )
  {
  this->startedThread()->windowList = list;
  return *this;
  }

#ifdef IC_WIN
/*------------------------------------------------------------------------------
| IThread::dialogControls                                                      |
|                                                                              |
| Returns the list of dialog controls from the associated started thread       |
------------------------------------------------------------------------------*/
IDialogControls  *IThread::dialogControls ( ) const
{
  if ( !(this->startedThread()->fdialogControls) )
     this->startedThread()->fdialogControls = new IDialogControls;
  return this->startedThread()->fdialogControls;
}
#endif

/*------------------------------------------------------------------------------
| IThread::isStarted                                                           |
|                                                                              |
| Returns true if the associated start thread has been started                 |
------------------------------------------------------------------------------*/
Boolean IThread :: isStarted ( ) const
  {
  return this->startedThread()->isStarted();
  }

/*------------------------------------------------------------------------------
| ICurrentThread::ICurrentThread                                               |
|                                                                              |
| Initialize referenced IStartedThread to zero.                                |
|                                                                              |
| Notes:                                                                       |
|   1. During the course of constructing this object, an IStartedThread        |
|      corresponding to thread 1 (the main application thread) is              |
|      created.  This IStartedThread is never used directly, however.          |
|   2. This constructor is protected to prevent creation of additional         |
|      objects of this class.  The only instance is the static                 |
|      currentThread object within this module (see above).                    |
------------------------------------------------------------------------------*/
ICurrentThread :: ICurrentThread ( )
#ifdef IC_WIN
  : IThread( (IHandle::Value)GetCurrentThreadId() )
#else
  : IThread( IStartedThread::fgPrimaryTID )
#endif
  {
  }

/*------------------------------------------------------------------------------
| ICurrentThread::anchorBlock                                                  |
|                                                                              |
| Return an IAnchorBlockHandle constructed from the current thread's HAB.      |
------------------------------------------------------------------------------*/
IAnchorBlockHandle ICurrentThread :: anchorBlock ( ) const
  {
  IStartedThread
   *p = this->startedThread();

  IAnchorBlockHandle
    hab( 0 );

  if ( !p->hab )
  {
#ifdef IC_PM
    if ( ( hab = WinInitialize( 0 ) ) != 0 )
       p->hab = hab;
    else
    {
      ITHROWGUIERROR2( "WinInitialize",
                       IBaseErrorInfo::accessError,
                       IException::unrecoverable );
    }
#endif
#ifdef IC_WIN
    // Unused value, but set to non-zero to avoid breaking existing code
    hab = IAnchorBlockHandle( (IHandle::Value)1 );
    p->hab = hab;
#endif
  }
  else
  {
    hab = p->hab;
  }
  return( hab );
  }

/*------------------------------------------------------------------------------
| ICurrentThread::isGUIInitialized                                              |
|                                                                              |
| Return true if PM has been initialized for this thread.                      |
------------------------------------------------------------------------------*/
Boolean ICurrentThread :: isGUIInitialized ( ) const
  {
  if ( this->anchorBlock() != 0
       &&
       this->messageQueue() != 0 )
    return true;
  else
    return false;
  }

/*------------------------------------------------------------------------------
| ICurrentThread::messageQueue                                                 |
|                                                                              |
| Return the hmq. First try the stored hmq ( IThread handles this)             |
| (if there is one), else try HMQ_CURRENT.                                     |
| Return 0 if no message queue is stored and HMQ_CURRENT is invalid.           |
------------------------------------------------------------------------------*/
IMessageQueueHandle ICurrentThread :: messageQueue ( )
  {
  IMessageQueueHandle hmq = this->Inherited::messageQueue();
  if ( !hmq )
    {
#ifdef IC_PM
    MQINFO info;
    hmq = HMQ_CURRENT;
    if ( !WinQueryQueueInfo( hmq, &info, sizeof info ) )
      hmq = 0;
#endif
#ifdef IC_WIN
    hmq = (IHandle::Value)GetCurrentThreadId(); // message queue id == thread id
#endif
    }
  return hmq;
  }

/*------------------------------------------------------------------------------
| ICurrentThread::messageQueue                                                 |
|                                                                              |
| Return the hmq. First try the stored hmq ( IThread handles this)             |
| (if there is one), else try HMQ_CURRENT.                                     |
| Return 0 if no message queue is stored and HMQ_CURRENT is invalid.           |
------------------------------------------------------------------------------*/
IMessageQueueHandle ICurrentThread :: messageQueue ( ) const
  {
  IMessageQueueHandle hmq = this->Inherited::messageQueue();
  if ( !hmq )
    {
#ifdef IC_PM
    MQINFO info;
    hmq = HMQ_CURRENT;
    if ( !WinQueryQueueInfo( hmq, &info, sizeof info ) )
      hmq = 0;
#endif
#ifdef IC_WIN
    hmq = (IHandle::Value)GetCurrentThreadId(); // message queue id == thread id
#endif
    }
  return hmq;
  }

/*------------------------------------------------------------------------------
| ICurrentThread::exit                                                         |
|                                                                              |
| Terminate the current thread with the given return code.  The thread is      |
| terminated via DosExit.                                                      |
------------------------------------------------------------------------------*/
void ICurrentThread :: exit ( unsigned long rc )
  {
  this->startedThread()->end();

  if ( id() == IStartedThread::fgPrimaryTID )
    ::exit( (int)rc );
  else
    {
  #ifdef IC_PM
    DosExit( EXIT_THREAD, rc );
  #endif
  #ifdef IC_WIN
    ExitThread(rc);
  #endif
    }
  }

/*------------------------------------------------------------------------------
| ICurrentThread::sleep                                                        |
|                                                                              |
| Suspend the current thread for the given number of milliseconds.  The thread |
| is suspended via DosSleep.                                                   |
------------------------------------------------------------------------------*/
ICurrentThread &ICurrentThread :: sleep ( unsigned long msecs )
  {
  ISLEEP( msecs );
  return *this;
  }

/*------------------------------------------------------------------------------
| ICurrentThread::waitFor                                                      |
|                                                                              |
| Issue DosWaitThread on the argument thread's id.                             |
------------------------------------------------------------------------------*/
ICurrentThread &ICurrentThread :: waitFor ( const IThread &anotherThread )
{
  IMODTRACE_DEVELOP( "ICurrentThread::waitFor" );
  IThreadId tid = anotherThread.id();

  IASSERTPARM( tid != IStartedThread::fgPrimaryTID );
  IASSERTPARM( tid != IThread::current().id() );
  if ( tid )
  {
#ifdef IC_PM
    unsigned long
      rc = DosWaitThread( (PTID)&tid, DCWW_WAIT );
    if ( rc != NO_ERROR && rc != ERROR_INVALID_THREADID )
    {
      ITHROWSYSTEMERROR( rc,
                         "DosWaitThread",
                         IBaseErrorInfo::accessError,
                         IException::recoverable );
    }
#endif
#ifdef IC_WIN
    if (IPlatform::isWin32S())
    {
      ITRACE_DEVELOP(IString("Win32s: waitFor() has no effect"));
    }
    else
    {
      if ( anotherThread.startedThread()->handle() )
      {
        if (WAIT_FAILED == WaitForSingleObject( anotherThread.handle(),
                                                (DWORD)INFINITE) )
        {
          ITHROWGUIERROR2( "WaitForSingleObject",
                           IBaseErrorInfo::accessError,
                           IException::recoverable );
        }
      }
    }
#endif
  }

  return( *this );
}

/*------------------------------------------------------------------------------
| ICurrentThread::waitForAllThreads                                            |
|                                                                              |
| Wait for any thread till there are no more...                                |
------------------------------------------------------------------------------*/
ICurrentThread &ICurrentThread :: waitForAllThreads ( )
{
  IMODTRACE_ALL( "ICurrentThread::waitForAllThreads" );
  IASSERTPARM( this->id() == IStartedThread::fgPrimaryTID );

#ifdef IC_PM
  Boolean loopFlag = true;
  IThreadId tid;
  while ( loopFlag )
  {
    /******************************************************************/
    /* Have the current thread wait until the next thread in the      */
    /* process ends                                                   */
    /******************************************************************/
    tid = 0;
    if ( DosWaitThread( (PTID)&tid, DCWW_WAIT ) != NO_ERROR )
      loopFlag = false;
  }
#endif
#ifdef IC_WIN
  if (IPlatform::isWin32S())
    ITRACE_DEVELOP(IString("Win32s: waitForAllThreads() has no effect"));
  else
  {
    // wait for all threads to terminate
    IStartedThread::waitForThreadEnd( true,       // wait for all
                                      this->id(), // except this one
                                      true );     // throw if error
  }
#endif

  return( *this );
}

/*------------------------------------------------------------------------------
| ICurrentThread::waitForAnyThread                                             |
|                                                                              |
| Issue DosWaitThread for thread id 0 and return the thread actually           |
| terminating.  Throw an exception in case of errors.                          |
------------------------------------------------------------------------------*/
IThreadId ICurrentThread :: waitForAnyThread ( )
{
  IMODTRACE_ALL( "ICurrentThread::waitForAnyThread" );
  IThreadId tid = 0;

#ifdef IC_PM
  unsigned long
    rc = DosWaitThread( (PTID)&tid, DCWW_WAIT );
  if ( rc != NO_ERROR )
  {
    ITHROWSYSTEMERROR( rc,
                      "DosWaitThread",
                      IBaseErrorInfo::accessError,
                      IException::recoverable );
  }
#endif

#ifdef IC_WIN
  if (IPlatform::isWin32S())
    ITRACE_DEVELOP(IString("Win32s: waitForAnyThread() has no effect"));
  else
  {
    // wait for one thread to terminate
    tid = IStartedThread::waitForThreadEnd( false,      // wait for one
                                            this->id(), // except this one
                                            true );     // throw if error
  }
#endif

  return( tid );
}

/*------------------------------------------------------------------------------
| ICurrentThread :: id                                                         |
|                                                                              |
| Return the current thread's id.                                              |
------------------------------------------------------------------------------*/
IThreadId ICurrentThread :: id ( ) const
  {
  return IThread::currentId();
  }

/*------------------------------------------------------------------------------
| ICurrentThread::handle                                                       |
------------------------------------------------------------------------------*/
IThreadHandle ICurrentThread :: handle ( ) const
  {
  return IThread::currentHandle( );
  }

/*------------------------------------------------------------------------------
| ICurrentThread::remainingStack                                               |
|                                                                              |
| Approximate stack remaining as the distance between the stack base           |
| (from the tib structure) and the current stack pointer (approximately        |
| the address of this function's argument).                                    |
------------------------------------------------------------------------------*/
unsigned long ICurrentThread :: remainingStack ( ) const
  {
#ifdef IC_PM
  tib_t
   *tib = this->startedThread()->tib();
  void
   *base = tib->tib_pstack,
   *cur  = &tib;
  return (char*)cur - (char*)base;
#endif
#ifdef IC_WIN
//#ifdef __BORLANDC__
// return stackavail();
//#else
 return 1000L;
//#endif
#endif
  }

/*------------------------------------------------------------------------------
| ICurrentThread::processMsgs                                                  |
|                                                                              |
| Loop, issuing WinGetMessage/WinDispatchMsg, until a WM_QUIT is received. If  |
| the quit has a non-NULLHANDLE hwnd in mp2, then it probably came from the    |
| task list so instead of quitting, post a WM_SYSCOMMAND+SC_CLOSE to the       |
| window in question.                                                          |
------------------------------------------------------------------------------*/
void ICurrentThread :: processMsgs ( )
{
  IMODTRACE_DEVELOP( "ICurrentThread::processMsgs" );

  IStartedThread
   *p = this->startedThread();
  if ( ! this->isGUIInitialized() )
    this->initializeGUI();

  IAnchorBlockHandle
    hab = this->anchorBlock();
  QMSG
    msg;

  p->inMsgLoop++;          //signal start of the loop
  ITRACE_DEVELOP( "Starting message loop - Threads in message loop = " +
                  IString( p->inMsgLoop ) );

#ifdef IC_PERF
   // Performance builds.
   // Environment variable supported
   //  ICLUI_PERF=INITONLY - stop tracing in IThread::processMsgs()
   //  ICLUI_PERF=INITEXIT - exit in IThread::processMsgs()
   char *envptr = getenv("ICLUI_PERF");
   if (envptr)
      {
      IString strTrace(envptr);
      strTrace.strip().upperCase();
      if (strTrace== "INITONLY")
         PerfStop();
      else if (strTrace== "INITEXIT")
         this->exit(0);
      }
#endif

  while ( true )
  {
#ifdef IC_WIN
    if ( (IGETMSG( hab, &msg, 0, 0, 0 ) > 0) && (msg.message != IC_UM_QUIT) )
    {
      ITRACE_ALL(
         IString("Event loop  hwnd=") +
         IString((unsigned long)msg.hwnd).d2x()+
         IString(", Msg=") + IString(msg.message).d2x()+
         IString(", wParam=") + IString(msg.wParam).d2x()+
         IString(", lParam=") + IString(msg.lParam).d2x() );

      // Check for accelerators.   If the translateAcclerator function
      // returns true we have handled this event.
      if (!( ((msg.message >= WM_KEYFIRST) && (msg.message <= WM_KEYLAST)) &&
             ICurrentThreadData::translateAccelerator( msg ) ) )
      {
         // Not accelerator.  Call TranslateMessage to create WM_CHARs if
         // appropriate and then dispatch the event.
         TranslateMessage( &msg );
         IDISPATCHMSG( hab, &msg );
      }
    }
    else
    {
      ITRACE_DEVELOP(
         IString("WM_QUIT hwnd=") +
         IString((unsigned long)msg.hwnd).d2x()+
         IString(", Msg=") + IString(msg.message).d2x()+
         IString(", wParam=") + IString(msg.wParam).d2x()+
         IString(", lParam=") + IString(msg.lParam).d2x() );

      // Check for WM_QUIT (from Windows) or IC_UM_QUIT and a value
      // in lParam equal to the current inMsgLoop value.  If either is
      // true exit the loop.  We allow lParam to be > inMsgLoop to allow
      // a showModally without a primary event loop.
      if ( (msg.message == WM_QUIT) ||
           (msg.message == IC_UM_QUIT) && (msg.lParam >= p->inMsgLoop) )
         {
         break;
         }
      else
         {
         // We have received an IC_UM_QUIT which is not the top of the
         // nesting stack.  We have already hidden the modal dialog and reenabled
         // the owner, so at this point we just have to remember the
         // fact that we were asked to exit.  We stack up the requests
         // and check the stack after each exit from this routine.  See
         // IFrameWindow::showModally.
         if (!p->fquitStack)
            p->fquitStack = new IWMQuitStack(10);
         p->fquitStack->add( msg.lParam );
         }
    }
#endif
#ifdef IC_PM
    if ( IGETMSG( hab, &msg, 0, 0, 0 ) )
    {
      IDISPATCHMSG( hab, &msg );
    }
    else
    {
      if ( msg.mp2 != NULL )
      {
        IWindowHandle frame( (unsigned long)( msg.mp2 ) );
        if ( frame.isValid() )
        {     // (A modal frame closed via the task list will
              // have its now invalid window handle here.)
           frame
            .postEvent( WM_SYSCOMMAND,
                        SC_CLOSE,
                        IEventData( CMDSRC_OTHER, false ) );
        }
      }
      else
        break;
    } // end getmsg
#endif
  } // end while

  p->inMsgLoop--;          //signal end of the loop

#ifdef IC_WIN
  // If we have a stack of unsatisfied requests to exit, inspect it
  // to see if we can satisfy a request now.
  if ( (p->fquitStack) && (p->fquitStack->lastElement() >= p->inMsgLoop ))
     {
     // Ready to exit.  Remove the request and post an IC_UM_QUIT message.
     // If the stack is now empty delete it.
     p->fquitStack->removeLast();
     p->hmq.postEvent( IC_UM_QUIT, 0, p->inMsgLoop );
     if (p->fquitStack->isEmpty() )
        {
        delete p->fquitStack;
        p->fquitStack = 0;
        }
     }
#endif

  ITRACE_DEVELOP( "Ending message loop - Threads in message loop = " +
                  IString( p->inMsgLoop ) );
}

/*------------------------------------------------------------------------------
| ICurrentThread::initializeGUI                                                 |
|                                                                              |
| Initialize the thread for PM by calling WinInitialize and WinCreateMsgQueue. |
| If an error occurs, throw an exception.                                      |
|                                                                              |
| Notes:                                                                       |
|   1. Increment the use count for the referenced IStartedThread (so the       |
|      thread-related hab/hmq are retained till terminateGUI).                 |
------------------------------------------------------------------------------*/
void ICurrentThread :: initializeGUI ( long queueSize )
{
  IMODTRACE_DEVELOP( "ICurrentThread::initializeGUI" );
  IStartedThread
   *p = this->startedThread();
  IAnchorBlockHandle hab = this->anchorBlock();

// Moved WinInitialize logic to anchorBlock() function

  IMessageQueueHandle hmq = this->messageQueue();


#ifdef IC_PM
  if ( !hmq )
  {
    if ( !queueSize )
      queueSize = this->queueSize();

    if ( ( hmq = WinCreateMsgQueue( hab, queueSize ) ) != 0 )
    {
       p->hmq = hmq;
       p->addRef();
    }
    else
      // WinCreateMsqQueue failed, throw exception...
    {
      ITHROWGUIERROR2( "WinCreateMsgQueue",
                       IBaseErrorInfo::accessError,
                       IException::unrecoverable );
    }
  }
#else
   // The message queue is to be set only the first time initializeGUI is
   // called.  a test on hmq is not sufficient, because it will always
   // be non-zero.  Test a private flag instead.
   if ( (!fsetMessageQueue) && (IPlatform::isWin32S()) )
     {
      // If the user passes in the queuesize, use it.  If the user
      // does not pass in the queue size, (its the default) use 96.
      // 96 is recommended minumum for applications that use OLE.

      if ((!queueSize) || (queueSize == IThread::dfltQueueSize))
        queueSize = 96;

      while ((SetMessageQueue((int)queueSize) == 0) && (queueSize >0) )
        queueSize -=8;

      fsetMessageQueue = true;
      setQueueSize(queueSize);
      ITRACE_DEVELOP(IString("Win32s: increase message queue size to: ")
                    +IString(queueSize) );
     }
   // Make sure hmq is stored.  Don't need to worry about addRef,
   // since we don't need to cleanup.  Queue size cannot be set.
   if (p->hmq != hmq)
      p->hmq = hmq;

   // Call a function in USER.LIB to insure GUI initialized.  This is
   // automatic but certain functions can fail without it (like
   // PostThreadMessage).
   GetDesktopWindow();
#endif
}

/*------------------------------------------------------------------------------
| ICurrentThread::terminateGUI                                                  |
|                                                                              |
| Issue WinDestroyMsgQueue and WinTerminte, as necessary.                      |
------------------------------------------------------------------------------*/
void ICurrentThread :: terminateGUI ( )
{
  IMODTRACE_DEVELOP( "ICurrentThread::terminateGUI" );
#ifdef IC_PM
  IStartedThread
   *p = this->startedThread();

  if ( this->messageQueue() )
  {
     WinDestroyMsgQueue( this->messageQueue() );
     if ( p->hmq )
     {             // Created and stored in initializeGUI.
        p->hmq = 0;
        p->removeRef();
     }
  }

  if ( this->anchorBlock() )
     WinTerminate( this->anchorBlock() );
#else
// Message queue deleted automatically in Windows so no-op
#endif
}

/*------------------------------------------------------------------------------
| IThread__watcher                                                             |
------------------------------------------------------------------------------*/
#ifndef IC_WIN
static void _System IThread__watcher ( unsigned long );
#endif

/*------------------------------------------------------------------------------
| IStartedThread::watch                                                        |
|                                                                              |
| This function is called when a non-IThread thread is encountered in          |
| IStartedThread::startedThread.  The first time it is called it launches      |
| a secondary thread to watch for thread terminations.  This thread will       |
| see if terminating threads are non-IThread and if so, call end() to          |
| clean them up and call removeRef() to free the IStartedThread resource       |
| (if no other references exist).                                              |
|                                                                              |
| The thread watching is done by the function IThread__watcher (see below).    |
|                                                                              |
| Notes:                                                                       |
|   1. Must be called with resources locked.                                   |
------------------------------------------------------------------------------*/
void IStartedThread :: watch ( )
{
  IMODTRACE_DEVELOP( "IStartedThread::watch" );

  // This function does nothing in Windows
#ifdef IC_PM
  TID tid;


  if ( !IStartedThread::watching )
  {
    unsigned long
      rc = DosCreateThread( &tid,
                            IThread__watcher,
                            0,
                            0,
                            16384 );
    if ( rc )
    {
      ITHROWSYSTEMERROR( rc,
                         "DosCreateThread",
                         IBaseErrorInfo::accessError,
                         IException::unrecoverable );
    }
    else
      IStartedThread::watching = true;
  }
#endif
}

/*------------------------------------------------------------------------------
| IThread__watcher                                                             |
|                                                                              |
| This function watches for thread terminations.  When a thread ends, the      |
| list of started threads is searched for this thread.  If found, then we      |
| remove it from the list and call end().  Also, the reference is count        |
| is decremented (twice if PM had been started).                               |
------------------------------------------------------------------------------*/
#ifndef IC_WIN
void _System IThread__watcher ( unsigned long )
{
  Boolean
    watching = true;
  unsigned long
    rc = 0;

  while ( watching && rc == 0 )
  {
    IThreadId tid = 0;
    ITRACE_DEVELOP( "Waiting for a thread to end..." );
    rc = DosWaitThread( (PTID)&tid, DCWW_WAIT );

    // Check to see if a thread really ended...
    if ( rc == 0 )
    {
       ITRACE_DEVELOP( "Thread " + IString( tid.asUnsigned() ) + " ended." );
       IResourceLock
         lock( IStartedThread::lock() );
       unsigned
         index = 0;

       // Check if the application managed to start another thread
       // with the same thread ID.
       Boolean
         reusedId = IStartedThread::isTIDValid( tid ),
         threadEnded = false;
       IStartedThread* restartedThread = 0;

       // Kill this thread unless no more not "startedHere" threads.
       watching = false;
       // Examine all started threads...
       while ( index < IStartedThread::arraySize )
       {
          IStartedThread
           *t = IStartedThread::threads[ index ];
          // Make sure slot is filled...
          if ( t )
          {
             // See if this is the entry for the thread that just ended.
             if ( t->id == tid )
             {        // Thread IDs match.
                if ( reusedId )
                {     // The ended thread has been replaced with
                      // another one with the same thread ID.  Be
                      // sure not to delete the wrong one.
                   if ( restartedThread )
                   {
                      if ( t->fCount > restartedThread->fCount )
                      {    // Keep the newest IStartedThread.
                         IStartedThread* swap = t;
                         t = restartedThread;     // End the old one.
                         restartedThread = swap;  // Keep the new one.
                      }
                      // Else t, restartedThread already ok.
                   }
                   else
                   {       // Decide later whether to end this one.
                      restartedThread = t;
                   }
                }

                if ( ! t->startedHere  &&
                     t != restartedThread )
                {     // Need to "end" the IStartedThread.
                   if ( restartedThread )
                   {
                      ITRACE_DEVELOP( "\tEnding oldest thread with same ID." );
                   }
                   ITRACE_DEVELOP( "\tRemoving from IStartedThread array!" );
                   t -> end();
                   threadEnded = true;
                }
             }
             else if ( ! t->startedHere )
             {     // If not started here, keep watching...
                watching = true;
             }
          }
          index++;
       }  // endwhile

       if ( restartedThread  &&
            threadEnded == false  &&
            ! restartedThread->startedHere )
       {  // Found only one IStartedThread with the ID to delete,
          // even though another thread with the same ID has been
          // started.  The IStartedThread could either be the one to
          // end (if it was not started via IThread) or the newly
          // started thread (if the ended thread was created via
          // IThread, and the new one wasn't but is now wrappered
          // with an IThread).
          ITRACE_DEVELOP( "\tOnly one of two started threads found..." );
          ITRACE_DEVELOP( "\tRemoving from IStartedThread array!" );
          restartedThread->end();
       }
    }
    else
    {
       ITRACE_DEVELOP( "Error watching threads, rc=" + IString( rc ) );
    }
  }

  ITRACE_DEVELOP( "No more threads to watch." );
  IStartedThread::watching = false;
}
#endif

/*------------------------------------------------------------------------------
| ICurrentThread::startedThread                                                |
|                                                                              |
| Returns the result of looking up the current TID.                            |
------------------------------------------------------------------------------*/
IStartedThread *ICurrentThread :: startedThread ( ) const
  {
  IStartedThread
   *result = IStartedThread::locate( this->id() );
  result -> removeRef();
  return result;
  }

/*------------------------------------------------------------------------------
| IThread::inGUISession                                                         |
|                                                                              |
| Query the current session type and return whether it is PM.                  |
------------------------------------------------------------------------------*/
Boolean IThread :: inGUISession ( )
{
  IMODTRACE_DEVELOP( "IThread::inGUISession" );

#ifdef IC_PM
  PTIB
    ptib;
  PPIB
    ppib;
  DosGetInfoBlocks( &ptib, &ppib );
  if ( ppib->pib_ultype == SSF_TYPE_PM )
    return( true );
  else
    return( false );
#endif
#ifdef IC_WIN
  if (IPlatform::isWin32S())
    ITRACE_DEVELOP(IString("Win32s: inGUISession() always returns true"));

  // Always true for any flavor of Windows.
  return( true );
#endif
}

/*------------------------------------------------------------------------------
| IThread::variable                                                            |
|                                                                              |
| Dispatch this request to the corresponding started thread.                   |
------------------------------------------------------------------------------*/
IString IThread :: variable ( const IString &key ) const
  {
  return this-> startedThread() -> variable( key );
  }

/*------------------------------------------------------------------------------
| IThread::setVariable                                                         |
|                                                                              |
| Dispatch this request to the corresponding started thread.                   |
------------------------------------------------------------------------------*/
IThread &IThread :: setVariable ( const IString &key,
                                  const IString &value )
  {
  this-> startedThread() -> setVariable( key, value );
  return *this;
  }

/*------------------------------------------------------------------------------
| IThread::defaultAutoInitGUI                                                  |
|                                                                              |
| Return static default value.                                                 |
------------------------------------------------------------------------------*/
Boolean IThread :: defaultAutoInitGUI ( )
  {
  if ( IThread::dfltAutoInitGUI == -1 )
    IThread::dfltAutoInitGUI = IThread::inGUISession();
  return IThread::dfltAutoInitGUI;
  }

/*------------------------------------------------------------------------------
| IThreadFn::IThreadFn                                                         |
|                                                                              |
| Empty ctor to prevent generated static version.                              |
------------------------------------------------------------------------------*/
IThreadFn :: IThreadFn ( )
  {
  }

/*------------------------------------------------------------------------------
| IThreadFn::~IThreadFn                                                        |
|                                                                              |
| Empty dtor to prevent generated static version.                              |
------------------------------------------------------------------------------*/
IThreadFn :: ~IThreadFn ( )
  {
  }

/*------------------------------------------------------------------------------
| IThread::current                                                             |
|                                                                              |
| Returns the current thread                                                   |
------------------------------------------------------------------------------*/
ICurrentThread &IThread :: current ( )
  {
  if ( !IThread::pCurrent )
    IThread::pCurrent = new ICurrentThread;
  return *IThread::pCurrent;
  }

/*------------------------------------------------------------------------------
| IThread::inMsgLoop                                                           |
|                                                                              |
| Get the receiver's IStartedThread and query its inMsgLoop flag.              |
------------------------------------------------------------------------------*/
Boolean IThread :: inMsgLoop ( ) const
  {
  return( this->startedThread()->inMsgLoop ? true : false );
  }

/*------------------------------------------------------------------------------
| IThread::Cursor::Cursor                                                      |
|                                                                              |
| Set data members:                                                            |
|   all      - set from constructor argument                                   |
|   created  - set to 0 (which is always invalid)                              |
------------------------------------------------------------------------------*/
IThread::Cursor :: Cursor ( Boolean allThreads )
  : all( allThreads ),
    iteratedAll( false ),
    created( 0 ),
    fCursorData( 0 )
  {
  }

/*------------------------------------------------------------------------------
| IThread::Cursor::~Cursor                                                     |
------------------------------------------------------------------------------*/
IThread::Cursor :: ~Cursor ( )
  {
//delete fCursorData; when it is defined.
  }

/*------------------------------------------------------------------------------
| IThread::Cursor::setToFirst                                                  |
|                                                                              |
| Point to first element of started thread array and reset this cursor's       |
| time stamp.                                                                  |
------------------------------------------------------------------------------*/
Boolean IThread::Cursor :: setToFirst ( )
  {
  this->created = IThread::Cursor::timeStamp;
  this->cursor  = (unsigned)-1;
  this->iteratedAll = false;
  return this -> advance();
  }

/*------------------------------------------------------------------------------
| IThread::Cursor::setToNext                                                   |
|                                                                              |
| Advance cursor and return whether it's still valid.                          |
------------------------------------------------------------------------------*/
Boolean IThread::Cursor :: setToNext ( )
  {
  return this -> advance();
  }

/*------------------------------------------------------------------------------
| IThread::Cursor::isValid                                                     |
|                                                                              |
| A cursor is valid iff its timestamp equals the static one (i.e., no          |
| threads have been started/stopped since it was reset).                       |
------------------------------------------------------------------------------*/
Boolean IThread::Cursor :: isValid ( ) const
  {
  return this->created == IThread::Cursor::timeStamp;
  }

/*------------------------------------------------------------------------------
| IThread::Cursor::invalidate                                                  |
|                                                                              |
| Invalidate the cursor by setting it's creation "timestamp" to an invalid     |
| value.                                                                       |
------------------------------------------------------------------------------*/
void IThread::Cursor :: invalidate ( )
  {
  this->created = 0;
  }

/*------------------------------------------------------------------------------
| IThread::Cursor::threadId                                                    |
|                                                                              |
| Check validity and return TID of started thread at cursor.                   |
------------------------------------------------------------------------------*/
IThreadId IThread::Cursor :: threadId ( ) const
  {
  IASSERTPARM( this->isValid() );
  return IStartedThread::threads[ this->cursor ] -> id;
  }

/*------------------------------------------------------------------------------
| IThread::Cursor::advance                                                     |
|                                                                              |
| Advance cursor to next IThread.  Skip non-PM if required.  If we run off     |
| the end, invalidate the cursor.                                              |
------------------------------------------------------------------------------*/
Boolean IThread::Cursor :: advance ( )
  {
  ICritSec
    lock;
  if ( this->isValid() )
    {
    for ( this->cursor++;
          this->cursor < IStartedThread::arraySize;
          this->cursor++ )
      {
      IStartedThread
       *p = IStartedThread::threads[ this->cursor ];
      if ( p && p->id )
        if ( this->all
             ||
             p->windowList )
          break;
      }
    if ( this->cursor == IStartedThread::arraySize )
      {
      this->iteratedAll = true;   // Flag no more items.
      this->invalidate();
      }
    }
  return this->isValid();
  }

#ifdef IC_WIN
/*------------------------------------------------------------------------------
| ICurrentThreadData::translateAccelerator                                     |
|                                                                              |
| Emulate the traversal of WM_TRANSLATEACCEL messages up the parent chain      |
| in PM, by chasing up the parent chain for an accelerator table.              |
------------------------------------------------------------------------------*/
IBase::Boolean ICurrentThreadData::translateAccelerator ( MSG& message )
{
  Boolean
    translated = false;
  IWindowHandle
    desktopHandle = IWindow::desktopWindow()->handle(),
    objectHandle = IWindow::objectWindow()->handle();

  for ( IWindowHandle hwnd = message.hwnd;
        translated == false  &&  hwnd  &&
          hwnd != desktopHandle  &&  hwnd != objectHandle;
        hwnd = IPARENTOF( hwnd ) )
  {  // Traverse up the parent chain to see if any window wants to
     // translate the message.
     IWindow* window = IWindow::windowWithHandle( hwnd );
     if ( window )
     {        // See if window has an accelerator table.
        IAccelTblHandle accelHandle = window->acceleratorHandle();
        if ( accelHandle )
        {
           translated = TranslateAccelerator( hwnd, accelHandle, &message );
        }
     }
  }
  return translated;
}
#endif

// Note: These functions are inlined on all platforms except Windows due to
//       variables they reference not being exported.
#ifdef IC_WIN
/*------------------------------------------------------------------------------
| IThread::defaultStatckSize                                                   |
------------------------------------------------------------------------------*/
#ifndef IC_MOTIF_FLAGNOP
unsigned long IThread :: defaultStackSize ( )
  {
  return IThread::dfltStackSize;
  }
#endif  // IC_MOTIF_FLAGNOP

/*------------------------------------------------------------------------------
| IThread::setDefaultStackSize                                                 |
------------------------------------------------------------------------------*/
#ifndef IC_MOTIF_FLAGNOP
void IThread :: setDefaultStackSize ( unsigned long newStkSize )
  {
  IThread::dfltStackSize = newStkSize;
  }
#endif  // IC_MOTIF_FLAGNOP

/*------------------------------------------------------------------------------
| IThread::setDefaultAutoInitGUI                                               |
------------------------------------------------------------------------------*/
void IThread :: setDefaultAutoInitGUI ( Boolean newFlag )
  {
  IThread::dfltAutoInitGUI = newFlag;
  }

/*------------------------------------------------------------------------------
| IThread::defaultQueueSize                                                    |
------------------------------------------------------------------------------*/
#ifndef IC_MOTIFWIN_FLAGNOP
long IThread :: defaultQueueSize ( )
  {
  return IThread::dfltQueueSize;
  }
#endif  // IC_MOTIFWIN_FLAGNOP

/*------------------------------------------------------------------------------
| IThread::setDefaultQueueSize                                                 |
------------------------------------------------------------------------------*/
#ifndef IC_MOTIFWIN_FLAGNOP
void IThread :: setDefaultQueueSize ( long newQSize )
  {
  IThread::dfltQueueSize = newQSize;
  }
#endif // IC_MOTIFWIN_FLAGNOP
#endif // IC_WIN
