/* NOSHIP */
#ifndef _IWINLSTS_
#define _IWINLSTS_
/*******************************************************************************
* FILE NAME: iwinlsts.hpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the definition of classes that are used to              *
*   manage the collection of windows and the collection of handlers.           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
#include <ihandler.hpp>
#include <ikeyset.h>
#include <ikssts.h>
#include <ireslock.hpp>
#include <iseq.h>
#include <ihshks.h>
#include <iwindow.hpp>

class IThreadId;

#pragma pack(4)

/*------------------------------------------------------------------------------
| IHandlerListObject class                                                     |
------------------------------------------------------------------------------*/
class IHandlerListObject : public IBase {
public:
  IHandlerListObject ( IHandler* handler );
 ~IHandlerListObject ();
  IHandlerListObject ( const IHandlerListObject& object );
IHandlerListObject
  &operator= ( const IHandlerListObject& object );
IHandler
 *fHandler;
Boolean
  fRemovePending;
};

/*------------------------------------------------------------------------------
| IHandlerList class                                                           |
------------------------------------------------------------------------------*/
class IHandlerList : public ISequence<IHandlerListObject> {
public:
  IHandlerList ();
 ~IHandlerList();

unsigned long
  totalHandlers,
  maxHandlers;
}; // IHandlerList

/*------------------------------------------------------------------------------
| IWindowList class                                                            |
------------------------------------------------------------------------------*/
class IWindowList : public IHashKeySet<IWindow*, unsigned long> {
public:
  IWindowList ( );
 ~IWindowList ( );

#ifdef IC_DEVELOP
static void
  incrementEventCount ( )
  { totalEvents++; }
static void
  incrementWindowAddedCount ( )
  { totalWindowsAdded++; }
static void
  incrementWindowDeletedCount ( )
  { totalWindowsDeleted++; }
static void
  incrementCacheCount ( )
  { cacheUsedCount++; }
static void
  dump ( );
#endif

static IWindowList
 *listForThread ( const IThreadId& threadId ),
 *current       ( );

static IPrivateResource
 &libraryKey ( );

#ifdef IC_DEVELOP
static unsigned long
  totalEvents,
  totalWindowsAdded,
  totalWindowsDeleted,
  cacheUsedCount;
#endif

static IPrivateResource
  libKey;
}; // IWindowList

/*------------------------------------------------------------------------------
| hash                                                                         |
------------------------------------------------------------------------------*/
inline unsigned long hash(unsigned long const& key, unsigned long size)
{
#ifdef IC_WIN
   // hwnds usually (always?) even.  Shift right then return modulo size.
   return (key >> 1) % size;
#else
   return key % size;
#endif

}

#ifdef IC_MOTIF
//============= Motif-unique collection: IRelatedHandlesList =================
// One per thread, this list holds elements which relate a widget to an IWindow.
// Only secondary widgets of compound controls are listed here.This collection
// is used in conjunction with this thread's IWindowList::current, which
// relates the primary widget of a compound control (or the only widget for
// a simple control) to it's IWindow instance. Unlike IWindowList, the elements
// here aren't actually IWindows (we don't want to create a lot of dummy ones);
// the elements are IRelatedHandleElement objects, defined here also.

/*------------------------------------------------------------------------------
| IRelatedHandleElement class                                                  |
------------------------------------------------------------------------------*/
class IRelatedHandleElement : public IBase {
public:
  IRelatedHandleElement (IWindowHandle aHwnd,
                         IWindow*      aPwin )
    : hwndKey((unsigned long)((void *)aHwnd) ), pwin(aPwin)
  { }
 ~IRelatedHandleElement ( )
  { }
IWindow
 *window ( )
  { return pwin; }
IWindowHandle
  relatedHandle ( )
  { return hwndKey; }

private:
  friend class IRelatedHandlesList;
  friend unsigned long const& key ( IRelatedHandleElement* const& pElement );

unsigned long
  hwndKey;
IWindow
 *pwin;
}; // IRelatedHandleElement

/*------------------------------------------------------------------------------
| IRelatedHandlesList class                                                    |
------------------------------------------------------------------------------*/
class IRelatedHandlesList : public IKeySet<IRelatedHandleElement*,
                                           unsigned long> {
public:
  IRelatedHandlesList ( );
 ~IRelatedHandlesList ( );

#ifdef IC_DEVELOP
static void
  incrementWindowAddedCount ( )
  { totalWindowsAdded++; }
static void
  dump ( );
#endif

static IPrivateResource
 &libraryKey ( );
static IRelatedHandlesList
 *current    ( );

static IPrivateResource
  libKey;
#ifdef IC_DEVELOP
static unsigned long
  totalWindowsAdded;
#endif
}; // IRelatedHandlesList
#endif  //IC_MOTIF

#ifdef IC_WIN
class IDialogControl : public IBase {
/*******************************************************************************
  This class represents controls in a dialog.

  In Windows, native dialog navigation depends on a function
  (IsDialogMessage) being called in the event loop.  This function requires
  that the window handle of the dialog be known to call it.  This generally
  conflicts with the class library model of making application frame windows
  and dialog windows basically interchangable.  Therefore, we don't call
  this function in the event loop ( IThread::processMsgs() ).

  For IWindow objects, we insure via default handlers that keystroke messages
  make it to the canvas or frame window to provide keyboard navigation.
  With dialogs created via a dialog template, however, we have unwrappered
  controls in the dialog which by default simply eat keyboard events,  perhaps
  after doing something with them.  The IDialogControl class is basically
  a "mini" IWindow whose function is to allow us to subclass dialog child
  controls.  In the subclass window procedure, we make sure keyboard
  events get to the dialog IWindow for processing.

*******************************************************************************/

public:
   IDialogControl   ( );
   IDialogControl   ( const IDialogControl& control) ;
   IDialogControl
      &operator=    ( const IDialogControl& control );

  ~IDialogControl   ();

  unsigned long fHwnd;           // HWND, as an unsigned long
  void*         fOriginalProc;   // WNDPROC, default window proc
  IWindowHandle fDialog;         // The parent dialog, typically IFrameWindow
};

/*******************************************************************************
// Key function for IDialogControl
*******************************************************************************/
inline unsigned long const& key (IDialogControl const& k)
{ return k.fHwnd; }


class IDialogControls : public IHashKeySet<IDialogControl, unsigned long> {
/*******************************************************************************
  This class represents a list of the controls in a dialog.  It is
  the list of IDialogControls for a frame window.

  The key into the collection is the window handle of the child control.

*******************************************************************************/
public:
  IDialogControls ();
  ~IDialogControls() ;
};



/*******************************************************************************
 pfnwpICControlProc is the subclass procedure for IDialogControl
*******************************************************************************/
unsigned long __stdcall pfnwpICControlProc(void*         hwnd,
                                           unsigned long ulMsg,
                                           void*         mp1,
                                           void*         mp2);

#endif

#pragma pack()

#endif // _IWINLSTS_
