/*******************************************************************************
* FILE NAME: ikeyevt.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in ikeyevt.hpp.                                                            *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

extern "C"
  {
  #define INCL_WININPUT
  #define INCL_NLS
  #include <iwindefs.h>
  }

#include <ikeyevt.hpp>
#include <icconst.h>
#include <iexcept.hpp>
#include <istring.hpp>


// Segment definitions
#ifdef IC_PAGETUNE
  #define _IKEYEVT_CPP_
  #include <ipagetun.h>
#endif


/*******************************************************************/
/* Initialize static flags used by the testing functions.          */
/*******************************************************************/
const unsigned long IKeyboardEvent :: ulAltFlag          ( 0x20000000 );
const unsigned long IKeyboardEvent :: ulUpTransitionFlag ( 0x80000000 );

/*------------------------------------------------------------------------------
| IKeyboardEvent::IKeyboardEvent                                               |
|                                                                              |
| Constructor here for page tuning.                                            |
------------------------------------------------------------------------------*/
IKeyboardEvent :: IKeyboardEvent ( const IEvent & evt )
  : IKeyboardEvent::Inherited ( evt ),
    fValidComposite(false),
    fInvalidComposite(false)
{ ; }


/*------------------------------------------------------------------------------
| IKeyboardEvent::~IKeyboardEvent                                              |
|                                                                              |
| Empty destructor here for page tuning.                                       |
------------------------------------------------------------------------------*/
IKeyboardEvent :: ~IKeyboardEvent ( )
{ ; }


/*------------------------------------------------------------------------------
| IKeyboardEvent::character                                                    |
|                                                                              |
| Return the SBCS character code for the event.  Before calling this function, |
| the user should first verify whether the event contains a character code by  |
| calling the isCharacter() function.                                          |
------------------------------------------------------------------------------*/
char  IKeyboardEvent::character ( ) const
{
  if (isCharacter() == false)
  {                               // no character data
     ITHROWLIBRARYERROR(IC_INVALIDCHARCODE,
                        IBaseErrorInfo::invalidRequest,
                        IException::recoverable);
  }

  return ( parameter1() );
}


/*------------------------------------------------------------------------------
| IKeyboardEvent::mixedCharacter                                               |
|                                                                              |
| Return the SBCS or DBCS character code for the event.  Before calling this   |
| function, the user should first verify whether the event contains a          |
| character code by calling the isCharacter() function.                        |
------------------------------------------------------------------------------*/
IString  IKeyboardEvent::mixedCharacter ( ) const
{
  if (isCharacter() == false)
  {                               // no character data
     ITHROWLIBRARYERROR(IC_INVALIDCHARCODE,
                        IBaseErrorInfo::invalidRequest,
                        IException::recoverable);
  } /* endif */

  IString strChar( (char)parameter1() );
  return strChar;
}


/*------------------------------------------------------------------------------
| IKeyboardEvent::scanCode                                                     |
|                                                                              |
| Get the hardware scan code for the key.  Before calling this function, the   |
| user should first verify whether the event contains a scan code by calling   |
| the isScanCode() function.                                                   |
------------------------------------------------------------------------------*/
unsigned long  IKeyboardEvent::scanCode ( ) const
{
  if (isScanCode() == false)
  {         // no scan code data (maybe event is from an input hook?)
     ITHROWLIBRARYERROR(IC_INVALIDSCANCODE,
                        IBaseErrorInfo::invalidRequest,
                        IException::recoverable);
  } /* endif */

  return parameter2().char3();
}


/*------------------------------------------------------------------------------
| IKeyboardEvent::repeatCount                                                  |
|                                                                              |
| Return the number of times this key event should be applied.                 |
------------------------------------------------------------------------------*/
unsigned long  IKeyboardEvent::repeatCount ( ) const
{
  return parameter2().number1();
}


/*------------------------------------------------------------------------------
| IKeyboardEvent::virtualKey                                                   |
|                                                                              |
| Get the virtual key value of the key event.  Before calling this function,   |
| the user should first verify whether the event contains a virtual key code   |
| by calling the isVirtual() function.                                         |
------------------------------------------------------------------------------*/
IKeyboardEvent::VirtualKey  IKeyboardEvent::virtualKey ( ) const
{
  VirtualKey vkey = other;        // assume no match found

  if (isVirtual() == false)
  {                               // not a virtual key
     ITHROWLIBRARYERROR(IC_INVALIDVIRTUALKEY,
                        IBaseErrorInfo::invalidRequest,
                        IException::recoverable);
  } /* endif */

  switch (parameter1())
  {                               // switch on the virtual key code
     case VK_BUTTON1:
        vkey = button1;
        break;
     case VK_BUTTON2:
        vkey = button2;
        break;
     case VK_BUTTON3:
        vkey = button3;
        break;
     case VK_BREAK:
        vkey = breakKey;
        break;
     case VK_BACKSPACE:
        vkey = backSpace;
        break;
     case VK_TAB:
        vkey = tab;
        // Check shift state to determine type of tab
        if ((isShiftDown()) && (isCtrlDown() == false) &&
            (isAltDown() == false))
          vkey = backTab;
        break;
     case VK_SHIFT:
        vkey = shift;
        break;
     case VK_CTRL:
        vkey = ctrl;
        break;
     case VK_PAUSE:
        vkey = pause;
        break;
     case VK_CAPSLOCK:
        vkey = capsLock;
        break;
     case VK_ESC:
        vkey = esc;
        break;
     case VK_SPACE:
        vkey = space;
        break;
     case VK_PAGEUP:
        vkey = pageUp;
        break;
     case VK_PAGEDOWN:
        vkey = pageDown;
        break;
     case VK_END:
        vkey = end;
        break;
     case VK_HOME:
        vkey = home;
        break;
     case VK_LEFT:
        vkey = left;
        break;
     case VK_UP:
        vkey = up;
        break;
     case VK_RIGHT:
        vkey = right;
        break;
     case VK_DOWN:
        vkey = down;
        break;
     case VK_INSERT:
        vkey = insert;
        break;
     case VK_DELETE:
        vkey = deleteKey;
        break;
     case VK_SCRLLOCK:
        vkey = scrollLock;
        break;
     case VK_NUMLOCK:
        vkey = numLock;
        break;
     case VK_ENTER:
        vkey = enter;
        // Check for extended key bit on
        if (parameter2().asUnsignedLong() & 0x01000000)
          vkey = newLine;
        break;
     case VK_F1:
        vkey = f1;
        break;
     case VK_F2:
        vkey = f2;
        break;
     case VK_F3:
        vkey = f3;
        break;
     case VK_F4:
        vkey = f4;
        break;
     case VK_F5:
        vkey = f5;
        break;
     case VK_F6:
        vkey = f6;
        break;
     case VK_F7:
        vkey = f7;
        break;
     case VK_F8:
        vkey = f8;
        break;
     case VK_F9:
        vkey = f9;
        break;
     case VK_F10:
        vkey = f10;
        break;
     case VK_F11:
        vkey = f11;
        break;
     case VK_F12:
        vkey = f12;
        break;
     case VK_F13:
        vkey = f13;
        break;
     case VK_F14:
        vkey = f14;
        break;
     case VK_F15:
        vkey = f15;
        break;
     case VK_F16:
        vkey = f16;
        break;
     case VK_F17:
        vkey = f17;
        break;
     case VK_F18:
        vkey = f18;
        break;
     case VK_F19:
        vkey = f19;
        break;
     case VK_F20:
        vkey = f20;
        break;
     case VK_F21:
        vkey = f21;
        break;
     case VK_F22:
        vkey = f22;
        break;
     case VK_F23:
        vkey = f23;
        break;
     case VK_F24:
        vkey = f24;
        break;

     // The below virtual keys are not supported because a
     // keyboard handler will not get the key press (F1, F10),
     // release (Alt), or both (PrintScreen, though maybe it can
     // get the key release) of these keys.  Supporting
     // enumerators for them implies that all presses and
     // releases of these keys can be processed.
     // The application can still process the non-missing events
     // for the below keys by overriding the key() function.
     case VK_ALT:                 // fall into default case
     case VK_PRINTSCRN:           // fall into default case
     default:
        break;
  } /* endswitch */

  return vkey;
}

/*------------------------------------------------------------------------------
| isSpecialCharacter                                                           |
|                                                                              |
| Local API for determining if the value passed is one of the keys which       |
| require special processing on our part.                                      |
------------------------------------------------------------------------------*/
Boolean isSpecialCharacter( unsigned long code )
{
  return (( code == VK_TAB ||
            code == VK_RETURN ||
            code == VK_BACK ||
            code == VK_ESCAPE ) ? true : false);
}

/*------------------------------------------------------------------------------
| IKeyboardEvent::isCharacter                                                  |
|                                                                              |
| Tell the caller whether this keyboard event is for a character key.          |
| To maintain compatibility with OS/2 design, certain keys which Windows       |
| converts into character messages will be excluded and will show up as        |
| virtual keys                                                                 |
------------------------------------------------------------------------------*/
IBase::Boolean IKeyboardEvent::isCharacter ( ) const
{
  return (( ((eventId() == WM_CHAR) ||
             (eventId() == WM_SYSCHAR)) &&
             (isSpecialCharacter( parameter1() ) == false)) ? true : false);
}

/*------------------------------------------------------------------------------
| IKeyboardEvent::isVirtual                                                    |
|                                                                              |
| Tell the caller whether this keyboard event is for a virtual key.            |
| To maintain compatibility with OS/2 design, certain keys which Windows       |
| converts into character messages will be included.                           |
------------------------------------------------------------------------------*/
IBase::Boolean IKeyboardEvent::isVirtual ( ) const
{
  Boolean brc = isSpecialCharacter( parameter1() );
  return (((((eventId() == WM_KEYDOWN) ||
             (eventId() == WM_SYSKEYDOWN) ||
             (eventId() == WM_KEYUP) ||
             (eventId() == WM_SYSKEYUP)) &&
            (brc == false)) ||
            (((eventId() == WM_CHAR) ||
              (eventId() == WM_SYSCHAR)) &&
             brc ) ) ? true : false);
}

/*------------------------------------------------------------------------------
| IBase::Boolean  IKeyboardEvent :: isShiftDown                                |
------------------------------------------------------------------------------*/
IBase::Boolean  IKeyboardEvent::isShiftDown ( ) const
{
  return (( GetKeyState(VK_SHIFT) & 0x8000 ) ? true : false );
}

/*------------------------------------------------------------------------------
| IBase::Boolean  IKeyboardEvent :: isCtrlDown                                 |
------------------------------------------------------------------------------*/
IBase::Boolean  IKeyboardEvent::isCtrlDown ( ) const
{
  return (( GetKeyState(VK_CONTROL) & 0x8000 ) ? true : false );
}

/*------------------------------------------------------------------------------
| IBase::Boolean  IKeyboardEvent :: isForComposite                             |
------------------------------------------------------------------------------*/
IBase::Boolean  IKeyboardEvent::isForComposite ( ) const
{
  return (((eventId() == WM_DEADCHAR) ||
           (eventId() == WM_SYSDEADCHAR)) ? true : false);
}

