/*******************************************************************************
* FILE NAME: iwindow0.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in iwindow.hpp that are common to all environments.                        *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
// Priority is 10 more (more negative) than controls or rest of IWindow,
// meaning statics in here get constructed before and destructed after
// controls or other IWindow statics.
#pragma priority( -2147482122 )

extern "C" {
  #include <iwindefs.h>
}

#include <iwindow.hpp>
#include <icconst.h>
#include <ievent.hpp>
#include <iexcept.hpp>
#include <ihandler.hpp>
#include <iplatfrm.hpp>
#include <ipoint.hpp>
#include <irect.hpp>
#include <istring.hpp>
#include <itrace.hpp>
#include <ithread.hpp>
#include <iwinlsts.hpp>
#include <iwinpriv.hpp>

// Segment definitions.
#ifdef IC_PAGETUNE
  #define _IWINDOW0_CPP_
  #include <ipagetun.h>
#endif


/*------------------------------------------------------------------------------
| Static Data.                                                                 |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLNonConst)
  IWindow::ExceptionFn* IWindowData::exceptionFunction = 0;
  IWindow::SiblingOrder IWindowData::siblingCreateOrder=IWindow::behindSiblings;

  ITheDesktopWindow* IWindowData::desktopWindow = 0;
  ITheObjectWindow*  IWindowData::objectWindow  = 0;
  IPrivateResource   IWindowList::libKey;

#ifdef IC_WIN
class IObjectWindowDeleter {
public:
 ~IObjectWindowDeleter ( )
  {
    if (IWindowData::objectWindow)
    {
       delete IWindowData::objectWindow;
       IWindowData::objectWindow = 0;
    }
  }
}; // IObjectWindowDeleter

IObjectWindowDeleter objectWindowDeleter;
unsigned long        propagationInProgress = 0;
#endif  // IC_WIN

#pragma data_seg()


#ifdef IC_DEVELOP
unsigned long IWindowList::totalEvents = 0;
unsigned long IWindowList::totalWindowsAdded = 0;
unsigned long IWindowList::totalWindowsDeleted = 0;
unsigned long IWindowList::cacheUsedCount = 0;
#endif


/*------------------------------------------------------------------------------
| ITheDesktopWindow::handle                                                    |
------------------------------------------------------------------------------*/
IWindowHandle ITheDesktopWindow::handle ( ) const
{
  return this->hwnd;
}

/*------------------------------------------------------------------------------
| ITheDesktopWindow::~ITheDesktopWindow                                        |
------------------------------------------------------------------------------*/
ITheDesktopWindow::~ITheDesktopWindow ( )
{ }


/*------------------------------------------------------------------------------
|ITheObjectWindow::~ITheObjectWindow                                           |
------------------------------------------------------------------------------*/
ITheObjectWindow::~ITheObjectWindow ( )
{
#ifdef IC_WIN
  IMODTRACE_DEVELOP("ITheObjectWindow::~ITheObjectWindow");
  IDESTROYWINDOW(this->hwnd);
#endif
}

/*------------------------------------------------------------------------------
|ITheObjectWindow::handle                                                      |
------------------------------------------------------------------------------*/
IWindowHandle ITheObjectWindow::handle ( ) const
{
  return this->hwnd;
}


/*------------------------------------------------------------------------------
| IWindow::handle                                                              |
------------------------------------------------------------------------------*/
IWindowHandle IWindow::handle ( ) const
{
  // Throw exception if window handle is not valid window.
  if ( !isValid() ) {
     ITHROWLIBRARYERROR( IC_INVALIDHANDLE,
                         IBaseErrorInfo::invalidRequest,
                         IException::recoverable );
  }
#ifdef IC_WIN
  return (HWND)this->hwnd;
#else
  return this->hwnd;
#endif
}

/*------------------------------------------------------------------------------
| IWindow::hide                                                                |
|                                                                              |
| Hide the window.                                                             |
------------------------------------------------------------------------------*/
IWindow& IWindow::hide ( )
{
  return show( false );
}

/*------------------------------------------------------------------------------
| IWindow::disableUpdate                                                       |
|                                                                              |
| Cause updates to the window not to be painted to the display.                |
------------------------------------------------------------------------------*/
IWindow& IWindow::disableUpdate ( )
{
  return enableUpdate( false );
}

/*------------------------------------------------------------------------------
| IWindow::disable                                                             |
|                                                                              |
| Disable the object - do not accept input.                                    |
------------------------------------------------------------------------------*/
IWindow& IWindow::disable ( )
{
  return enable( false );
}

/*------------------------------------------------------------------------------
| IWindow::setAutoDeleteObject                                                 |
|                                                                              |
| Set flag to delete the object (this pointer) when the                        |
| window handle is destroyed.                                                  |
------------------------------------------------------------------------------*/
IWindow& IWindow::setAutoDeleteObject ( Boolean autoDelete )
{
  if ( autoDelete )
     pWindowData->state |= autoDeleteObject;
  else
     pWindowData->state &= (unsigned long)(~autoDeleteObject);
  return *this;
}

/*------------------------------------------------------------------------------
| IWindow::setAutoDestroyWindow                                                |
|                                                                              |
| Set flag to destroy the window handle when the object                        |
| is destructed.                                                               |
------------------------------------------------------------------------------*/
IWindow& IWindow::setAutoDestroyWindow ( Boolean autoDestroy )
{
   if ( autoDestroy )
      pWindowData->state |= autoDestroyWindow;
   else
      pWindowData->state &= (unsigned long)(~autoDestroyWindow);
   return *this;
}

/*------------------------------------------------------------------------------
| IWindow::startHandlingEventsFor                                                  |
|                                                                              |
| This function subclasses the PM window procedure to                          |
| begin dispatch events.                                                       |
------------------------------------------------------------------------------*/
IWindow& IWindow::startHandlingEventsFor ( unsigned long identifier,
                                           IWindow*      parent )
{
  // Assertions on input parms.
  IASSERTPARM( parent != 0 );

  IWindowHandle wndh =
            IWindow::handleWithParent( identifier, parent->handle() );
  if ( wndh == 0 )
     ITHROWLIBRARYERROR( IC_NO_WINDOW_WITHID,
                         IBaseErrorInfo::invalidRequest,
                         IException::recoverable );
#ifdef IC_WIN
  // Set up the parent dialog as owner to simulate PM message routing.
  if ( ISTYLEOF( wndh ) & WS_CHILD )
  {
     // Owner is simulated.
     pWindowData->fhwndOwner = parent->handle();
  }
#endif

  return startHandlingEventsFor( wndh );
}

/*------------------------------------------------------------------------------
| IWindow::cleanUpHandler                                                      |
|                                                                              |
| Remove the argument handler from any windows it still might be trying        |
| to handle events for.                                                        |
|                                                                              |
| Currently, just searches the current thread's window list.  If the handler   |
| is handling events for windows on other threads, then log a "warning."       |
| At some point, we should try to remove this handler from windows on          |
| other threads.                                                               |
|                                                                              |
| This code presumes the window list will not be twiddled with by code on      |
| another thread.                                                              |
------------------------------------------------------------------------------*/
void IWindow::cleanUpHandler ( IHandler &handler )
{
  IMODTRACE_DEVELOP( "IWindow::cleanUpHandler" );
  IWindowList
   *pwinlist = IWindowList::current();
  if ( pwinlist !=0 && pwinlist->numberOfElements() != 0 )
  {
     IWindowList::Cursor cursor( *pwinlist );
     for ( cursor.setToFirst();
           cursor.isValid() && handler.numWindows;
           cursor.setToNext() )
        pwinlist->elementAt( cursor ) -> removeHandler( &handler );
     if ( handler.numWindows )
     {
        ITRACE_DEVELOP( "Active Handler Destroyed!" );
     }
  }
}

/*------------------------------------------------------------------------------
| IWindow::windowWithOwner                                                     |
|                                                                              |
| Search the internal collection for the iwindow with                          |
| the passed id and owner.                                                     |
------------------------------------------------------------------------------*/
IWindow* IWindow::windowWithOwner ( unsigned long identifier,
                                    const IWindow* owner,
                                    Boolean allThreads )
{
  IWindow* pwin;
  IWindowList* pwinlist;

  if ( allThreads )
  {
     IThreadId tid;
     IThread::Cursor threadCursor( false );
     while ( ! threadCursor.iteratedAll )
     {     // Repeat if not able to iterate through all threads
           // (cursor may have been invalidated by add/remove
           // done from another thread).
        for ( threadCursor.setToFirst();
              threadCursor.isValid();
              threadCursor.setToNext() )
        {
           tid = threadCursor.threadId();
           pwinlist = IWindowList::listForThread( tid );
           IWindowList::Cursor cursor( *pwinlist );
           for ( cursor.setToFirst(); cursor.isValid(); cursor.setToNext() )
           {
              pwin = pwinlist->elementAt( cursor );

              // Ignore entries without a PM window.  Calling
              // IThread::stopProcessingMsgs, or falling out of
              // IThread::processMsgs with an exception, can leave
              // invalid entries for a short time.
              if ( pwin->isValid()  &&
                   pwin->id() == identifier  &&
                   pwin->owner() == owner )
              {
                 return pwin;
              }
           } /* endfor */
        } /* endfor */
     } /* endwhile */
  }
  else
  {
     pwinlist = IWindowList::current();
     IWindowList::Cursor cursor( *pwinlist );
     unsigned long windowsRemaining;
     do
     {     // Repeat if not able to iterate through all windows
           // (cursor may have been invalidated by add/remove
           // done from another thread).
        windowsRemaining = pwinlist->numberOfElements();
        for ( cursor.setToFirst(); cursor.isValid(); cursor.setToNext() )
        {
           pwin = pwinlist->elementAt( cursor );
           if ( pwin->isValid()  &&
                pwin->id() == identifier  &&
                pwin->owner() == owner )
           {
              return pwin;
           }
           windowsRemaining--;
        } /* endfor */
     } /* enddo */
     while ( windowsRemaining );
  }  // end if/then/else

  return 0;
}

/*------------------------------------------------------------------------------
| IWindow::windowWithParent                                                    |
|                                                                              |
| Search the internal collection for the iwindow with                          |
| the passed id and parent.                                                    |
------------------------------------------------------------------------------*/
IWindow* IWindow::windowWithParent ( unsigned long identifier,
                                     const IWindow* parent,
                                     Boolean allThreads )
{
  IASSERTPARM( parent != 0 );
  return (windowWithHandle( IWINDOWFROMID( parent->handle(), identifier ),
                            allThreads ));
}

/*------------------------------------------------------------------------------
| IWindow::isWindowValid                                                       |
|                                                                              |
| Searches the internal collections for the iwindow to ensure it is valid.     |
------------------------------------------------------------------------------*/
Boolean IWindow::isWindowValid ( const IWindow* window )
{
  IWindow* pwin;
  IWindowList* pwinlist;

  IThreadId tid;
  IThread::Cursor threadCursor( false );
  while ( ! threadCursor.iteratedAll )
  {      // Repeat if not able to iterate through all threads
         // (cursor may have been invalidated by add/remove
         // done from another thread).
     for ( threadCursor.setToFirst();
           threadCursor.isValid();
           threadCursor.setToNext() )
     {
        tid = threadCursor.threadId();
        pwinlist = IWindowList::listForThread( tid );
        IWindowList::Cursor cursor( *pwinlist );
        for ( cursor.setToFirst(); cursor.isValid(); cursor.setToNext() )
        {
           pwin = pwinlist->elementAt( cursor );
           if ( pwin == window )
           {
              return true;
           }
        } /* endfor */
     } /* endfor */
  } /* endwhile */

  return false;
}

/*------------------------------------------------------------------------------
| IHandlerListObject::IHandlerListObject                                       |
------------------------------------------------------------------------------*/
IHandlerListObject::IHandlerListObject ( IHandler* handler )
  : fHandler ( handler ),
    fRemovePending ( false )
{ }

/*------------------------------------------------------------------------------
| IHandlerListObject::~IHandlerListObject                                      |
------------------------------------------------------------------------------*/
IHandlerListObject::~IHandlerListObject ( )
{
}


/*------------------------------------------------------------------------------
| IHandlerListObject::IHandlerListObject (copy constructor)                    |
------------------------------------------------------------------------------*/
IHandlerListObject::IHandlerListObject ( const IHandlerListObject& object )
  : fHandler ( object.fHandler ),
    fRemovePending ( object.fRemovePending )
{
}

/*------------------------------------------------------------------------------
| IHandlerListObject::operator=                                                |
------------------------------------------------------------------------------*/
IHandlerListObject& IHandlerListObject::operator= ( const IHandlerListObject& object )
{
  fHandler = object.fHandler;
  fRemovePending = object.fRemovePending;
  return *this;
}

/*------------------------------------------------------------------------------
| IHandlerList::IHandlerList                                                   |
------------------------------------------------------------------------------*/
IHandlerList::IHandlerList ( )
  : ISequence<IHandlerListObject>( 10 ),
    totalHandlers( 0 ),
    maxHandlers( 0 )
{ }

/*------------------------------------------------------------------------------
| IHandlerList::~IHandlerList                                                  |
|                                                                              |
| IHandlerList Destructor - total statistics.                                  |
------------------------------------------------------------------------------*/
IHandlerList::~IHandlerList ( )
{
  IMODTRACE_DEVELOP( "handlerList::~HandlerList" );
  ITRACE_DEVELOP( IString( "Total Handlers added: " ) +
                  IString( this->totalHandlers ));
  ITRACE_DEVELOP( IString( "Max Handlers: " ) + IString( this->maxHandlers ));
}

/*------------------------------------------------------------------------------
| IWindow::isLayoutDistorted                                                   |
------------------------------------------------------------------------------*/
IBase::Boolean IWindow::isLayoutDistorted( unsigned long layoutAttribute ) const
{
   if ( layoutAttribute & pWindowData->layoutChange )
      return true;
   else
      return false;
}

/*------------------------------------------------------------------------------
| IWindow::calcMinimumSize                                                     |
|                                                                              |
| Virtual function returning the minimun allowable size of                     |
| the window, used by canvas layout routined. Default                          |
| operation returns the minimun size set by users. Typically,                  |
| this should calculate the minimun size based on text and                     |
| current font setting.                                                        |
------------------------------------------------------------------------------*/
ISize IWindow::calcMinimumSize ( ) const
{
  IMODTRACE_DEVELOP( "Ctl::calcMinimumSize" );
  ITRACE_DEVELOP( "Default minimize size WIDTH=100 and HEIGHT=100 is used." );

  return ISize( 100,100 );
}

/*------------------------------------------------------------------------------
| IWindow::layoutAdjustment                                                    |
|                                                                              |
| Virtual function returning an IRectangle used to adjust the position         |
| and size of a window after layout.                                           |
------------------------------------------------------------------------------*/
IRectangle IWindow::layoutAdjustment ( ) const
{
  return IRectangle();
}

/*------------------------------------------------------------------------------
| IWindow::visibleRectangle                                                    |
|                                                                              |
| Virtual function returning an IRectangle representing the visible (painted)  |
| area of the control.                                                         |
------------------------------------------------------------------------------*/
IRectangle IWindow::visibleRectangle ( ) const
{
  return rect();
}

/*------------------------------------------------------------------------------
| IWindow::ExceptionFn::~ExceptionFn                                           |
|                                                                              |
| Keep the compiler from generating destructors.                               |
------------------------------------------------------------------------------*/
IWindow::ExceptionFn::~ExceptionFn ( )
{ }

/*------------------------------------------------------------------------------
| IWindow::setExceptionFunction                                                |
------------------------------------------------------------------------------*/
IWindow::ExceptionFn*
  IWindow::setExceptionFunction ( IWindow::ExceptionFn* pexcFn )
{
  IWindow::ExceptionFn* pcurrent = IWindowData::exceptionFunction;
  IWindowData::exceptionFunction = pexcFn;
  return pcurrent;
}

/*------------------------------------------------------------------------------
| IWindow::exceptionFunction                                                   |
------------------------------------------------------------------------------*/
IWindow::ExceptionFn* IWindow::exceptionFunction ( )
{
   return IWindowData::exceptionFunction;
}

/*------------------------------------------------------------------------------
| IWindow::handleException                                                     |
------------------------------------------------------------------------------*/
IBase::Boolean IWindow::handleException ( IException& exc,
                                          IEvent&     event )
{
  IWindow::ExceptionFn* pexcFn = exceptionFunction();
  if ( pexcFn )
     return pexcFn->handleException( exc, event );
  else
     return false;
}

/*------------------------------------------------------------------------------
| IWindow::childAt                                                             |
|                                                                              |
| Return the windowhandle from the argument cursor.                            |
------------------------------------------------------------------------------*/
IWindowHandle IWindow::childAt ( const ChildCursor &cursor ) const
{
  return cursor.hwnd;
}

/*------------------------------------------------------------------------------
| IWindow::setDefaultOrdering                                                  |
|                                                                              |
| Static function that sets the default placement of new windows.              |
------------------------------------------------------------------------------*/
void IWindow::setDefaultOrdering ( IWindow::SiblingOrder order )
{
  IWindowData::siblingCreateOrder = order;
}

/*------------------------------------------------------------------------------
| IWindow::defaultOrdering                                                     |
|                                                                              |
| Static function that returns the default placement of new windows.           |
------------------------------------------------------------------------------*/
IWindow::SiblingOrder IWindow::defaultOrdering ( )
{
  return IWindowData::siblingCreateOrder;
}

/*------------------------------------------------------------------------------
| IWindow::showSourceEmphasis                                                  |
|                                                                              |
| Overloaded to show source emphasis.                                          |
------------------------------------------------------------------------------*/
IWindow& IWindow::showSourceEmphasis ( Boolean /*unused*/ )
{
  return *this;
}

/*------------------------------------------------------------------------------
| IWindow::hideSourceEmphasis                                                  |
|                                                                              |
| Overloaded to remove source emphasis.                                        |
------------------------------------------------------------------------------*/
IWindow& IWindow::hideSourceEmphasis ( )
{
  return *this;
}

/*------------------------------------------------------------------------------
| IWindow::postEvent                                                           |
|                                                                              |
| Construct an IEvent from the arguments and post it to this window.           |
------------------------------------------------------------------------------*/
const IWindow& IWindow::postEvent ( unsigned long         eventId,
                                    const IEventParameter1 &parm1,
                                    const IEventParameter2 &parm2 ) const
{
  handle().postEvent( eventId, parm1, parm2 );
  return *this;
}

/*------------------------------------------------------------------------------
| IWindow::postEvent                                                           |
|                                                                              |
| Construct an IEvent from the arguments and post it to this window.           |
------------------------------------------------------------------------------*/
const IWindow& IWindow::postEvent ( const IEvent& event) const
{
  handle().postEvent( event.eventId(),
                      event.parameter1(),
                      event.parameter2() );
  return *this;
}

/*------------------------------------------------------------------------------
| IWindow::sendEvent                                                           |
|                                                                              |
| Construct an IEvent from the arguments and send it to this window.  Return   |
| the event result.                                                            |
------------------------------------------------------------------------------*/
IEventResult IWindow::sendEvent ( const IEvent& event ) const
{
  return handle().sendEvent( event.eventId(),
                             event.parameter1(),
                             event.parameter2() );
}

/*------------------------------------------------------------------------------
| IWindow::sendEvent                                                           |
|                                                                              |
| Construct an IEvent from the arguments and send it to this window.  Return   |
| the event result.                                                            |
------------------------------------------------------------------------------*/
IEventResult IWindow::sendEvent ( unsigned long   eventId,
                                  const IEventParameter1 &parm1,
                                  const IEventParameter2 &parm2 ) const
{
  return handle().sendEvent( eventId, parm1, parm2 );
}

/*------------------------------------------------------------------------------
| IWindow::setDefaultProcedure                                                 |
|                                                                              |
| Simply set the "default procedure" data member.  Returns the previous        |
| value.                                                                       |
------------------------------------------------------------------------------*/
IWinProc* IWindow::setDefaultProcedure ( IWinProc *newDefaultProc )
{
  IWinProc
   *oldDefaultProc = pWindowData->defaultProc;
  pWindowData->defaultProc = newDefaultProc;
  return oldDefaultProc;
}

/*------------------------------------------------------------------------------
| IWindow::isAutoDeleteObject                                                  |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IWindow ::isAutoDeleteObject ( ) const
{
  return ((pWindowData->state & IWindow::autoDeleteObject) ? true : false);
}

/*------------------------------------------------------------------------------
| IWindow::isAutoDestroyWindow                                                 |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IWindow::isAutoDestroyWindow ( ) const
{
  return ((pWindowData->state & IWindow::autoDestroyWindow) ? true : false);
}

/*------------------------------------------------------------------------------
| IWindow::deleteIsInProcess                                                   |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IWindow::deleteIsInProcess ( ) const
{
  return ((pWindowData->state & IWindow::deleteInProcess) ? true : false);
}

/*------------------------------------------------------------------------------
| IWindow::isPrimaryWindow                                                     |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IWindow::isPrimaryWindow ( ) const
{
  return ((pWindowData->state & IWindow::primaryWindow) ? true : false);
}

/*------------------------------------------------------------------------------
| IWindow::setDeleteInProcess                                                  |
|                                                                              |
------------------------------------------------------------------------------*/
IWindow& IWindow::setDeleteInProcess ( )
{
  pWindowData->state |= IWindow::deleteInProcess;
  return *this;
}

/*------------------------------------------------------------------------------
| IWindow::resetMinimumSize                                                    |
|                                                                              |
|  Reset the minimum size to the default.                                      |
------------------------------------------------------------------------------*/
IWindow& IWindow::resetMinimumSize ( )
{
  pWindowData->minimumSize = ISize( -1, -1 );
#ifdef IC_PMWIN
  this->saveMinimumSize( ISize( 0, 0 ) );
#endif
  setLayoutDistorted( IWindow::minimumSizeChanged, 0 );
  return *this;
}

/*------------------------------------------------------------------------------
| IWindow::setMinimumSize                                                      |
|                                                                              |
|  Set the minimum allowable size of the window. In the                        |
|  library this only apply to window on a canvas.                              |
------------------------------------------------------------------------------*/
IWindow& IWindow::setMinimumSize ( const ISize& sizMin )
{
  pWindowData->fSetMinSizeInProgress = true;
  pWindowData->minimumSize = sizMin;
  setLayoutDistorted( IWindow::minimumSizeChanged, 0 );
  pWindowData->fSetMinSizeInProgress = false;
  return *this;
}

/*------------------------------------------------------------------------------
| IWindow::minimumSize                                                         |
|                                                                              |
|  Returns the minimum allowable size set by the user's class                  |
------------------------------------------------------------------------------*/
ISize IWindow::minimumSize ( Boolean useCalculatedSize ) const
{
  ISize minSize;

  if ( useCalculatedSize  ||
       ( pWindowData->minimumSize.width() == -1  &&
         pWindowData->minimumSize.height() == -1 ) )
  {           // Get the value calculated by the window.
#ifdef IC_PMWIN
     if ( this->isLayoutDistorted( IWindow::minimumSizeChanged
                                    | IWindow::layoutChanged )  ||
          this->savedMinimumSize() == ISize( 0, 0 )  ||
          !( this->isMinimumSizeCachingEnabled() ) )
     {        // Need new calculated value.
        // No previously cached calculated value, the cached value is
        // no longer valid, or don't use cached values for this window.
#endif
        minSize = this->calcMinimumSize();
#ifdef IC_PMWIN
        (*(IWindow*)this)
         .saveMinimumSize( minSize )
         .setLayoutDistorted( 0, IWindow::minimumSizeChanged );
     }
     else
     {  // The last calculated value is still valid, so return it.
        minSize = this->savedMinimumSize();
     }
#endif
  }
  else
  {
     // Return the value stored by the application via
     // IWindow::setMInimumSize.
     minSize = pWindowData->minimumSize;
#ifdef IC_PMWIN
     (*(IWindow*)this)
      .setLayoutDistorted( 0, IWindow::minimumSizeChanged );
#endif
  }

  return minSize;
}

#ifdef IC_PMWIN
/*------------------------------------------------------------------------------
| IWindow::enableMinimumSizeCaching                                            |
|                                                                              |
|  Enable or disable minimum size caching as an optimization.  Enabling        |
|  caching causes IWindow::minimumSize to return the next calculated size      |
|  until the IWindow::minimumSizeChanged flag gets turned on.  Disabling       |
|  caching causes IWindow::minimumSize to call calcMinimumSize.                |
------------------------------------------------------------------------------*/
IWindow& IWindow::enableMinimumSizeCaching ( Boolean enableCaching )
{
  if ( enableCaching )
  {
     pWindowData->state |= IWindow::cacheMinimumSize;
  }
  else
  {
     pWindowData->state &= (unsigned long)(~IWindow::cacheMinimumSize);
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IWindow::disableMinimumSizeCaching                                           |
|                                                                              |
|  Disable minimum size caching as an optimization, so that                    |
|  IWindow::minimumSize will not bypass calcMinimumSize.                       |
------------------------------------------------------------------------------*/
IWindow& IWindow::disableMinimumSizeCaching ( )
{
  return this->enableMinimumSizeCaching( false );
}

/*------------------------------------------------------------------------------
| IWindow::isMinimumSizeCachingEnabled                                         |
|                                                                              |
|  Disable minimum size caching as an optimization, so that                    |
|  IWindow::minimumSize will not bypass calcMinimumSize.                       |
------------------------------------------------------------------------------*/
IBase::Boolean IWindow::isMinimumSizeCachingEnabled ( ) const
{
  return ( pWindowData->state & IWindow::cacheMinimumSize ) ? true : false;
}

/*------------------------------------------------------------------------------
| IWindow::saveMinimumSize                                                     |
|                                                                              |
|  Save the minimum size as an optimization.  IWindow::minimumSize             |
|  will return this value until the IWindow::minimumSizeChanged                |
|  flag gets turned on.                                                        |
------------------------------------------------------------------------------*/
IWindow& IWindow::saveMinimumSize ( const ISize& size )
{
  pWindowData->fCachedMinimumSize = size;
  return *this;
}

/*------------------------------------------------------------------------------
| IWindow::savedMinimumSize                                                    |
|                                                                              |
|  Returns the minimum size saved via IWindow::saveMinimumSize.                |
|  This function is used as an optimization step to avoid having               |
|  to constantly call calcMinimumSize.                                         |
------------------------------------------------------------------------------*/
ISize IWindow::savedMinimumSize ( ) const
{
  return pWindowData->fCachedMinimumSize;
}
#endif

/*------------------------------------------------------------------------------
| IWindow::helpId                                                              |
|                                                                              |
| Get the context help topic identifier.                                       |
------------------------------------------------------------------------------*/
unsigned long IWindow::helpId() const
{
  unsigned long ulId = pWindowData->helpId;
#ifdef IC_WIN
  if (!IPlatform::isWin32S())
    ulId = GetWindowContextHelpId( this->handle() );
#endif
  return ulId;
}

/*------------------------------------------------------------------------------
| IWindow::setHelpId                                                           |
|                                                                              |
| Set the context help topic identifier.                                       |
------------------------------------------------------------------------------*/
IWindow& IWindow::setHelpId( unsigned long ulId )
{
#ifdef IC_WIN
  if (!IPlatform::isWin32S())
    SetWindowContextHelpId( this->handle(), ulId );
#endif
  pWindowData->helpId = ulId;
  return *this;
}

/*------------------------------------------------------------------------------
| IWindow::defaultProcedure                                                    |
|                                                                              |
------------------------------------------------------------------------------*/
IWindow& IWindow::defaultProcedure ( IEvent& evt )
{
#ifdef IC_WIN
  IEventResult er( (unsigned long)
     CallWindowProc((FARPROC)pWindowData->defaultProc,
                                    (void*)evt.handle(),  (unsigned int)
#else
  IEventResult er( pWindowData->defaultProc(
                                    evt.handle().asUnsigned(),
#endif
                                    evt.eventId(),
                                    evt.parameter1(),
                                    evt.parameter2()) );
  evt.setResult( er );
  return *this;
}


#pragma info(none)
/*------------------------------------------------------------------------------
| IWindowList::IWindowList                                                     |
------------------------------------------------------------------------------*/
IWindowList::IWindowList ( )
  : IKeySetOnSortedTabularSequence<IWindow*, unsigned long>( 300 )
{ }
#pragma info(restore)

/*------------------------------------------------------------------------------
| IWindowList::~IWindowList                                                    |
|                                                                              |
| IWindowList Destructor - total statistics.                                   |
------------------------------------------------------------------------------*/
IWindowList::~IWindowList ( )
{
  IMODTRACE_DEVELOP( "IWindowList::~IWindowList" );

  this->removeAll();

#ifdef IC_DEVELOP
  ITRACE_DEVELOP( IString( "Total Windows added: " ) +
                  IString( this->totalWindowsAdded ));
  ITRACE_DEVELOP( IString( "Total Events: " ) +
                  IString( this->totalEvents ));
  ITRACE_DEVELOP( IString( "Successful Cache Requests: " ) +
                  IString( this->cacheUsedCount ));
#endif
}

/*------------------------------------------------------------------------------
| IWindowList::listForThread                                                   |
|                                                                              |
| Return the window list for the specified thread.                             |
------------------------------------------------------------------------------*/
IWindowList* IWindowList::listForThread( const IThreadId& threadId )
{
  IThread thread( threadId );
  IWindowList* list = thread.windowList();
  if ( list == 0 )
  {           // Create if it doesn't exist.
     list = new IWindowList();
     thread.setWindowList( list );
  }
  return list;
}

/*------------------------------------------------------------------------------
| IWindowList::current                                                         |
|                                                                              |
| Return the window list for the current thread.                               |
------------------------------------------------------------------------------*/
IWindowList* IWindowList::current ( )
{
  return IWindowList::listForThread( IThread::currentId() );
}

/*------------------------------------------------------------------------------
| IWindowList::libraryKey                                                      |
------------------------------------------------------------------------------*/
IPrivateResource& IWindowList::libraryKey ( )
{
  return libKey;
}

#ifdef IC_DEVELOP
/*------------------------------------------------------------------------------
| IWindowList::dump                                                            |
|                                                                              |
| Debug function to write the contents of the current windowlist.              |
------------------------------------------------------------------------------*/
void IWindowList::dump ( )
{
  IMODTRACE_DEVELOP( "IWindowList::dump" );
  IWindowList* pwinlist = IWindowList::current();
  IWindowList::Cursor cursor( *pwinlist );
  for ( cursor.setToFirst(); cursor.isValid(); cursor.setToNext() )
  {
     IWindow* pwin = pwinlist->elementAt( cursor );
     if ( pwin->isValid() )
     {            // Entry represents a valid window.
        ITRACE_DEVELOP( IString( "pwin =")+
                        IString( (unsigned long)pwin ).d2x() +
                        IString( " id =") +
                        IString( pwin->id() ) +
                        IString( " handle =")+
                        IString( pwin->handle().asUnsigned() ).d2x());
     }
     else
     {            // Window has been destroyed.
        ITRACE_DEVELOP( IString( "pwin =") +
                        IString( (unsigned long)pwin ).d2x() +
                        IString( " handle =none" ));
     }
  }
  return;
}
#endif  //IC_DEVELOP
